import { createTypeSpecLibrary, paramMessage, listServices, getEffectiveModelType, getVisibility, getProjectedName, getFriendlyName, resolveEncodedName, getNamespaceFullName, createDiagnosticCollector, ignoreDiagnostics, getDoc, getSummary, isErrorModel, getEncode, isNeverType, getKnownValues, getDiscriminator, SyntaxKind, isService, isTemplateDeclaration, isTemplateDeclarationOrInstance, projectProgram, getDeprecationDetails, isNullType, isVoidType, NoTarget, resolvePath, isGlobalNamespace, getService, isList, getPagingOperation, emitFile } from '@typespec/compiler';
import { getOperationId } from '@typespec/openapi';
import { getVersions, buildVersionProjections, getAddedOnVersions, getRemovedOnVersions, resolveVersions } from '@typespec/versioning';
import { isMetadata, getHttpPart, getHttpOperation, isPathParam, isQueryParam, isHeader, isBody, isCookieParam, getPathParamName, getCookieParamOptions, getQueryParamName, getHeaderFieldName, getQueryParamOptions, getHeaderFieldOptions, Visibility, getAuthentication, isStatusCode, getServers, isOrExtendsHttpFile } from '@typespec/http';
import { getUnionAsEnum, getLroMetadata, getPagedResult } from '@azure-tools/typespec-azure-core';
import { DuplicateTracker } from '@typespec/compiler/utils';

const ALIAS = Symbol.for('yaml.alias');
const DOC = Symbol.for('yaml.document');
const MAP = Symbol.for('yaml.map');
const PAIR = Symbol.for('yaml.pair');
const SCALAR = Symbol.for('yaml.scalar');
const SEQ = Symbol.for('yaml.seq');
const NODE_TYPE = Symbol.for('yaml.node.type');
const isAlias = (node) => !!node && typeof node === 'object' && node[NODE_TYPE] === ALIAS;
const isDocument = (node) => !!node && typeof node === 'object' && node[NODE_TYPE] === DOC;
const isMap = (node) => !!node && typeof node === 'object' && node[NODE_TYPE] === MAP;
const isPair = (node) => !!node && typeof node === 'object' && node[NODE_TYPE] === PAIR;
const isScalar = (node) => !!node && typeof node === 'object' && node[NODE_TYPE] === SCALAR;
const isSeq = (node) => !!node && typeof node === 'object' && node[NODE_TYPE] === SEQ;
function isCollection(node) {
    if (node && typeof node === 'object')
        switch (node[NODE_TYPE]) {
            case MAP:
            case SEQ:
                return true;
        }
    return false;
}
function isNode(node) {
    if (node && typeof node === 'object')
        switch (node[NODE_TYPE]) {
            case ALIAS:
            case MAP:
            case SCALAR:
            case SEQ:
                return true;
        }
    return false;
}
const hasAnchor = (node) => (isScalar(node) || isCollection(node)) && !!node.anchor;

const BREAK = Symbol('break visit');
const SKIP = Symbol('skip children');
const REMOVE = Symbol('remove node');
/**
 * Apply a visitor to an AST node or document.
 *
 * Walks through the tree (depth-first) starting from `node`, calling a
 * `visitor` function with three arguments:
 *   - `key`: For sequence values and map `Pair`, the node's index in the
 *     collection. Within a `Pair`, `'key'` or `'value'`, correspondingly.
 *     `null` for the root node.
 *   - `node`: The current node.
 *   - `path`: The ancestry of the current node.
 *
 * The return value of the visitor may be used to control the traversal:
 *   - `undefined` (default): Do nothing and continue
 *   - `visit.SKIP`: Do not visit the children of this node, continue with next
 *     sibling
 *   - `visit.BREAK`: Terminate traversal completely
 *   - `visit.REMOVE`: Remove the current node, then continue with the next one
 *   - `Node`: Replace the current node, then continue by visiting it
 *   - `number`: While iterating the items of a sequence or map, set the index
 *     of the next step. This is useful especially if the index of the current
 *     node has changed.
 *
 * If `visitor` is a single function, it will be called with all values
 * encountered in the tree, including e.g. `null` values. Alternatively,
 * separate visitor functions may be defined for each `Map`, `Pair`, `Seq`,
 * `Alias` and `Scalar` node. To define the same visitor function for more than
 * one node type, use the `Collection` (map and seq), `Value` (map, seq & scalar)
 * and `Node` (alias, map, seq & scalar) targets. Of all these, only the most
 * specific defined one will be used for each node.
 */
function visit(node, visitor) {
    const visitor_ = initVisitor(visitor);
    if (isDocument(node)) {
        const cd = visit_(null, node.contents, visitor_, Object.freeze([node]));
        if (cd === REMOVE)
            node.contents = null;
    }
    else
        visit_(null, node, visitor_, Object.freeze([]));
}
// Without the `as symbol` casts, TS declares these in the `visit`
// namespace using `var`, but then complains about that because
// `unique symbol` must be `const`.
/** Terminate visit traversal completely */
visit.BREAK = BREAK;
/** Do not visit the children of the current node */
visit.SKIP = SKIP;
/** Remove the current node */
visit.REMOVE = REMOVE;
function visit_(key, node, visitor, path) {
    const ctrl = callVisitor(key, node, visitor, path);
    if (isNode(ctrl) || isPair(ctrl)) {
        replaceNode(key, path, ctrl);
        return visit_(key, ctrl, visitor, path);
    }
    if (typeof ctrl !== 'symbol') {
        if (isCollection(node)) {
            path = Object.freeze(path.concat(node));
            for (let i = 0; i < node.items.length; ++i) {
                const ci = visit_(i, node.items[i], visitor, path);
                if (typeof ci === 'number')
                    i = ci - 1;
                else if (ci === BREAK)
                    return BREAK;
                else if (ci === REMOVE) {
                    node.items.splice(i, 1);
                    i -= 1;
                }
            }
        }
        else if (isPair(node)) {
            path = Object.freeze(path.concat(node));
            const ck = visit_('key', node.key, visitor, path);
            if (ck === BREAK)
                return BREAK;
            else if (ck === REMOVE)
                node.key = null;
            const cv = visit_('value', node.value, visitor, path);
            if (cv === BREAK)
                return BREAK;
            else if (cv === REMOVE)
                node.value = null;
        }
    }
    return ctrl;
}
function initVisitor(visitor) {
    if (typeof visitor === 'object' &&
        (visitor.Collection || visitor.Node || visitor.Value)) {
        return Object.assign({
            Alias: visitor.Node,
            Map: visitor.Node,
            Scalar: visitor.Node,
            Seq: visitor.Node
        }, visitor.Value && {
            Map: visitor.Value,
            Scalar: visitor.Value,
            Seq: visitor.Value
        }, visitor.Collection && {
            Map: visitor.Collection,
            Seq: visitor.Collection
        }, visitor);
    }
    return visitor;
}
function callVisitor(key, node, visitor, path) {
    if (typeof visitor === 'function')
        return visitor(key, node, path);
    if (isMap(node))
        return visitor.Map?.(key, node, path);
    if (isSeq(node))
        return visitor.Seq?.(key, node, path);
    if (isPair(node))
        return visitor.Pair?.(key, node, path);
    if (isScalar(node))
        return visitor.Scalar?.(key, node, path);
    if (isAlias(node))
        return visitor.Alias?.(key, node, path);
    return undefined;
}
function replaceNode(key, path, node) {
    const parent = path[path.length - 1];
    if (isCollection(parent)) {
        parent.items[key] = node;
    }
    else if (isPair(parent)) {
        if (key === 'key')
            parent.key = node;
        else
            parent.value = node;
    }
    else if (isDocument(parent)) {
        parent.contents = node;
    }
    else {
        const pt = isAlias(parent) ? 'alias' : 'scalar';
        throw new Error(`Cannot replace node with ${pt} parent`);
    }
}

const escapeChars = {
    '!': '%21',
    ',': '%2C',
    '[': '%5B',
    ']': '%5D',
    '{': '%7B',
    '}': '%7D'
};
const escapeTagName = (tn) => tn.replace(/[!,[\]{}]/g, ch => escapeChars[ch]);
class Directives {
    constructor(yaml, tags) {
        /**
         * The directives-end/doc-start marker `---`. If `null`, a marker may still be
         * included in the document's stringified representation.
         */
        this.docStart = null;
        /** The doc-end marker `...`.  */
        this.docEnd = false;
        this.yaml = Object.assign({}, Directives.defaultYaml, yaml);
        this.tags = Object.assign({}, Directives.defaultTags, tags);
    }
    clone() {
        const copy = new Directives(this.yaml, this.tags);
        copy.docStart = this.docStart;
        return copy;
    }
    /**
     * During parsing, get a Directives instance for the current document and
     * update the stream state according to the current version's spec.
     */
    atDocument() {
        const res = new Directives(this.yaml, this.tags);
        switch (this.yaml.version) {
            case '1.1':
                this.atNextDocument = true;
                break;
            case '1.2':
                this.atNextDocument = false;
                this.yaml = {
                    explicit: Directives.defaultYaml.explicit,
                    version: '1.2'
                };
                this.tags = Object.assign({}, Directives.defaultTags);
                break;
        }
        return res;
    }
    /**
     * @param onError - May be called even if the action was successful
     * @returns `true` on success
     */
    add(line, onError) {
        if (this.atNextDocument) {
            this.yaml = { explicit: Directives.defaultYaml.explicit, version: '1.1' };
            this.tags = Object.assign({}, Directives.defaultTags);
            this.atNextDocument = false;
        }
        const parts = line.trim().split(/[ \t]+/);
        const name = parts.shift();
        switch (name) {
            case '%TAG': {
                if (parts.length !== 2) {
                    onError(0, '%TAG directive should contain exactly two parts');
                    if (parts.length < 2)
                        return false;
                }
                const [handle, prefix] = parts;
                this.tags[handle] = prefix;
                return true;
            }
            case '%YAML': {
                this.yaml.explicit = true;
                if (parts.length !== 1) {
                    onError(0, '%YAML directive should contain exactly one part');
                    return false;
                }
                const [version] = parts;
                if (version === '1.1' || version === '1.2') {
                    this.yaml.version = version;
                    return true;
                }
                else {
                    const isValid = /^\d+\.\d+$/.test(version);
                    onError(6, `Unsupported YAML version ${version}`, isValid);
                    return false;
                }
            }
            default:
                onError(0, `Unknown directive ${name}`, true);
                return false;
        }
    }
    /**
     * Resolves a tag, matching handles to those defined in %TAG directives.
     *
     * @returns Resolved tag, which may also be the non-specific tag `'!'` or a
     *   `'!local'` tag, or `null` if unresolvable.
     */
    tagName(source, onError) {
        if (source === '!')
            return '!'; // non-specific tag
        if (source[0] !== '!') {
            onError(`Not a valid tag: ${source}`);
            return null;
        }
        if (source[1] === '<') {
            const verbatim = source.slice(2, -1);
            if (verbatim === '!' || verbatim === '!!') {
                onError(`Verbatim tags aren't resolved, so ${source} is invalid.`);
                return null;
            }
            if (source[source.length - 1] !== '>')
                onError('Verbatim tags must end with a >');
            return verbatim;
        }
        const [, handle, suffix] = source.match(/^(.*!)([^!]*)$/s);
        if (!suffix)
            onError(`The ${source} tag has no suffix`);
        const prefix = this.tags[handle];
        if (prefix) {
            try {
                return prefix + decodeURIComponent(suffix);
            }
            catch (error) {
                onError(String(error));
                return null;
            }
        }
        if (handle === '!')
            return source; // local tag
        onError(`Could not resolve tag: ${source}`);
        return null;
    }
    /**
     * Given a fully resolved tag, returns its printable string form,
     * taking into account current tag prefixes and defaults.
     */
    tagString(tag) {
        for (const [handle, prefix] of Object.entries(this.tags)) {
            if (tag.startsWith(prefix))
                return handle + escapeTagName(tag.substring(prefix.length));
        }
        return tag[0] === '!' ? tag : `!<${tag}>`;
    }
    toString(doc) {
        const lines = this.yaml.explicit
            ? [`%YAML ${this.yaml.version || '1.2'}`]
            : [];
        const tagEntries = Object.entries(this.tags);
        let tagNames;
        if (doc && tagEntries.length > 0 && isNode(doc.contents)) {
            const tags = {};
            visit(doc.contents, (_key, node) => {
                if (isNode(node) && node.tag)
                    tags[node.tag] = true;
            });
            tagNames = Object.keys(tags);
        }
        else
            tagNames = [];
        for (const [handle, prefix] of tagEntries) {
            if (handle === '!!' && prefix === 'tag:yaml.org,2002:')
                continue;
            if (!doc || tagNames.some(tn => tn.startsWith(prefix)))
                lines.push(`%TAG ${handle} ${prefix}`);
        }
        return lines.join('\n');
    }
}
Directives.defaultYaml = { explicit: false, version: '1.2' };
Directives.defaultTags = { '!!': 'tag:yaml.org,2002:' };

/**
 * Verify that the input string is a valid anchor.
 *
 * Will throw on errors.
 */
function anchorIsValid(anchor) {
    if (/[\x00-\x19\s,[\]{}]/.test(anchor)) {
        const sa = JSON.stringify(anchor);
        const msg = `Anchor must not contain whitespace or control characters: ${sa}`;
        throw new Error(msg);
    }
    return true;
}
function anchorNames(root) {
    const anchors = new Set();
    visit(root, {
        Value(_key, node) {
            if (node.anchor)
                anchors.add(node.anchor);
        }
    });
    return anchors;
}
/** Find a new anchor name with the given `prefix` and a one-indexed suffix. */
function findNewAnchor(prefix, exclude) {
    for (let i = 1; true; ++i) {
        const name = `${prefix}${i}`;
        if (!exclude.has(name))
            return name;
    }
}
function createNodeAnchors(doc, prefix) {
    const aliasObjects = [];
    const sourceObjects = new Map();
    let prevAnchors = null;
    return {
        onAnchor: (source) => {
            aliasObjects.push(source);
            if (!prevAnchors)
                prevAnchors = anchorNames(doc);
            const anchor = findNewAnchor(prefix, prevAnchors);
            prevAnchors.add(anchor);
            return anchor;
        },
        /**
         * With circular references, the source node is only resolved after all
         * of its child nodes are. This is why anchors are set only after all of
         * the nodes have been created.
         */
        setAnchors: () => {
            for (const source of aliasObjects) {
                const ref = sourceObjects.get(source);
                if (typeof ref === 'object' &&
                    ref.anchor &&
                    (isScalar(ref.node) || isCollection(ref.node))) {
                    ref.node.anchor = ref.anchor;
                }
                else {
                    const error = new Error('Failed to resolve repeated object (this should not happen)');
                    error.source = source;
                    throw error;
                }
            }
        },
        sourceObjects
    };
}

/**
 * Applies the JSON.parse reviver algorithm as defined in the ECMA-262 spec,
 * in section 24.5.1.1 "Runtime Semantics: InternalizeJSONProperty" of the
 * 2021 edition: https://tc39.es/ecma262/#sec-json.parse
 *
 * Includes extensions for handling Map and Set objects.
 */
function applyReviver(reviver, obj, key, val) {
    if (val && typeof val === 'object') {
        if (Array.isArray(val)) {
            for (let i = 0, len = val.length; i < len; ++i) {
                const v0 = val[i];
                const v1 = applyReviver(reviver, val, String(i), v0);
                // eslint-disable-next-line @typescript-eslint/no-array-delete
                if (v1 === undefined)
                    delete val[i];
                else if (v1 !== v0)
                    val[i] = v1;
            }
        }
        else if (val instanceof Map) {
            for (const k of Array.from(val.keys())) {
                const v0 = val.get(k);
                const v1 = applyReviver(reviver, val, k, v0);
                if (v1 === undefined)
                    val.delete(k);
                else if (v1 !== v0)
                    val.set(k, v1);
            }
        }
        else if (val instanceof Set) {
            for (const v0 of Array.from(val)) {
                const v1 = applyReviver(reviver, val, v0, v0);
                if (v1 === undefined)
                    val.delete(v0);
                else if (v1 !== v0) {
                    val.delete(v0);
                    val.add(v1);
                }
            }
        }
        else {
            for (const [k, v0] of Object.entries(val)) {
                const v1 = applyReviver(reviver, val, k, v0);
                if (v1 === undefined)
                    delete val[k];
                else if (v1 !== v0)
                    val[k] = v1;
            }
        }
    }
    return reviver.call(obj, key, val);
}

/**
 * Recursively convert any node or its contents to native JavaScript
 *
 * @param value - The input value
 * @param arg - If `value` defines a `toJSON()` method, use this
 *   as its first argument
 * @param ctx - Conversion context, originally set in Document#toJS(). If
 *   `{ keep: true }` is not set, output should be suitable for JSON
 *   stringification.
 */
function toJS(value, arg, ctx) {
    // eslint-disable-next-line @typescript-eslint/no-unsafe-return
    if (Array.isArray(value))
        return value.map((v, i) => toJS(v, String(i), ctx));
    if (value && typeof value.toJSON === 'function') {
        // eslint-disable-next-line @typescript-eslint/no-unsafe-call
        if (!ctx || !hasAnchor(value))
            return value.toJSON(arg, ctx);
        const data = { aliasCount: 0, count: 1, res: undefined };
        ctx.anchors.set(value, data);
        ctx.onCreate = res => {
            data.res = res;
            delete ctx.onCreate;
        };
        const res = value.toJSON(arg, ctx);
        if (ctx.onCreate)
            ctx.onCreate(res);
        return res;
    }
    if (typeof value === 'bigint' && !ctx?.keep)
        return Number(value);
    return value;
}

class NodeBase {
    constructor(type) {
        Object.defineProperty(this, NODE_TYPE, { value: type });
    }
    /** Create a copy of this node.  */
    clone() {
        const copy = Object.create(Object.getPrototypeOf(this), Object.getOwnPropertyDescriptors(this));
        if (this.range)
            copy.range = this.range.slice();
        return copy;
    }
    /** A plain JavaScript representation of this node. */
    toJS(doc, { mapAsMap, maxAliasCount, onAnchor, reviver } = {}) {
        if (!isDocument(doc))
            throw new TypeError('A document argument is required');
        const ctx = {
            anchors: new Map(),
            doc,
            keep: true,
            mapAsMap: mapAsMap === true,
            mapKeyWarned: false,
            maxAliasCount: typeof maxAliasCount === 'number' ? maxAliasCount : 100
        };
        const res = toJS(this, '', ctx);
        if (typeof onAnchor === 'function')
            for (const { count, res } of ctx.anchors.values())
                onAnchor(res, count);
        return typeof reviver === 'function'
            ? applyReviver(reviver, { '': res }, '', res)
            : res;
    }
}

class Alias extends NodeBase {
    constructor(source) {
        super(ALIAS);
        this.source = source;
        Object.defineProperty(this, 'tag', {
            set() {
                throw new Error('Alias nodes cannot have tags');
            }
        });
    }
    /**
     * Resolve the value of this alias within `doc`, finding the last
     * instance of the `source` anchor before this node.
     */
    resolve(doc) {
        let found = undefined;
        visit(doc, {
            Node: (_key, node) => {
                if (node === this)
                    return visit.BREAK;
                if (node.anchor === this.source)
                    found = node;
            }
        });
        return found;
    }
    toJSON(_arg, ctx) {
        if (!ctx)
            return { source: this.source };
        const { anchors, doc, maxAliasCount } = ctx;
        const source = this.resolve(doc);
        if (!source) {
            const msg = `Unresolved alias (the anchor must be set before the alias): ${this.source}`;
            throw new ReferenceError(msg);
        }
        let data = anchors.get(source);
        if (!data) {
            // Resolve anchors for Node.prototype.toJS()
            toJS(source, null, ctx);
            data = anchors.get(source);
        }
        /* istanbul ignore if */
        if (!data || data.res === undefined) {
            const msg = 'This should not happen: Alias anchor was not resolved?';
            throw new ReferenceError(msg);
        }
        if (maxAliasCount >= 0) {
            data.count += 1;
            if (data.aliasCount === 0)
                data.aliasCount = getAliasCount(doc, source, anchors);
            if (data.count * data.aliasCount > maxAliasCount) {
                const msg = 'Excessive alias count indicates a resource exhaustion attack';
                throw new ReferenceError(msg);
            }
        }
        return data.res;
    }
    toString(ctx, _onComment, _onChompKeep) {
        const src = `*${this.source}`;
        if (ctx) {
            anchorIsValid(this.source);
            if (ctx.options.verifyAliasOrder && !ctx.anchors.has(this.source)) {
                const msg = `Unresolved alias (the anchor must be set before the alias): ${this.source}`;
                throw new Error(msg);
            }
            if (ctx.implicitKey)
                return `${src} `;
        }
        return src;
    }
}
function getAliasCount(doc, node, anchors) {
    if (isAlias(node)) {
        const source = node.resolve(doc);
        const anchor = anchors && source && anchors.get(source);
        return anchor ? anchor.count * anchor.aliasCount : 0;
    }
    else if (isCollection(node)) {
        let count = 0;
        for (const item of node.items) {
            const c = getAliasCount(doc, item, anchors);
            if (c > count)
                count = c;
        }
        return count;
    }
    else if (isPair(node)) {
        const kc = getAliasCount(doc, node.key, anchors);
        const vc = getAliasCount(doc, node.value, anchors);
        return Math.max(kc, vc);
    }
    return 1;
}

const isScalarValue = (value) => !value || (typeof value !== 'function' && typeof value !== 'object');
class Scalar extends NodeBase {
    constructor(value) {
        super(SCALAR);
        this.value = value;
    }
    toJSON(arg, ctx) {
        return ctx?.keep ? this.value : toJS(this.value, arg, ctx);
    }
    toString() {
        return String(this.value);
    }
}
Scalar.BLOCK_FOLDED = 'BLOCK_FOLDED';
Scalar.BLOCK_LITERAL = 'BLOCK_LITERAL';
Scalar.PLAIN = 'PLAIN';
Scalar.QUOTE_DOUBLE = 'QUOTE_DOUBLE';
Scalar.QUOTE_SINGLE = 'QUOTE_SINGLE';

const defaultTagPrefix = 'tag:yaml.org,2002:';
function findTagObject(value, tagName, tags) {
    if (tagName) {
        const match = tags.filter(t => t.tag === tagName);
        const tagObj = match.find(t => !t.format) ?? match[0];
        if (!tagObj)
            throw new Error(`Tag ${tagName} not found`);
        return tagObj;
    }
    return tags.find(t => t.identify?.(value) && !t.format);
}
function createNode(value, tagName, ctx) {
    if (isDocument(value))
        value = value.contents;
    if (isNode(value))
        return value;
    if (isPair(value)) {
        const map = ctx.schema[MAP].createNode?.(ctx.schema, null, ctx);
        map.items.push(value);
        return map;
    }
    if (value instanceof String ||
        value instanceof Number ||
        value instanceof Boolean ||
        (typeof BigInt !== 'undefined' && value instanceof BigInt) // not supported everywhere
    ) {
        // https://tc39.es/ecma262/#sec-serializejsonproperty
        value = value.valueOf();
    }
    const { aliasDuplicateObjects, onAnchor, onTagObj, schema, sourceObjects } = ctx;
    // Detect duplicate references to the same object & use Alias nodes for all
    // after first. The `ref` wrapper allows for circular references to resolve.
    let ref = undefined;
    if (aliasDuplicateObjects && value && typeof value === 'object') {
        ref = sourceObjects.get(value);
        if (ref) {
            if (!ref.anchor)
                ref.anchor = onAnchor(value);
            return new Alias(ref.anchor);
        }
        else {
            ref = { anchor: null, node: null };
            sourceObjects.set(value, ref);
        }
    }
    if (tagName?.startsWith('!!'))
        tagName = defaultTagPrefix + tagName.slice(2);
    let tagObj = findTagObject(value, tagName, schema.tags);
    if (!tagObj) {
        if (value && typeof value.toJSON === 'function') {
            // eslint-disable-next-line @typescript-eslint/no-unsafe-call
            value = value.toJSON();
        }
        if (!value || typeof value !== 'object') {
            const node = new Scalar(value);
            if (ref)
                ref.node = node;
            return node;
        }
        tagObj =
            value instanceof Map
                ? schema[MAP]
                : Symbol.iterator in Object(value)
                    ? schema[SEQ]
                    : schema[MAP];
    }
    if (onTagObj) {
        onTagObj(tagObj);
        delete ctx.onTagObj;
    }
    const node = tagObj?.createNode
        ? tagObj.createNode(ctx.schema, value, ctx)
        : typeof tagObj?.nodeClass?.from === 'function'
            ? tagObj.nodeClass.from(ctx.schema, value, ctx)
            : new Scalar(value);
    if (tagName)
        node.tag = tagName;
    else if (!tagObj.default)
        node.tag = tagObj.tag;
    if (ref)
        ref.node = node;
    return node;
}

function collectionFromPath(schema, path, value) {
    let v = value;
    for (let i = path.length - 1; i >= 0; --i) {
        const k = path[i];
        if (typeof k === 'number' && Number.isInteger(k) && k >= 0) {
            const a = [];
            a[k] = v;
            v = a;
        }
        else {
            v = new Map([[k, v]]);
        }
    }
    return createNode(v, undefined, {
        aliasDuplicateObjects: false,
        keepUndefined: false,
        onAnchor: () => {
            throw new Error('This should not happen, please report a bug.');
        },
        schema,
        sourceObjects: new Map()
    });
}
// Type guard is intentionally a little wrong so as to be more useful,
// as it does not cover untypable empty non-string iterables (e.g. []).
const isEmptyPath = (path) => path == null ||
    (typeof path === 'object' && !!path[Symbol.iterator]().next().done);
class Collection extends NodeBase {
    constructor(type, schema) {
        super(type);
        Object.defineProperty(this, 'schema', {
            value: schema,
            configurable: true,
            enumerable: false,
            writable: true
        });
    }
    /**
     * Create a copy of this collection.
     *
     * @param schema - If defined, overwrites the original's schema
     */
    clone(schema) {
        const copy = Object.create(Object.getPrototypeOf(this), Object.getOwnPropertyDescriptors(this));
        if (schema)
            copy.schema = schema;
        copy.items = copy.items.map(it => isNode(it) || isPair(it) ? it.clone(schema) : it);
        if (this.range)
            copy.range = this.range.slice();
        return copy;
    }
    /**
     * Adds a value to the collection. For `!!map` and `!!omap` the value must
     * be a Pair instance or a `{ key, value }` object, which may not have a key
     * that already exists in the map.
     */
    addIn(path, value) {
        if (isEmptyPath(path))
            this.add(value);
        else {
            const [key, ...rest] = path;
            const node = this.get(key, true);
            if (isCollection(node))
                node.addIn(rest, value);
            else if (node === undefined && this.schema)
                this.set(key, collectionFromPath(this.schema, rest, value));
            else
                throw new Error(`Expected YAML collection at ${key}. Remaining path: ${rest}`);
        }
    }
    /**
     * Removes a value from the collection.
     * @returns `true` if the item was found and removed.
     */
    deleteIn(path) {
        const [key, ...rest] = path;
        if (rest.length === 0)
            return this.delete(key);
        const node = this.get(key, true);
        if (isCollection(node))
            return node.deleteIn(rest);
        else
            throw new Error(`Expected YAML collection at ${key}. Remaining path: ${rest}`);
    }
    /**
     * Returns item at `key`, or `undefined` if not found. By default unwraps
     * scalar values from their surrounding node; to disable set `keepScalar` to
     * `true` (collections are always returned intact).
     */
    getIn(path, keepScalar) {
        const [key, ...rest] = path;
        const node = this.get(key, true);
        if (rest.length === 0)
            return !keepScalar && isScalar(node) ? node.value : node;
        else
            return isCollection(node) ? node.getIn(rest, keepScalar) : undefined;
    }
    hasAllNullValues(allowScalar) {
        return this.items.every(node => {
            if (!isPair(node))
                return false;
            const n = node.value;
            return (n == null ||
                (allowScalar &&
                    isScalar(n) &&
                    n.value == null &&
                    !n.commentBefore &&
                    !n.comment &&
                    !n.tag));
        });
    }
    /**
     * Checks if the collection includes a value with the key `key`.
     */
    hasIn(path) {
        const [key, ...rest] = path;
        if (rest.length === 0)
            return this.has(key);
        const node = this.get(key, true);
        return isCollection(node) ? node.hasIn(rest) : false;
    }
    /**
     * Sets a value in this collection. For `!!set`, `value` needs to be a
     * boolean to add/remove the item from the set.
     */
    setIn(path, value) {
        const [key, ...rest] = path;
        if (rest.length === 0) {
            this.set(key, value);
        }
        else {
            const node = this.get(key, true);
            if (isCollection(node))
                node.setIn(rest, value);
            else if (node === undefined && this.schema)
                this.set(key, collectionFromPath(this.schema, rest, value));
            else
                throw new Error(`Expected YAML collection at ${key}. Remaining path: ${rest}`);
        }
    }
}

/**
 * Stringifies a comment.
 *
 * Empty comment lines are left empty,
 * lines consisting of a single space are replaced by `#`,
 * and all other lines are prefixed with a `#`.
 */
const stringifyComment = (str) => str.replace(/^(?!$)(?: $)?/gm, '#');
function indentComment(comment, indent) {
    if (/^\n+$/.test(comment))
        return comment.substring(1);
    return indent ? comment.replace(/^(?! *$)/gm, indent) : comment;
}
const lineComment = (str, indent, comment) => str.endsWith('\n')
    ? indentComment(comment, indent)
    : comment.includes('\n')
        ? '\n' + indentComment(comment, indent)
        : (str.endsWith(' ') ? '' : ' ') + comment;

const FOLD_FLOW = 'flow';
const FOLD_BLOCK = 'block';
const FOLD_QUOTED = 'quoted';
/**
 * Tries to keep input at up to `lineWidth` characters, splitting only on spaces
 * not followed by newlines or spaces unless `mode` is `'quoted'`. Lines are
 * terminated with `\n` and started with `indent`.
 */
function foldFlowLines(text, indent, mode = 'flow', { indentAtStart, lineWidth = 80, minContentWidth = 20, onFold, onOverflow } = {}) {
    if (!lineWidth || lineWidth < 0)
        return text;
    if (lineWidth < minContentWidth)
        minContentWidth = 0;
    const endStep = Math.max(1 + minContentWidth, 1 + lineWidth - indent.length);
    if (text.length <= endStep)
        return text;
    const folds = [];
    const escapedFolds = {};
    let end = lineWidth - indent.length;
    if (typeof indentAtStart === 'number') {
        if (indentAtStart > lineWidth - Math.max(2, minContentWidth))
            folds.push(0);
        else
            end = lineWidth - indentAtStart;
    }
    let split = undefined;
    let prev = undefined;
    let overflow = false;
    let i = -1;
    let escStart = -1;
    let escEnd = -1;
    if (mode === FOLD_BLOCK) {
        i = consumeMoreIndentedLines(text, i, indent.length);
        if (i !== -1)
            end = i + endStep;
    }
    for (let ch; (ch = text[(i += 1)]);) {
        if (mode === FOLD_QUOTED && ch === '\\') {
            escStart = i;
            switch (text[i + 1]) {
                case 'x':
                    i += 3;
                    break;
                case 'u':
                    i += 5;
                    break;
                case 'U':
                    i += 9;
                    break;
                default:
                    i += 1;
            }
            escEnd = i;
        }
        if (ch === '\n') {
            if (mode === FOLD_BLOCK)
                i = consumeMoreIndentedLines(text, i, indent.length);
            end = i + indent.length + endStep;
            split = undefined;
        }
        else {
            if (ch === ' ' &&
                prev &&
                prev !== ' ' &&
                prev !== '\n' &&
                prev !== '\t') {
                // space surrounded by non-space can be replaced with newline + indent
                const next = text[i + 1];
                if (next && next !== ' ' && next !== '\n' && next !== '\t')
                    split = i;
            }
            if (i >= end) {
                if (split) {
                    folds.push(split);
                    end = split + endStep;
                    split = undefined;
                }
                else if (mode === FOLD_QUOTED) {
                    // white-space collected at end may stretch past lineWidth
                    while (prev === ' ' || prev === '\t') {
                        prev = ch;
                        ch = text[(i += 1)];
                        overflow = true;
                    }
                    // Account for newline escape, but don't break preceding escape
                    const j = i > escEnd + 1 ? i - 2 : escStart - 1;
                    // Bail out if lineWidth & minContentWidth are shorter than an escape string
                    if (escapedFolds[j])
                        return text;
                    folds.push(j);
                    escapedFolds[j] = true;
                    end = j + endStep;
                    split = undefined;
                }
                else {
                    overflow = true;
                }
            }
        }
        prev = ch;
    }
    if (overflow && onOverflow)
        onOverflow();
    if (folds.length === 0)
        return text;
    if (onFold)
        onFold();
    let res = text.slice(0, folds[0]);
    for (let i = 0; i < folds.length; ++i) {
        const fold = folds[i];
        const end = folds[i + 1] || text.length;
        if (fold === 0)
            res = `\n${indent}${text.slice(0, end)}`;
        else {
            if (mode === FOLD_QUOTED && escapedFolds[fold])
                res += `${text[fold]}\\`;
            res += `\n${indent}${text.slice(fold + 1, end)}`;
        }
    }
    return res;
}
/**
 * Presumes `i + 1` is at the start of a line
 * @returns index of last newline in more-indented block
 */
function consumeMoreIndentedLines(text, i, indent) {
    let end = i;
    let start = i + 1;
    let ch = text[start];
    while (ch === ' ' || ch === '\t') {
        if (i < start + indent) {
            ch = text[++i];
        }
        else {
            do {
                ch = text[++i];
            } while (ch && ch !== '\n');
            end = i;
            start = i + 1;
            ch = text[start];
        }
    }
    return end;
}

const getFoldOptions = (ctx, isBlock) => ({
    indentAtStart: isBlock ? ctx.indent.length : ctx.indentAtStart,
    lineWidth: ctx.options.lineWidth,
    minContentWidth: ctx.options.minContentWidth
});
// Also checks for lines starting with %, as parsing the output as YAML 1.1 will
// presume that's starting a new document.
const containsDocumentMarker = (str) => /^(%|---|\.\.\.)/m.test(str);
function lineLengthOverLimit(str, lineWidth, indentLength) {
    if (!lineWidth || lineWidth < 0)
        return false;
    const limit = lineWidth - indentLength;
    const strLen = str.length;
    if (strLen <= limit)
        return false;
    for (let i = 0, start = 0; i < strLen; ++i) {
        if (str[i] === '\n') {
            if (i - start > limit)
                return true;
            start = i + 1;
            if (strLen - start <= limit)
                return false;
        }
    }
    return true;
}
function doubleQuotedString(value, ctx) {
    const json = JSON.stringify(value);
    if (ctx.options.doubleQuotedAsJSON)
        return json;
    const { implicitKey } = ctx;
    const minMultiLineLength = ctx.options.doubleQuotedMinMultiLineLength;
    const indent = ctx.indent || (containsDocumentMarker(value) ? '  ' : '');
    let str = '';
    let start = 0;
    for (let i = 0, ch = json[i]; ch; ch = json[++i]) {
        if (ch === ' ' && json[i + 1] === '\\' && json[i + 2] === 'n') {
            // space before newline needs to be escaped to not be folded
            str += json.slice(start, i) + '\\ ';
            i += 1;
            start = i;
            ch = '\\';
        }
        if (ch === '\\')
            switch (json[i + 1]) {
                case 'u':
                    {
                        str += json.slice(start, i);
                        const code = json.substr(i + 2, 4);
                        switch (code) {
                            case '0000':
                                str += '\\0';
                                break;
                            case '0007':
                                str += '\\a';
                                break;
                            case '000b':
                                str += '\\v';
                                break;
                            case '001b':
                                str += '\\e';
                                break;
                            case '0085':
                                str += '\\N';
                                break;
                            case '00a0':
                                str += '\\_';
                                break;
                            case '2028':
                                str += '\\L';
                                break;
                            case '2029':
                                str += '\\P';
                                break;
                            default:
                                if (code.substr(0, 2) === '00')
                                    str += '\\x' + code.substr(2);
                                else
                                    str += json.substr(i, 6);
                        }
                        i += 5;
                        start = i + 1;
                    }
                    break;
                case 'n':
                    if (implicitKey ||
                        json[i + 2] === '"' ||
                        json.length < minMultiLineLength) {
                        i += 1;
                    }
                    else {
                        // folding will eat first newline
                        str += json.slice(start, i) + '\n\n';
                        while (json[i + 2] === '\\' &&
                            json[i + 3] === 'n' &&
                            json[i + 4] !== '"') {
                            str += '\n';
                            i += 2;
                        }
                        str += indent;
                        // space after newline needs to be escaped to not be folded
                        if (json[i + 2] === ' ')
                            str += '\\';
                        i += 1;
                        start = i + 1;
                    }
                    break;
                default:
                    i += 1;
            }
    }
    str = start ? str + json.slice(start) : json;
    return implicitKey
        ? str
        : foldFlowLines(str, indent, FOLD_QUOTED, getFoldOptions(ctx, false));
}
function singleQuotedString(value, ctx) {
    if (ctx.options.singleQuote === false ||
        (ctx.implicitKey && value.includes('\n')) ||
        /[ \t]\n|\n[ \t]/.test(value) // single quoted string can't have leading or trailing whitespace around newline
    )
        return doubleQuotedString(value, ctx);
    const indent = ctx.indent || (containsDocumentMarker(value) ? '  ' : '');
    const res = "'" + value.replace(/'/g, "''").replace(/\n+/g, `$&\n${indent}`) + "'";
    return ctx.implicitKey
        ? res
        : foldFlowLines(res, indent, FOLD_FLOW, getFoldOptions(ctx, false));
}
function quotedString(value, ctx) {
    const { singleQuote } = ctx.options;
    let qs;
    if (singleQuote === false)
        qs = doubleQuotedString;
    else {
        const hasDouble = value.includes('"');
        const hasSingle = value.includes("'");
        if (hasDouble && !hasSingle)
            qs = singleQuotedString;
        else if (hasSingle && !hasDouble)
            qs = doubleQuotedString;
        else
            qs = singleQuote ? singleQuotedString : doubleQuotedString;
    }
    return qs(value, ctx);
}
// The negative lookbehind avoids a polynomial search,
// but isn't supported yet on Safari: https://caniuse.com/js-regexp-lookbehind
let blockEndNewlines;
try {
    blockEndNewlines = new RegExp('(^|(?<!\n))\n+(?!\n|$)', 'g');
}
catch {
    blockEndNewlines = /\n+(?!\n|$)/g;
}
function blockString({ comment, type, value }, ctx, onComment, onChompKeep) {
    const { blockQuote, commentString, lineWidth } = ctx.options;
    // 1. Block can't end in whitespace unless the last line is non-empty.
    // 2. Strings consisting of only whitespace are best rendered explicitly.
    if (!blockQuote || /\n[\t ]+$/.test(value) || /^\s*$/.test(value)) {
        return quotedString(value, ctx);
    }
    const indent = ctx.indent ||
        (ctx.forceBlockIndent || containsDocumentMarker(value) ? '  ' : '');
    const literal = blockQuote === 'literal'
        ? true
        : blockQuote === 'folded' || type === Scalar.BLOCK_FOLDED
            ? false
            : type === Scalar.BLOCK_LITERAL
                ? true
                : !lineLengthOverLimit(value, lineWidth, indent.length);
    if (!value)
        return literal ? '|\n' : '>\n';
    // determine chomping from whitespace at value end
    let chomp;
    let endStart;
    for (endStart = value.length; endStart > 0; --endStart) {
        const ch = value[endStart - 1];
        if (ch !== '\n' && ch !== '\t' && ch !== ' ')
            break;
    }
    let end = value.substring(endStart);
    const endNlPos = end.indexOf('\n');
    if (endNlPos === -1) {
        chomp = '-'; // strip
    }
    else if (value === end || endNlPos !== end.length - 1) {
        chomp = '+'; // keep
        if (onChompKeep)
            onChompKeep();
    }
    else {
        chomp = ''; // clip
    }
    if (end) {
        value = value.slice(0, -end.length);
        if (end[end.length - 1] === '\n')
            end = end.slice(0, -1);
        end = end.replace(blockEndNewlines, `$&${indent}`);
    }
    // determine indent indicator from whitespace at value start
    let startWithSpace = false;
    let startEnd;
    let startNlPos = -1;
    for (startEnd = 0; startEnd < value.length; ++startEnd) {
        const ch = value[startEnd];
        if (ch === ' ')
            startWithSpace = true;
        else if (ch === '\n')
            startNlPos = startEnd;
        else
            break;
    }
    let start = value.substring(0, startNlPos < startEnd ? startNlPos + 1 : startEnd);
    if (start) {
        value = value.substring(start.length);
        start = start.replace(/\n+/g, `$&${indent}`);
    }
    const indentSize = indent ? '2' : '1'; // root is at -1
    let header = (literal ? '|' : '>') + (startWithSpace ? indentSize : '') + chomp;
    if (comment) {
        header += ' ' + commentString(comment.replace(/ ?[\r\n]+/g, ' '));
        if (onComment)
            onComment();
    }
    if (literal) {
        value = value.replace(/\n+/g, `$&${indent}`);
        return `${header}\n${indent}${start}${value}${end}`;
    }
    value = value
        .replace(/\n+/g, '\n$&')
        .replace(/(?:^|\n)([\t ].*)(?:([\n\t ]*)\n(?![\n\t ]))?/g, '$1$2') // more-indented lines aren't folded
        //                ^ more-ind. ^ empty     ^ capture next empty lines only at end of indent
        .replace(/\n+/g, `$&${indent}`);
    const body = foldFlowLines(`${start}${value}${end}`, indent, FOLD_BLOCK, getFoldOptions(ctx, true));
    return `${header}\n${indent}${body}`;
}
function plainString(item, ctx, onComment, onChompKeep) {
    const { type, value } = item;
    const { actualString, implicitKey, indent, indentStep, inFlow } = ctx;
    if ((implicitKey && value.includes('\n')) ||
        (inFlow && /[[\]{},]/.test(value))) {
        return quotedString(value, ctx);
    }
    if (!value ||
        /^[\n\t ,[\]{}#&*!|>'"%@`]|^[?-]$|^[?-][ \t]|[\n:][ \t]|[ \t]\n|[\n\t ]#|[\n\t :]$/.test(value)) {
        // not allowed:
        // - empty string, '-' or '?'
        // - start with an indicator character (except [?:-]) or /[?-] /
        // - '\n ', ': ' or ' \n' anywhere
        // - '#' not preceded by a non-space char
        // - end with ' ' or ':'
        return implicitKey || inFlow || !value.includes('\n')
            ? quotedString(value, ctx)
            : blockString(item, ctx, onComment, onChompKeep);
    }
    if (!implicitKey &&
        !inFlow &&
        type !== Scalar.PLAIN &&
        value.includes('\n')) {
        // Where allowed & type not set explicitly, prefer block style for multiline strings
        return blockString(item, ctx, onComment, onChompKeep);
    }
    if (containsDocumentMarker(value)) {
        if (indent === '') {
            ctx.forceBlockIndent = true;
            return blockString(item, ctx, onComment, onChompKeep);
        }
        else if (implicitKey && indent === indentStep) {
            return quotedString(value, ctx);
        }
    }
    const str = value.replace(/\n+/g, `$&\n${indent}`);
    // Verify that output will be parsed as a string, as e.g. plain numbers and
    // booleans get parsed with those types in v1.2 (e.g. '42', 'true' & '0.9e-3'),
    // and others in v1.1.
    if (actualString) {
        const test = (tag) => tag.default && tag.tag !== 'tag:yaml.org,2002:str' && tag.test?.test(str);
        const { compat, tags } = ctx.doc.schema;
        if (tags.some(test) || compat?.some(test))
            return quotedString(value, ctx);
    }
    return implicitKey
        ? str
        : foldFlowLines(str, indent, FOLD_FLOW, getFoldOptions(ctx, false));
}
function stringifyString(item, ctx, onComment, onChompKeep) {
    const { implicitKey, inFlow } = ctx;
    const ss = typeof item.value === 'string'
        ? item
        : Object.assign({}, item, { value: String(item.value) });
    let { type } = item;
    if (type !== Scalar.QUOTE_DOUBLE) {
        // force double quotes on control characters & unpaired surrogates
        if (/[\x00-\x08\x0b-\x1f\x7f-\x9f\u{D800}-\u{DFFF}]/u.test(ss.value))
            type = Scalar.QUOTE_DOUBLE;
    }
    const _stringify = (_type) => {
        switch (_type) {
            case Scalar.BLOCK_FOLDED:
            case Scalar.BLOCK_LITERAL:
                return implicitKey || inFlow
                    ? quotedString(ss.value, ctx) // blocks are not valid inside flow containers
                    : blockString(ss, ctx, onComment, onChompKeep);
            case Scalar.QUOTE_DOUBLE:
                return doubleQuotedString(ss.value, ctx);
            case Scalar.QUOTE_SINGLE:
                return singleQuotedString(ss.value, ctx);
            case Scalar.PLAIN:
                return plainString(ss, ctx, onComment, onChompKeep);
            default:
                return null;
        }
    };
    let res = _stringify(type);
    if (res === null) {
        const { defaultKeyType, defaultStringType } = ctx.options;
        const t = (implicitKey && defaultKeyType) || defaultStringType;
        res = _stringify(t);
        if (res === null)
            throw new Error(`Unsupported default string type ${t}`);
    }
    return res;
}

function createStringifyContext(doc, options) {
    const opt = Object.assign({
        blockQuote: true,
        commentString: stringifyComment,
        defaultKeyType: null,
        defaultStringType: 'PLAIN',
        directives: null,
        doubleQuotedAsJSON: false,
        doubleQuotedMinMultiLineLength: 40,
        falseStr: 'false',
        flowCollectionPadding: true,
        indentSeq: true,
        lineWidth: 80,
        minContentWidth: 20,
        nullStr: 'null',
        simpleKeys: false,
        singleQuote: null,
        trueStr: 'true',
        verifyAliasOrder: true
    }, doc.schema.toStringOptions, options);
    let inFlow;
    switch (opt.collectionStyle) {
        case 'block':
            inFlow = false;
            break;
        case 'flow':
            inFlow = true;
            break;
        default:
            inFlow = null;
    }
    return {
        anchors: new Set(),
        doc,
        flowCollectionPadding: opt.flowCollectionPadding ? ' ' : '',
        indent: '',
        indentStep: typeof opt.indent === 'number' ? ' '.repeat(opt.indent) : '  ',
        inFlow,
        options: opt
    };
}
function getTagObject(tags, item) {
    if (item.tag) {
        const match = tags.filter(t => t.tag === item.tag);
        if (match.length > 0)
            return match.find(t => t.format === item.format) ?? match[0];
    }
    let tagObj = undefined;
    let obj;
    if (isScalar(item)) {
        obj = item.value;
        const match = tags.filter(t => t.identify?.(obj));
        tagObj =
            match.find(t => t.format === item.format) ?? match.find(t => !t.format);
    }
    else {
        obj = item;
        tagObj = tags.find(t => t.nodeClass && obj instanceof t.nodeClass);
    }
    if (!tagObj) {
        const name = obj?.constructor?.name ?? typeof obj;
        throw new Error(`Tag not resolved for ${name} value`);
    }
    return tagObj;
}
// needs to be called before value stringifier to allow for circular anchor refs
function stringifyProps(node, tagObj, { anchors, doc }) {
    if (!doc.directives)
        return '';
    const props = [];
    const anchor = (isScalar(node) || isCollection(node)) && node.anchor;
    if (anchor && anchorIsValid(anchor)) {
        anchors.add(anchor);
        props.push(`&${anchor}`);
    }
    const tag = node.tag ? node.tag : tagObj.default ? null : tagObj.tag;
    if (tag)
        props.push(doc.directives.tagString(tag));
    return props.join(' ');
}
function stringify$1(item, ctx, onComment, onChompKeep) {
    if (isPair(item))
        return item.toString(ctx, onComment, onChompKeep);
    if (isAlias(item)) {
        if (ctx.doc.directives)
            return item.toString(ctx);
        if (ctx.resolvedAliases?.has(item)) {
            throw new TypeError(`Cannot stringify circular structure without alias nodes`);
        }
        else {
            if (ctx.resolvedAliases)
                ctx.resolvedAliases.add(item);
            else
                ctx.resolvedAliases = new Set([item]);
            item = item.resolve(ctx.doc);
        }
    }
    let tagObj = undefined;
    const node = isNode(item)
        ? item
        : ctx.doc.createNode(item, { onTagObj: o => (tagObj = o) });
    if (!tagObj)
        tagObj = getTagObject(ctx.doc.schema.tags, node);
    const props = stringifyProps(node, tagObj, ctx);
    if (props.length > 0)
        ctx.indentAtStart = (ctx.indentAtStart ?? 0) + props.length + 1;
    const str = typeof tagObj.stringify === 'function'
        ? tagObj.stringify(node, ctx, onComment, onChompKeep)
        : isScalar(node)
            ? stringifyString(node, ctx, onComment, onChompKeep)
            : node.toString(ctx, onComment, onChompKeep);
    if (!props)
        return str;
    return isScalar(node) || str[0] === '{' || str[0] === '['
        ? `${props} ${str}`
        : `${props}\n${ctx.indent}${str}`;
}

function stringifyPair({ key, value }, ctx, onComment, onChompKeep) {
    const { allNullValues, doc, indent, indentStep, options: { commentString, indentSeq, simpleKeys } } = ctx;
    let keyComment = (isNode(key) && key.comment) || null;
    if (simpleKeys) {
        if (keyComment) {
            throw new Error('With simple keys, key nodes cannot have comments');
        }
        if (isCollection(key) || (!isNode(key) && typeof key === 'object')) {
            const msg = 'With simple keys, collection cannot be used as a key value';
            throw new Error(msg);
        }
    }
    let explicitKey = !simpleKeys &&
        (!key ||
            (keyComment && value == null && !ctx.inFlow) ||
            isCollection(key) ||
            (isScalar(key)
                ? key.type === Scalar.BLOCK_FOLDED || key.type === Scalar.BLOCK_LITERAL
                : typeof key === 'object'));
    ctx = Object.assign({}, ctx, {
        allNullValues: false,
        implicitKey: !explicitKey && (simpleKeys || !allNullValues),
        indent: indent + indentStep
    });
    let keyCommentDone = false;
    let chompKeep = false;
    let str = stringify$1(key, ctx, () => (keyCommentDone = true), () => (chompKeep = true));
    if (!explicitKey && !ctx.inFlow && str.length > 1024) {
        if (simpleKeys)
            throw new Error('With simple keys, single line scalar must not span more than 1024 characters');
        explicitKey = true;
    }
    if (ctx.inFlow) {
        if (allNullValues || value == null) {
            if (keyCommentDone && onComment)
                onComment();
            return str === '' ? '?' : explicitKey ? `? ${str}` : str;
        }
    }
    else if ((allNullValues && !simpleKeys) || (value == null && explicitKey)) {
        str = `? ${str}`;
        if (keyComment && !keyCommentDone) {
            str += lineComment(str, ctx.indent, commentString(keyComment));
        }
        else if (chompKeep && onChompKeep)
            onChompKeep();
        return str;
    }
    if (keyCommentDone)
        keyComment = null;
    if (explicitKey) {
        if (keyComment)
            str += lineComment(str, ctx.indent, commentString(keyComment));
        str = `? ${str}\n${indent}:`;
    }
    else {
        str = `${str}:`;
        if (keyComment)
            str += lineComment(str, ctx.indent, commentString(keyComment));
    }
    let vsb, vcb, valueComment;
    if (isNode(value)) {
        vsb = !!value.spaceBefore;
        vcb = value.commentBefore;
        valueComment = value.comment;
    }
    else {
        vsb = false;
        vcb = null;
        valueComment = null;
        if (value && typeof value === 'object')
            value = doc.createNode(value);
    }
    ctx.implicitKey = false;
    if (!explicitKey && !keyComment && isScalar(value))
        ctx.indentAtStart = str.length + 1;
    chompKeep = false;
    if (!indentSeq &&
        indentStep.length >= 2 &&
        !ctx.inFlow &&
        !explicitKey &&
        isSeq(value) &&
        !value.flow &&
        !value.tag &&
        !value.anchor) {
        // If indentSeq === false, consider '- ' as part of indentation where possible
        ctx.indent = ctx.indent.substring(2);
    }
    let valueCommentDone = false;
    const valueStr = stringify$1(value, ctx, () => (valueCommentDone = true), () => (chompKeep = true));
    let ws = ' ';
    if (keyComment || vsb || vcb) {
        ws = vsb ? '\n' : '';
        if (vcb) {
            const cs = commentString(vcb);
            ws += `\n${indentComment(cs, ctx.indent)}`;
        }
        if (valueStr === '' && !ctx.inFlow) {
            if (ws === '\n')
                ws = '\n\n';
        }
        else {
            ws += `\n${ctx.indent}`;
        }
    }
    else if (!explicitKey && isCollection(value)) {
        const vs0 = valueStr[0];
        const nl0 = valueStr.indexOf('\n');
        const hasNewline = nl0 !== -1;
        const flow = ctx.inFlow ?? value.flow ?? value.items.length === 0;
        if (hasNewline || !flow) {
            let hasPropsLine = false;
            if (hasNewline && (vs0 === '&' || vs0 === '!')) {
                let sp0 = valueStr.indexOf(' ');
                if (vs0 === '&' &&
                    sp0 !== -1 &&
                    sp0 < nl0 &&
                    valueStr[sp0 + 1] === '!') {
                    sp0 = valueStr.indexOf(' ', sp0 + 1);
                }
                if (sp0 === -1 || nl0 < sp0)
                    hasPropsLine = true;
            }
            if (!hasPropsLine)
                ws = `\n${ctx.indent}`;
        }
    }
    else if (valueStr === '' || valueStr[0] === '\n') {
        ws = '';
    }
    str += ws + valueStr;
    if (ctx.inFlow) {
        if (valueCommentDone && onComment)
            onComment();
    }
    else if (valueComment && !valueCommentDone) {
        str += lineComment(str, ctx.indent, commentString(valueComment));
    }
    else if (chompKeep && onChompKeep) {
        onChompKeep();
    }
    return str;
}

function warn(logLevel, warning) {
    if (logLevel === 'debug' || logLevel === 'warn') {
        if (typeof process !== 'undefined' && process.emitWarning)
            process.emitWarning(warning);
        else
            console.warn(warning);
    }
}

const MERGE_KEY = '<<';
function addPairToJSMap(ctx, map, { key, value }) {
    if (ctx?.doc.schema.merge && isMergeKey(key)) {
        value = isAlias(value) ? value.resolve(ctx.doc) : value;
        if (isSeq(value))
            for (const it of value.items)
                mergeToJSMap(ctx, map, it);
        else if (Array.isArray(value))
            for (const it of value)
                mergeToJSMap(ctx, map, it);
        else
            mergeToJSMap(ctx, map, value);
    }
    else {
        const jsKey = toJS(key, '', ctx);
        if (map instanceof Map) {
            map.set(jsKey, toJS(value, jsKey, ctx));
        }
        else if (map instanceof Set) {
            map.add(jsKey);
        }
        else {
            const stringKey = stringifyKey(key, jsKey, ctx);
            const jsValue = toJS(value, stringKey, ctx);
            if (stringKey in map)
                Object.defineProperty(map, stringKey, {
                    value: jsValue,
                    writable: true,
                    enumerable: true,
                    configurable: true
                });
            else
                map[stringKey] = jsValue;
        }
    }
    return map;
}
const isMergeKey = (key) => key === MERGE_KEY ||
    (isScalar(key) &&
        key.value === MERGE_KEY &&
        (!key.type || key.type === Scalar.PLAIN));
// If the value associated with a merge key is a single mapping node, each of
// its key/value pairs is inserted into the current mapping, unless the key
// already exists in it. If the value associated with the merge key is a
// sequence, then this sequence is expected to contain mapping nodes and each
// of these nodes is merged in turn according to its order in the sequence.
// Keys in mapping nodes earlier in the sequence override keys specified in
// later mapping nodes. -- http://yaml.org/type/merge.html
function mergeToJSMap(ctx, map, value) {
    const source = ctx && isAlias(value) ? value.resolve(ctx.doc) : value;
    if (!isMap(source))
        throw new Error('Merge sources must be maps or map aliases');
    const srcMap = source.toJSON(null, ctx, Map);
    for (const [key, value] of srcMap) {
        if (map instanceof Map) {
            if (!map.has(key))
                map.set(key, value);
        }
        else if (map instanceof Set) {
            map.add(key);
        }
        else if (!Object.prototype.hasOwnProperty.call(map, key)) {
            Object.defineProperty(map, key, {
                value,
                writable: true,
                enumerable: true,
                configurable: true
            });
        }
    }
    return map;
}
function stringifyKey(key, jsKey, ctx) {
    if (jsKey === null)
        return '';
    if (typeof jsKey !== 'object')
        return String(jsKey);
    if (isNode(key) && ctx?.doc) {
        const strCtx = createStringifyContext(ctx.doc, {});
        strCtx.anchors = new Set();
        for (const node of ctx.anchors.keys())
            strCtx.anchors.add(node.anchor);
        strCtx.inFlow = true;
        strCtx.inStringifyKey = true;
        const strKey = key.toString(strCtx);
        if (!ctx.mapKeyWarned) {
            let jsonStr = JSON.stringify(strKey);
            if (jsonStr.length > 40)
                jsonStr = jsonStr.substring(0, 36) + '..."';
            warn(ctx.doc.options.logLevel, `Keys with collection values will be stringified due to JS Object restrictions: ${jsonStr}. Set mapAsMap: true to use object keys.`);
            ctx.mapKeyWarned = true;
        }
        return strKey;
    }
    return JSON.stringify(jsKey);
}

function createPair(key, value, ctx) {
    const k = createNode(key, undefined, ctx);
    const v = createNode(value, undefined, ctx);
    return new Pair(k, v);
}
class Pair {
    constructor(key, value = null) {
        Object.defineProperty(this, NODE_TYPE, { value: PAIR });
        this.key = key;
        this.value = value;
    }
    clone(schema) {
        let { key, value } = this;
        if (isNode(key))
            key = key.clone(schema);
        if (isNode(value))
            value = value.clone(schema);
        return new Pair(key, value);
    }
    toJSON(_, ctx) {
        const pair = ctx?.mapAsMap ? new Map() : {};
        return addPairToJSMap(ctx, pair, this);
    }
    toString(ctx, onComment, onChompKeep) {
        return ctx?.doc
            ? stringifyPair(this, ctx, onComment, onChompKeep)
            : JSON.stringify(this);
    }
}

function stringifyCollection(collection, ctx, options) {
    const flow = ctx.inFlow ?? collection.flow;
    const stringify = flow ? stringifyFlowCollection : stringifyBlockCollection;
    return stringify(collection, ctx, options);
}
function stringifyBlockCollection({ comment, items }, ctx, { blockItemPrefix, flowChars, itemIndent, onChompKeep, onComment }) {
    const { indent, options: { commentString } } = ctx;
    const itemCtx = Object.assign({}, ctx, { indent: itemIndent, type: null });
    let chompKeep = false; // flag for the preceding node's status
    const lines = [];
    for (let i = 0; i < items.length; ++i) {
        const item = items[i];
        let comment = null;
        if (isNode(item)) {
            if (!chompKeep && item.spaceBefore)
                lines.push('');
            addCommentBefore(ctx, lines, item.commentBefore, chompKeep);
            if (item.comment)
                comment = item.comment;
        }
        else if (isPair(item)) {
            const ik = isNode(item.key) ? item.key : null;
            if (ik) {
                if (!chompKeep && ik.spaceBefore)
                    lines.push('');
                addCommentBefore(ctx, lines, ik.commentBefore, chompKeep);
            }
        }
        chompKeep = false;
        let str = stringify$1(item, itemCtx, () => (comment = null), () => (chompKeep = true));
        if (comment)
            str += lineComment(str, itemIndent, commentString(comment));
        if (chompKeep && comment)
            chompKeep = false;
        lines.push(blockItemPrefix + str);
    }
    let str;
    if (lines.length === 0) {
        str = flowChars.start + flowChars.end;
    }
    else {
        str = lines[0];
        for (let i = 1; i < lines.length; ++i) {
            const line = lines[i];
            str += line ? `\n${indent}${line}` : '\n';
        }
    }
    if (comment) {
        str += '\n' + indentComment(commentString(comment), indent);
        if (onComment)
            onComment();
    }
    else if (chompKeep && onChompKeep)
        onChompKeep();
    return str;
}
function stringifyFlowCollection({ items }, ctx, { flowChars, itemIndent }) {
    const { indent, indentStep, flowCollectionPadding: fcPadding, options: { commentString } } = ctx;
    itemIndent += indentStep;
    const itemCtx = Object.assign({}, ctx, {
        indent: itemIndent,
        inFlow: true,
        type: null
    });
    let reqNewline = false;
    let linesAtValue = 0;
    const lines = [];
    for (let i = 0; i < items.length; ++i) {
        const item = items[i];
        let comment = null;
        if (isNode(item)) {
            if (item.spaceBefore)
                lines.push('');
            addCommentBefore(ctx, lines, item.commentBefore, false);
            if (item.comment)
                comment = item.comment;
        }
        else if (isPair(item)) {
            const ik = isNode(item.key) ? item.key : null;
            if (ik) {
                if (ik.spaceBefore)
                    lines.push('');
                addCommentBefore(ctx, lines, ik.commentBefore, false);
                if (ik.comment)
                    reqNewline = true;
            }
            const iv = isNode(item.value) ? item.value : null;
            if (iv) {
                if (iv.comment)
                    comment = iv.comment;
                if (iv.commentBefore)
                    reqNewline = true;
            }
            else if (item.value == null && ik?.comment) {
                comment = ik.comment;
            }
        }
        if (comment)
            reqNewline = true;
        let str = stringify$1(item, itemCtx, () => (comment = null));
        if (i < items.length - 1)
            str += ',';
        if (comment)
            str += lineComment(str, itemIndent, commentString(comment));
        if (!reqNewline && (lines.length > linesAtValue || str.includes('\n')))
            reqNewline = true;
        lines.push(str);
        linesAtValue = lines.length;
    }
    const { start, end } = flowChars;
    if (lines.length === 0) {
        return start + end;
    }
    else {
        if (!reqNewline) {
            const len = lines.reduce((sum, line) => sum + line.length + 2, 2);
            reqNewline = ctx.options.lineWidth > 0 && len > ctx.options.lineWidth;
        }
        if (reqNewline) {
            let str = start;
            for (const line of lines)
                str += line ? `\n${indentStep}${indent}${line}` : '\n';
            return `${str}\n${indent}${end}`;
        }
        else {
            return `${start}${fcPadding}${lines.join(' ')}${fcPadding}${end}`;
        }
    }
}
function addCommentBefore({ indent, options: { commentString } }, lines, comment, chompKeep) {
    if (comment && chompKeep)
        comment = comment.replace(/^\n+/, '');
    if (comment) {
        const ic = indentComment(commentString(comment), indent);
        lines.push(ic.trimStart()); // Avoid double indent on first line
    }
}

function findPair(items, key) {
    const k = isScalar(key) ? key.value : key;
    for (const it of items) {
        if (isPair(it)) {
            if (it.key === key || it.key === k)
                return it;
            if (isScalar(it.key) && it.key.value === k)
                return it;
        }
    }
    return undefined;
}
class YAMLMap extends Collection {
    static get tagName() {
        return 'tag:yaml.org,2002:map';
    }
    constructor(schema) {
        super(MAP, schema);
        this.items = [];
    }
    /**
     * A generic collection parsing method that can be extended
     * to other node classes that inherit from YAMLMap
     */
    static from(schema, obj, ctx) {
        const { keepUndefined, replacer } = ctx;
        const map = new this(schema);
        const add = (key, value) => {
            if (typeof replacer === 'function')
                value = replacer.call(obj, key, value);
            else if (Array.isArray(replacer) && !replacer.includes(key))
                return;
            if (value !== undefined || keepUndefined)
                map.items.push(createPair(key, value, ctx));
        };
        if (obj instanceof Map) {
            for (const [key, value] of obj)
                add(key, value);
        }
        else if (obj && typeof obj === 'object') {
            for (const key of Object.keys(obj))
                add(key, obj[key]);
        }
        if (typeof schema.sortMapEntries === 'function') {
            map.items.sort(schema.sortMapEntries);
        }
        return map;
    }
    /**
     * Adds a value to the collection.
     *
     * @param overwrite - If not set `true`, using a key that is already in the
     *   collection will throw. Otherwise, overwrites the previous value.
     */
    add(pair, overwrite) {
        let _pair;
        if (isPair(pair))
            _pair = pair;
        else if (!pair || typeof pair !== 'object' || !('key' in pair)) {
            // In TypeScript, this never happens.
            _pair = new Pair(pair, pair?.value);
        }
        else
            _pair = new Pair(pair.key, pair.value);
        const prev = findPair(this.items, _pair.key);
        const sortEntries = this.schema?.sortMapEntries;
        if (prev) {
            if (!overwrite)
                throw new Error(`Key ${_pair.key} already set`);
            // For scalars, keep the old node & its comments and anchors
            if (isScalar(prev.value) && isScalarValue(_pair.value))
                prev.value.value = _pair.value;
            else
                prev.value = _pair.value;
        }
        else if (sortEntries) {
            const i = this.items.findIndex(item => sortEntries(_pair, item) < 0);
            if (i === -1)
                this.items.push(_pair);
            else
                this.items.splice(i, 0, _pair);
        }
        else {
            this.items.push(_pair);
        }
    }
    delete(key) {
        const it = findPair(this.items, key);
        if (!it)
            return false;
        const del = this.items.splice(this.items.indexOf(it), 1);
        return del.length > 0;
    }
    get(key, keepScalar) {
        const it = findPair(this.items, key);
        const node = it?.value;
        return (!keepScalar && isScalar(node) ? node.value : node) ?? undefined;
    }
    has(key) {
        return !!findPair(this.items, key);
    }
    set(key, value) {
        this.add(new Pair(key, value), true);
    }
    /**
     * @param ctx - Conversion context, originally set in Document#toJS()
     * @param {Class} Type - If set, forces the returned collection type
     * @returns Instance of Type, Map, or Object
     */
    toJSON(_, ctx, Type) {
        const map = Type ? new Type() : ctx?.mapAsMap ? new Map() : {};
        if (ctx?.onCreate)
            ctx.onCreate(map);
        for (const item of this.items)
            addPairToJSMap(ctx, map, item);
        return map;
    }
    toString(ctx, onComment, onChompKeep) {
        if (!ctx)
            return JSON.stringify(this);
        for (const item of this.items) {
            if (!isPair(item))
                throw new Error(`Map items must all be pairs; found ${JSON.stringify(item)} instead`);
        }
        if (!ctx.allNullValues && this.hasAllNullValues(false))
            ctx = Object.assign({}, ctx, { allNullValues: true });
        return stringifyCollection(this, ctx, {
            blockItemPrefix: '',
            flowChars: { start: '{', end: '}' },
            itemIndent: ctx.indent || '',
            onChompKeep,
            onComment
        });
    }
}

const map = {
    collection: 'map',
    default: true,
    nodeClass: YAMLMap,
    tag: 'tag:yaml.org,2002:map',
    resolve(map, onError) {
        if (!isMap(map))
            onError('Expected a mapping for this tag');
        return map;
    },
    createNode: (schema, obj, ctx) => YAMLMap.from(schema, obj, ctx)
};

class YAMLSeq extends Collection {
    static get tagName() {
        return 'tag:yaml.org,2002:seq';
    }
    constructor(schema) {
        super(SEQ, schema);
        this.items = [];
    }
    add(value) {
        this.items.push(value);
    }
    /**
     * Removes a value from the collection.
     *
     * `key` must contain a representation of an integer for this to succeed.
     * It may be wrapped in a `Scalar`.
     *
     * @returns `true` if the item was found and removed.
     */
    delete(key) {
        const idx = asItemIndex(key);
        if (typeof idx !== 'number')
            return false;
        const del = this.items.splice(idx, 1);
        return del.length > 0;
    }
    get(key, keepScalar) {
        const idx = asItemIndex(key);
        if (typeof idx !== 'number')
            return undefined;
        const it = this.items[idx];
        return !keepScalar && isScalar(it) ? it.value : it;
    }
    /**
     * Checks if the collection includes a value with the key `key`.
     *
     * `key` must contain a representation of an integer for this to succeed.
     * It may be wrapped in a `Scalar`.
     */
    has(key) {
        const idx = asItemIndex(key);
        return typeof idx === 'number' && idx < this.items.length;
    }
    /**
     * Sets a value in this collection. For `!!set`, `value` needs to be a
     * boolean to add/remove the item from the set.
     *
     * If `key` does not contain a representation of an integer, this will throw.
     * It may be wrapped in a `Scalar`.
     */
    set(key, value) {
        const idx = asItemIndex(key);
        if (typeof idx !== 'number')
            throw new Error(`Expected a valid index, not ${key}.`);
        const prev = this.items[idx];
        if (isScalar(prev) && isScalarValue(value))
            prev.value = value;
        else
            this.items[idx] = value;
    }
    toJSON(_, ctx) {
        const seq = [];
        if (ctx?.onCreate)
            ctx.onCreate(seq);
        let i = 0;
        for (const item of this.items)
            seq.push(toJS(item, String(i++), ctx));
        return seq;
    }
    toString(ctx, onComment, onChompKeep) {
        if (!ctx)
            return JSON.stringify(this);
        return stringifyCollection(this, ctx, {
            blockItemPrefix: '- ',
            flowChars: { start: '[', end: ']' },
            itemIndent: (ctx.indent || '') + '  ',
            onChompKeep,
            onComment
        });
    }
    static from(schema, obj, ctx) {
        const { replacer } = ctx;
        const seq = new this(schema);
        if (obj && Symbol.iterator in Object(obj)) {
            let i = 0;
            for (let it of obj) {
                if (typeof replacer === 'function') {
                    const key = obj instanceof Set ? it : String(i++);
                    it = replacer.call(obj, key, it);
                }
                seq.items.push(createNode(it, undefined, ctx));
            }
        }
        return seq;
    }
}
function asItemIndex(key) {
    let idx = isScalar(key) ? key.value : key;
    if (idx && typeof idx === 'string')
        idx = Number(idx);
    return typeof idx === 'number' && Number.isInteger(idx) && idx >= 0
        ? idx
        : null;
}

const seq = {
    collection: 'seq',
    default: true,
    nodeClass: YAMLSeq,
    tag: 'tag:yaml.org,2002:seq',
    resolve(seq, onError) {
        if (!isSeq(seq))
            onError('Expected a sequence for this tag');
        return seq;
    },
    createNode: (schema, obj, ctx) => YAMLSeq.from(schema, obj, ctx)
};

const string = {
    identify: value => typeof value === 'string',
    default: true,
    tag: 'tag:yaml.org,2002:str',
    resolve: str => str,
    stringify(item, ctx, onComment, onChompKeep) {
        ctx = Object.assign({ actualString: true }, ctx);
        return stringifyString(item, ctx, onComment, onChompKeep);
    }
};

const nullTag = {
    identify: value => value == null,
    createNode: () => new Scalar(null),
    default: true,
    tag: 'tag:yaml.org,2002:null',
    test: /^(?:~|[Nn]ull|NULL)?$/,
    resolve: () => new Scalar(null),
    stringify: ({ source }, ctx) => typeof source === 'string' && nullTag.test.test(source)
        ? source
        : ctx.options.nullStr
};

const boolTag = {
    identify: value => typeof value === 'boolean',
    default: true,
    tag: 'tag:yaml.org,2002:bool',
    test: /^(?:[Tt]rue|TRUE|[Ff]alse|FALSE)$/,
    resolve: str => new Scalar(str[0] === 't' || str[0] === 'T'),
    stringify({ source, value }, ctx) {
        if (source && boolTag.test.test(source)) {
            const sv = source[0] === 't' || source[0] === 'T';
            if (value === sv)
                return source;
        }
        return value ? ctx.options.trueStr : ctx.options.falseStr;
    }
};

function stringifyNumber({ format, minFractionDigits, tag, value }) {
    if (typeof value === 'bigint')
        return String(value);
    const num = typeof value === 'number' ? value : Number(value);
    if (!isFinite(num))
        return isNaN(num) ? '.nan' : num < 0 ? '-.inf' : '.inf';
    let n = JSON.stringify(value);
    if (!format &&
        minFractionDigits &&
        (!tag || tag === 'tag:yaml.org,2002:float') &&
        /^\d/.test(n)) {
        let i = n.indexOf('.');
        if (i < 0) {
            i = n.length;
            n += '.';
        }
        let d = minFractionDigits - (n.length - i - 1);
        while (d-- > 0)
            n += '0';
    }
    return n;
}

const floatNaN$1 = {
    identify: value => typeof value === 'number',
    default: true,
    tag: 'tag:yaml.org,2002:float',
    test: /^(?:[-+]?\.(?:inf|Inf|INF)|\.nan|\.NaN|\.NAN)$/,
    resolve: str => str.slice(-3).toLowerCase() === 'nan'
        ? NaN
        : str[0] === '-'
            ? Number.NEGATIVE_INFINITY
            : Number.POSITIVE_INFINITY,
    stringify: stringifyNumber
};
const floatExp$1 = {
    identify: value => typeof value === 'number',
    default: true,
    tag: 'tag:yaml.org,2002:float',
    format: 'EXP',
    test: /^[-+]?(?:\.[0-9]+|[0-9]+(?:\.[0-9]*)?)[eE][-+]?[0-9]+$/,
    resolve: str => parseFloat(str),
    stringify(node) {
        const num = Number(node.value);
        return isFinite(num) ? num.toExponential() : stringifyNumber(node);
    }
};
const float$1 = {
    identify: value => typeof value === 'number',
    default: true,
    tag: 'tag:yaml.org,2002:float',
    test: /^[-+]?(?:\.[0-9]+|[0-9]+\.[0-9]*)$/,
    resolve(str) {
        const node = new Scalar(parseFloat(str));
        const dot = str.indexOf('.');
        if (dot !== -1 && str[str.length - 1] === '0')
            node.minFractionDigits = str.length - dot - 1;
        return node;
    },
    stringify: stringifyNumber
};

const intIdentify$2 = (value) => typeof value === 'bigint' || Number.isInteger(value);
const intResolve$1 = (str, offset, radix, { intAsBigInt }) => (intAsBigInt ? BigInt(str) : parseInt(str.substring(offset), radix));
function intStringify$1(node, radix, prefix) {
    const { value } = node;
    if (intIdentify$2(value) && value >= 0)
        return prefix + value.toString(radix);
    return stringifyNumber(node);
}
const intOct$1 = {
    identify: value => intIdentify$2(value) && value >= 0,
    default: true,
    tag: 'tag:yaml.org,2002:int',
    format: 'OCT',
    test: /^0o[0-7]+$/,
    resolve: (str, _onError, opt) => intResolve$1(str, 2, 8, opt),
    stringify: node => intStringify$1(node, 8, '0o')
};
const int$1 = {
    identify: intIdentify$2,
    default: true,
    tag: 'tag:yaml.org,2002:int',
    test: /^[-+]?[0-9]+$/,
    resolve: (str, _onError, opt) => intResolve$1(str, 0, 10, opt),
    stringify: stringifyNumber
};
const intHex$1 = {
    identify: value => intIdentify$2(value) && value >= 0,
    default: true,
    tag: 'tag:yaml.org,2002:int',
    format: 'HEX',
    test: /^0x[0-9a-fA-F]+$/,
    resolve: (str, _onError, opt) => intResolve$1(str, 2, 16, opt),
    stringify: node => intStringify$1(node, 16, '0x')
};

const schema$2 = [
    map,
    seq,
    string,
    nullTag,
    boolTag,
    intOct$1,
    int$1,
    intHex$1,
    floatNaN$1,
    floatExp$1,
    float$1
];

function intIdentify$1(value) {
    return typeof value === 'bigint' || Number.isInteger(value);
}
const stringifyJSON = ({ value }) => JSON.stringify(value);
const jsonScalars = [
    {
        identify: value => typeof value === 'string',
        default: true,
        tag: 'tag:yaml.org,2002:str',
        resolve: str => str,
        stringify: stringifyJSON
    },
    {
        identify: value => value == null,
        createNode: () => new Scalar(null),
        default: true,
        tag: 'tag:yaml.org,2002:null',
        test: /^null$/,
        resolve: () => null,
        stringify: stringifyJSON
    },
    {
        identify: value => typeof value === 'boolean',
        default: true,
        tag: 'tag:yaml.org,2002:bool',
        test: /^true|false$/,
        resolve: str => str === 'true',
        stringify: stringifyJSON
    },
    {
        identify: intIdentify$1,
        default: true,
        tag: 'tag:yaml.org,2002:int',
        test: /^-?(?:0|[1-9][0-9]*)$/,
        resolve: (str, _onError, { intAsBigInt }) => intAsBigInt ? BigInt(str) : parseInt(str, 10),
        stringify: ({ value }) => intIdentify$1(value) ? value.toString() : JSON.stringify(value)
    },
    {
        identify: value => typeof value === 'number',
        default: true,
        tag: 'tag:yaml.org,2002:float',
        test: /^-?(?:0|[1-9][0-9]*)(?:\.[0-9]*)?(?:[eE][-+]?[0-9]+)?$/,
        resolve: str => parseFloat(str),
        stringify: stringifyJSON
    }
];
const jsonError = {
    default: true,
    tag: '',
    test: /^/,
    resolve(str, onError) {
        onError(`Unresolved plain scalar ${JSON.stringify(str)}`);
        return str;
    }
};
const schema$1 = [map, seq].concat(jsonScalars, jsonError);

const binary = {
    identify: value => value instanceof Uint8Array, // Buffer inherits from Uint8Array
    default: false,
    tag: 'tag:yaml.org,2002:binary',
    /**
     * Returns a Buffer in node and an Uint8Array in browsers
     *
     * To use the resulting buffer as an image, you'll want to do something like:
     *
     *   const blob = new Blob([buffer], { type: 'image/jpeg' })
     *   document.querySelector('#photo').src = URL.createObjectURL(blob)
     */
    resolve(src, onError) {
        if (typeof Buffer === 'function') {
            return Buffer.from(src, 'base64');
        }
        else if (typeof atob === 'function') {
            // On IE 11, atob() can't handle newlines
            const str = atob(src.replace(/[\n\r]/g, ''));
            const buffer = new Uint8Array(str.length);
            for (let i = 0; i < str.length; ++i)
                buffer[i] = str.charCodeAt(i);
            return buffer;
        }
        else {
            onError('This environment does not support reading binary tags; either Buffer or atob is required');
            return src;
        }
    },
    stringify({ comment, type, value }, ctx, onComment, onChompKeep) {
        const buf = value; // checked earlier by binary.identify()
        let str;
        if (typeof Buffer === 'function') {
            str =
                buf instanceof Buffer
                    ? buf.toString('base64')
                    : Buffer.from(buf.buffer).toString('base64');
        }
        else if (typeof btoa === 'function') {
            let s = '';
            for (let i = 0; i < buf.length; ++i)
                s += String.fromCharCode(buf[i]);
            str = btoa(s);
        }
        else {
            throw new Error('This environment does not support writing binary tags; either Buffer or btoa is required');
        }
        if (!type)
            type = Scalar.BLOCK_LITERAL;
        if (type !== Scalar.QUOTE_DOUBLE) {
            const lineWidth = Math.max(ctx.options.lineWidth - ctx.indent.length, ctx.options.minContentWidth);
            const n = Math.ceil(str.length / lineWidth);
            const lines = new Array(n);
            for (let i = 0, o = 0; i < n; ++i, o += lineWidth) {
                lines[i] = str.substr(o, lineWidth);
            }
            str = lines.join(type === Scalar.BLOCK_LITERAL ? '\n' : ' ');
        }
        return stringifyString({ comment, type, value: str }, ctx, onComment, onChompKeep);
    }
};

function resolvePairs(seq, onError) {
    if (isSeq(seq)) {
        for (let i = 0; i < seq.items.length; ++i) {
            let item = seq.items[i];
            if (isPair(item))
                continue;
            else if (isMap(item)) {
                if (item.items.length > 1)
                    onError('Each pair must have its own sequence indicator');
                const pair = item.items[0] || new Pair(new Scalar(null));
                if (item.commentBefore)
                    pair.key.commentBefore = pair.key.commentBefore
                        ? `${item.commentBefore}\n${pair.key.commentBefore}`
                        : item.commentBefore;
                if (item.comment) {
                    const cn = pair.value ?? pair.key;
                    cn.comment = cn.comment
                        ? `${item.comment}\n${cn.comment}`
                        : item.comment;
                }
                item = pair;
            }
            seq.items[i] = isPair(item) ? item : new Pair(item);
        }
    }
    else
        onError('Expected a sequence for this tag');
    return seq;
}
function createPairs(schema, iterable, ctx) {
    const { replacer } = ctx;
    const pairs = new YAMLSeq(schema);
    pairs.tag = 'tag:yaml.org,2002:pairs';
    let i = 0;
    if (iterable && Symbol.iterator in Object(iterable))
        for (let it of iterable) {
            if (typeof replacer === 'function')
                it = replacer.call(iterable, String(i++), it);
            let key, value;
            if (Array.isArray(it)) {
                if (it.length === 2) {
                    key = it[0];
                    value = it[1];
                }
                else
                    throw new TypeError(`Expected [key, value] tuple: ${it}`);
            }
            else if (it && it instanceof Object) {
                const keys = Object.keys(it);
                if (keys.length === 1) {
                    key = keys[0];
                    value = it[key];
                }
                else {
                    throw new TypeError(`Expected tuple with one key, not ${keys.length} keys`);
                }
            }
            else {
                key = it;
            }
            pairs.items.push(createPair(key, value, ctx));
        }
    return pairs;
}
const pairs = {
    collection: 'seq',
    default: false,
    tag: 'tag:yaml.org,2002:pairs',
    resolve: resolvePairs,
    createNode: createPairs
};

class YAMLOMap extends YAMLSeq {
    constructor() {
        super();
        this.add = YAMLMap.prototype.add.bind(this);
        this.delete = YAMLMap.prototype.delete.bind(this);
        this.get = YAMLMap.prototype.get.bind(this);
        this.has = YAMLMap.prototype.has.bind(this);
        this.set = YAMLMap.prototype.set.bind(this);
        this.tag = YAMLOMap.tag;
    }
    /**
     * If `ctx` is given, the return type is actually `Map<unknown, unknown>`,
     * but TypeScript won't allow widening the signature of a child method.
     */
    toJSON(_, ctx) {
        if (!ctx)
            return super.toJSON(_);
        const map = new Map();
        if (ctx?.onCreate)
            ctx.onCreate(map);
        for (const pair of this.items) {
            let key, value;
            if (isPair(pair)) {
                key = toJS(pair.key, '', ctx);
                value = toJS(pair.value, key, ctx);
            }
            else {
                key = toJS(pair, '', ctx);
            }
            if (map.has(key))
                throw new Error('Ordered maps must not include duplicate keys');
            map.set(key, value);
        }
        return map;
    }
    static from(schema, iterable, ctx) {
        const pairs = createPairs(schema, iterable, ctx);
        const omap = new this();
        omap.items = pairs.items;
        return omap;
    }
}
YAMLOMap.tag = 'tag:yaml.org,2002:omap';
const omap = {
    collection: 'seq',
    identify: value => value instanceof Map,
    nodeClass: YAMLOMap,
    default: false,
    tag: 'tag:yaml.org,2002:omap',
    resolve(seq, onError) {
        const pairs = resolvePairs(seq, onError);
        const seenKeys = [];
        for (const { key } of pairs.items) {
            if (isScalar(key)) {
                if (seenKeys.includes(key.value)) {
                    onError(`Ordered maps must not include duplicate keys: ${key.value}`);
                }
                else {
                    seenKeys.push(key.value);
                }
            }
        }
        return Object.assign(new YAMLOMap(), pairs);
    },
    createNode: (schema, iterable, ctx) => YAMLOMap.from(schema, iterable, ctx)
};

function boolStringify({ value, source }, ctx) {
    const boolObj = value ? trueTag : falseTag;
    if (source && boolObj.test.test(source))
        return source;
    return value ? ctx.options.trueStr : ctx.options.falseStr;
}
const trueTag = {
    identify: value => value === true,
    default: true,
    tag: 'tag:yaml.org,2002:bool',
    test: /^(?:Y|y|[Yy]es|YES|[Tt]rue|TRUE|[Oo]n|ON)$/,
    resolve: () => new Scalar(true),
    stringify: boolStringify
};
const falseTag = {
    identify: value => value === false,
    default: true,
    tag: 'tag:yaml.org,2002:bool',
    test: /^(?:N|n|[Nn]o|NO|[Ff]alse|FALSE|[Oo]ff|OFF)$/,
    resolve: () => new Scalar(false),
    stringify: boolStringify
};

const floatNaN = {
    identify: value => typeof value === 'number',
    default: true,
    tag: 'tag:yaml.org,2002:float',
    test: /^(?:[-+]?\.(?:inf|Inf|INF)|\.nan|\.NaN|\.NAN)$/,
    resolve: (str) => str.slice(-3).toLowerCase() === 'nan'
        ? NaN
        : str[0] === '-'
            ? Number.NEGATIVE_INFINITY
            : Number.POSITIVE_INFINITY,
    stringify: stringifyNumber
};
const floatExp = {
    identify: value => typeof value === 'number',
    default: true,
    tag: 'tag:yaml.org,2002:float',
    format: 'EXP',
    test: /^[-+]?(?:[0-9][0-9_]*)?(?:\.[0-9_]*)?[eE][-+]?[0-9]+$/,
    resolve: (str) => parseFloat(str.replace(/_/g, '')),
    stringify(node) {
        const num = Number(node.value);
        return isFinite(num) ? num.toExponential() : stringifyNumber(node);
    }
};
const float = {
    identify: value => typeof value === 'number',
    default: true,
    tag: 'tag:yaml.org,2002:float',
    test: /^[-+]?(?:[0-9][0-9_]*)?\.[0-9_]*$/,
    resolve(str) {
        const node = new Scalar(parseFloat(str.replace(/_/g, '')));
        const dot = str.indexOf('.');
        if (dot !== -1) {
            const f = str.substring(dot + 1).replace(/_/g, '');
            if (f[f.length - 1] === '0')
                node.minFractionDigits = f.length;
        }
        return node;
    },
    stringify: stringifyNumber
};

const intIdentify = (value) => typeof value === 'bigint' || Number.isInteger(value);
function intResolve(str, offset, radix, { intAsBigInt }) {
    const sign = str[0];
    if (sign === '-' || sign === '+')
        offset += 1;
    str = str.substring(offset).replace(/_/g, '');
    if (intAsBigInt) {
        switch (radix) {
            case 2:
                str = `0b${str}`;
                break;
            case 8:
                str = `0o${str}`;
                break;
            case 16:
                str = `0x${str}`;
                break;
        }
        const n = BigInt(str);
        return sign === '-' ? BigInt(-1) * n : n;
    }
    const n = parseInt(str, radix);
    return sign === '-' ? -1 * n : n;
}
function intStringify(node, radix, prefix) {
    const { value } = node;
    if (intIdentify(value)) {
        const str = value.toString(radix);
        return value < 0 ? '-' + prefix + str.substr(1) : prefix + str;
    }
    return stringifyNumber(node);
}
const intBin = {
    identify: intIdentify,
    default: true,
    tag: 'tag:yaml.org,2002:int',
    format: 'BIN',
    test: /^[-+]?0b[0-1_]+$/,
    resolve: (str, _onError, opt) => intResolve(str, 2, 2, opt),
    stringify: node => intStringify(node, 2, '0b')
};
const intOct = {
    identify: intIdentify,
    default: true,
    tag: 'tag:yaml.org,2002:int',
    format: 'OCT',
    test: /^[-+]?0[0-7_]+$/,
    resolve: (str, _onError, opt) => intResolve(str, 1, 8, opt),
    stringify: node => intStringify(node, 8, '0')
};
const int = {
    identify: intIdentify,
    default: true,
    tag: 'tag:yaml.org,2002:int',
    test: /^[-+]?[0-9][0-9_]*$/,
    resolve: (str, _onError, opt) => intResolve(str, 0, 10, opt),
    stringify: stringifyNumber
};
const intHex = {
    identify: intIdentify,
    default: true,
    tag: 'tag:yaml.org,2002:int',
    format: 'HEX',
    test: /^[-+]?0x[0-9a-fA-F_]+$/,
    resolve: (str, _onError, opt) => intResolve(str, 2, 16, opt),
    stringify: node => intStringify(node, 16, '0x')
};

class YAMLSet extends YAMLMap {
    constructor(schema) {
        super(schema);
        this.tag = YAMLSet.tag;
    }
    add(key) {
        let pair;
        if (isPair(key))
            pair = key;
        else if (key &&
            typeof key === 'object' &&
            'key' in key &&
            'value' in key &&
            key.value === null)
            pair = new Pair(key.key, null);
        else
            pair = new Pair(key, null);
        const prev = findPair(this.items, pair.key);
        if (!prev)
            this.items.push(pair);
    }
    /**
     * If `keepPair` is `true`, returns the Pair matching `key`.
     * Otherwise, returns the value of that Pair's key.
     */
    get(key, keepPair) {
        const pair = findPair(this.items, key);
        return !keepPair && isPair(pair)
            ? isScalar(pair.key)
                ? pair.key.value
                : pair.key
            : pair;
    }
    set(key, value) {
        if (typeof value !== 'boolean')
            throw new Error(`Expected boolean value for set(key, value) in a YAML set, not ${typeof value}`);
        const prev = findPair(this.items, key);
        if (prev && !value) {
            this.items.splice(this.items.indexOf(prev), 1);
        }
        else if (!prev && value) {
            this.items.push(new Pair(key));
        }
    }
    toJSON(_, ctx) {
        return super.toJSON(_, ctx, Set);
    }
    toString(ctx, onComment, onChompKeep) {
        if (!ctx)
            return JSON.stringify(this);
        if (this.hasAllNullValues(true))
            return super.toString(Object.assign({}, ctx, { allNullValues: true }), onComment, onChompKeep);
        else
            throw new Error('Set items must all have null values');
    }
    static from(schema, iterable, ctx) {
        const { replacer } = ctx;
        const set = new this(schema);
        if (iterable && Symbol.iterator in Object(iterable))
            for (let value of iterable) {
                if (typeof replacer === 'function')
                    value = replacer.call(iterable, value, value);
                set.items.push(createPair(value, null, ctx));
            }
        return set;
    }
}
YAMLSet.tag = 'tag:yaml.org,2002:set';
const set = {
    collection: 'map',
    identify: value => value instanceof Set,
    nodeClass: YAMLSet,
    default: false,
    tag: 'tag:yaml.org,2002:set',
    createNode: (schema, iterable, ctx) => YAMLSet.from(schema, iterable, ctx),
    resolve(map, onError) {
        if (isMap(map)) {
            if (map.hasAllNullValues(true))
                return Object.assign(new YAMLSet(), map);
            else
                onError('Set items must all have null values');
        }
        else
            onError('Expected a mapping for this tag');
        return map;
    }
};

/** Internal types handle bigint as number, because TS can't figure it out. */
function parseSexagesimal(str, asBigInt) {
    const sign = str[0];
    const parts = sign === '-' || sign === '+' ? str.substring(1) : str;
    const num = (n) => asBigInt ? BigInt(n) : Number(n);
    const res = parts
        .replace(/_/g, '')
        .split(':')
        .reduce((res, p) => res * num(60) + num(p), num(0));
    return (sign === '-' ? num(-1) * res : res);
}
/**
 * hhhh:mm:ss.sss
 *
 * Internal types handle bigint as number, because TS can't figure it out.
 */
function stringifySexagesimal(node) {
    let { value } = node;
    let num = (n) => n;
    if (typeof value === 'bigint')
        num = n => BigInt(n);
    else if (isNaN(value) || !isFinite(value))
        return stringifyNumber(node);
    let sign = '';
    if (value < 0) {
        sign = '-';
        value *= num(-1);
    }
    const _60 = num(60);
    const parts = [value % _60]; // seconds, including ms
    if (value < 60) {
        parts.unshift(0); // at least one : is required
    }
    else {
        value = (value - parts[0]) / _60;
        parts.unshift(value % _60); // minutes
        if (value >= 60) {
            value = (value - parts[0]) / _60;
            parts.unshift(value); // hours
        }
    }
    return (sign +
        parts
            .map(n => String(n).padStart(2, '0'))
            .join(':')
            .replace(/000000\d*$/, '') // % 60 may introduce error
    );
}
const intTime = {
    identify: value => typeof value === 'bigint' || Number.isInteger(value),
    default: true,
    tag: 'tag:yaml.org,2002:int',
    format: 'TIME',
    test: /^[-+]?[0-9][0-9_]*(?::[0-5]?[0-9])+$/,
    resolve: (str, _onError, { intAsBigInt }) => parseSexagesimal(str, intAsBigInt),
    stringify: stringifySexagesimal
};
const floatTime = {
    identify: value => typeof value === 'number',
    default: true,
    tag: 'tag:yaml.org,2002:float',
    format: 'TIME',
    test: /^[-+]?[0-9][0-9_]*(?::[0-5]?[0-9])+\.[0-9_]*$/,
    resolve: str => parseSexagesimal(str, false),
    stringify: stringifySexagesimal
};
const timestamp = {
    identify: value => value instanceof Date,
    default: true,
    tag: 'tag:yaml.org,2002:timestamp',
    // If the time zone is omitted, the timestamp is assumed to be specified in UTC. The time part
    // may be omitted altogether, resulting in a date format. In such a case, the time part is
    // assumed to be 00:00:00Z (start of day, UTC).
    test: RegExp('^([0-9]{4})-([0-9]{1,2})-([0-9]{1,2})' + // YYYY-Mm-Dd
        '(?:' + // time is optional
        '(?:t|T|[ \\t]+)' + // t | T | whitespace
        '([0-9]{1,2}):([0-9]{1,2}):([0-9]{1,2}(\\.[0-9]+)?)' + // Hh:Mm:Ss(.ss)?
        '(?:[ \\t]*(Z|[-+][012]?[0-9](?::[0-9]{2})?))?' + // Z | +5 | -03:30
        ')?$'),
    resolve(str) {
        const match = str.match(timestamp.test);
        if (!match)
            throw new Error('!!timestamp expects a date, starting with yyyy-mm-dd');
        const [, year, month, day, hour, minute, second] = match.map(Number);
        const millisec = match[7] ? Number((match[7] + '00').substr(1, 3)) : 0;
        let date = Date.UTC(year, month - 1, day, hour || 0, minute || 0, second || 0, millisec);
        const tz = match[8];
        if (tz && tz !== 'Z') {
            let d = parseSexagesimal(tz, false);
            if (Math.abs(d) < 30)
                d *= 60;
            date -= 60000 * d;
        }
        return new Date(date);
    },
    stringify: ({ value }) => value.toISOString().replace(/((T00:00)?:00)?\.000Z$/, '')
};

const schema = [
    map,
    seq,
    string,
    nullTag,
    trueTag,
    falseTag,
    intBin,
    intOct,
    int,
    intHex,
    floatNaN,
    floatExp,
    float,
    binary,
    omap,
    pairs,
    set,
    intTime,
    floatTime,
    timestamp
];

const schemas = new Map([
    ['core', schema$2],
    ['failsafe', [map, seq, string]],
    ['json', schema$1],
    ['yaml11', schema],
    ['yaml-1.1', schema]
]);
const tagsByName = {
    binary,
    bool: boolTag,
    float: float$1,
    floatExp: floatExp$1,
    floatNaN: floatNaN$1,
    floatTime,
    int: int$1,
    intHex: intHex$1,
    intOct: intOct$1,
    intTime,
    map,
    null: nullTag,
    omap,
    pairs,
    seq,
    set,
    timestamp
};
const coreKnownTags = {
    'tag:yaml.org,2002:binary': binary,
    'tag:yaml.org,2002:omap': omap,
    'tag:yaml.org,2002:pairs': pairs,
    'tag:yaml.org,2002:set': set,
    'tag:yaml.org,2002:timestamp': timestamp
};
function getTags(customTags, schemaName) {
    let tags = schemas.get(schemaName);
    if (!tags) {
        if (Array.isArray(customTags))
            tags = [];
        else {
            const keys = Array.from(schemas.keys())
                .filter(key => key !== 'yaml11')
                .map(key => JSON.stringify(key))
                .join(', ');
            throw new Error(`Unknown schema "${schemaName}"; use one of ${keys} or define customTags array`);
        }
    }
    if (Array.isArray(customTags)) {
        for (const tag of customTags)
            tags = tags.concat(tag);
    }
    else if (typeof customTags === 'function') {
        tags = customTags(tags.slice());
    }
    return tags.map(tag => {
        if (typeof tag !== 'string')
            return tag;
        const tagObj = tagsByName[tag];
        if (tagObj)
            return tagObj;
        const keys = Object.keys(tagsByName)
            .map(key => JSON.stringify(key))
            .join(', ');
        throw new Error(`Unknown custom tag "${tag}"; use one of ${keys}`);
    });
}

const sortMapEntriesByKey = (a, b) => a.key < b.key ? -1 : a.key > b.key ? 1 : 0;
class Schema {
    constructor({ compat, customTags, merge, resolveKnownTags, schema, sortMapEntries, toStringDefaults }) {
        this.compat = Array.isArray(compat)
            ? getTags(compat, 'compat')
            : compat
                ? getTags(null, compat)
                : null;
        this.merge = !!merge;
        this.name = (typeof schema === 'string' && schema) || 'core';
        this.knownTags = resolveKnownTags ? coreKnownTags : {};
        this.tags = getTags(customTags, this.name);
        this.toStringOptions = toStringDefaults ?? null;
        Object.defineProperty(this, MAP, { value: map });
        Object.defineProperty(this, SCALAR, { value: string });
        Object.defineProperty(this, SEQ, { value: seq });
        // Used by createMap()
        this.sortMapEntries =
            typeof sortMapEntries === 'function'
                ? sortMapEntries
                : sortMapEntries === true
                    ? sortMapEntriesByKey
                    : null;
    }
    clone() {
        const copy = Object.create(Schema.prototype, Object.getOwnPropertyDescriptors(this));
        copy.tags = this.tags.slice();
        return copy;
    }
}

function stringifyDocument(doc, options) {
    const lines = [];
    let hasDirectives = options.directives === true;
    if (options.directives !== false && doc.directives) {
        const dir = doc.directives.toString(doc);
        if (dir) {
            lines.push(dir);
            hasDirectives = true;
        }
        else if (doc.directives.docStart)
            hasDirectives = true;
    }
    if (hasDirectives)
        lines.push('---');
    const ctx = createStringifyContext(doc, options);
    const { commentString } = ctx.options;
    if (doc.commentBefore) {
        if (lines.length !== 1)
            lines.unshift('');
        const cs = commentString(doc.commentBefore);
        lines.unshift(indentComment(cs, ''));
    }
    let chompKeep = false;
    let contentComment = null;
    if (doc.contents) {
        if (isNode(doc.contents)) {
            if (doc.contents.spaceBefore && hasDirectives)
                lines.push('');
            if (doc.contents.commentBefore) {
                const cs = commentString(doc.contents.commentBefore);
                lines.push(indentComment(cs, ''));
            }
            // top-level block scalars need to be indented if followed by a comment
            ctx.forceBlockIndent = !!doc.comment;
            contentComment = doc.contents.comment;
        }
        const onChompKeep = contentComment ? undefined : () => (chompKeep = true);
        let body = stringify$1(doc.contents, ctx, () => (contentComment = null), onChompKeep);
        if (contentComment)
            body += lineComment(body, '', commentString(contentComment));
        if ((body[0] === '|' || body[0] === '>') &&
            lines[lines.length - 1] === '---') {
            // Top-level block scalars with a preceding doc marker ought to use the
            // same line for their header.
            lines[lines.length - 1] = `--- ${body}`;
        }
        else
            lines.push(body);
    }
    else {
        lines.push(stringify$1(doc.contents, ctx));
    }
    if (doc.directives?.docEnd) {
        if (doc.comment) {
            const cs = commentString(doc.comment);
            if (cs.includes('\n')) {
                lines.push('...');
                lines.push(indentComment(cs, ''));
            }
            else {
                lines.push(`... ${cs}`);
            }
        }
        else {
            lines.push('...');
        }
    }
    else {
        let dc = doc.comment;
        if (dc && chompKeep)
            dc = dc.replace(/^\n+/, '');
        if (dc) {
            if ((!chompKeep || contentComment) && lines[lines.length - 1] !== '')
                lines.push('');
            lines.push(indentComment(commentString(dc), ''));
        }
    }
    return lines.join('\n') + '\n';
}

class Document {
    constructor(value, replacer, options) {
        /** A comment before this Document */
        this.commentBefore = null;
        /** A comment immediately after this Document */
        this.comment = null;
        /** Errors encountered during parsing. */
        this.errors = [];
        /** Warnings encountered during parsing. */
        this.warnings = [];
        Object.defineProperty(this, NODE_TYPE, { value: DOC });
        let _replacer = null;
        if (typeof replacer === 'function' || Array.isArray(replacer)) {
            _replacer = replacer;
        }
        else if (options === undefined && replacer) {
            options = replacer;
            replacer = undefined;
        }
        const opt = Object.assign({
            intAsBigInt: false,
            keepSourceTokens: false,
            logLevel: 'warn',
            prettyErrors: true,
            strict: true,
            uniqueKeys: true,
            version: '1.2'
        }, options);
        this.options = opt;
        let { version } = opt;
        if (options?._directives) {
            this.directives = options._directives.atDocument();
            if (this.directives.yaml.explicit)
                version = this.directives.yaml.version;
        }
        else
            this.directives = new Directives({ version });
        this.setSchema(version, options);
        // @ts-expect-error We can't really know that this matches Contents.
        this.contents =
            value === undefined ? null : this.createNode(value, _replacer, options);
    }
    /**
     * Create a deep copy of this Document and its contents.
     *
     * Custom Node values that inherit from `Object` still refer to their original instances.
     */
    clone() {
        const copy = Object.create(Document.prototype, {
            [NODE_TYPE]: { value: DOC }
        });
        copy.commentBefore = this.commentBefore;
        copy.comment = this.comment;
        copy.errors = this.errors.slice();
        copy.warnings = this.warnings.slice();
        copy.options = Object.assign({}, this.options);
        if (this.directives)
            copy.directives = this.directives.clone();
        copy.schema = this.schema.clone();
        // @ts-expect-error We can't really know that this matches Contents.
        copy.contents = isNode(this.contents)
            ? this.contents.clone(copy.schema)
            : this.contents;
        if (this.range)
            copy.range = this.range.slice();
        return copy;
    }
    /** Adds a value to the document. */
    add(value) {
        if (assertCollection(this.contents))
            this.contents.add(value);
    }
    /** Adds a value to the document. */
    addIn(path, value) {
        if (assertCollection(this.contents))
            this.contents.addIn(path, value);
    }
    /**
     * Create a new `Alias` node, ensuring that the target `node` has the required anchor.
     *
     * If `node` already has an anchor, `name` is ignored.
     * Otherwise, the `node.anchor` value will be set to `name`,
     * or if an anchor with that name is already present in the document,
     * `name` will be used as a prefix for a new unique anchor.
     * If `name` is undefined, the generated anchor will use 'a' as a prefix.
     */
    createAlias(node, name) {
        if (!node.anchor) {
            const prev = anchorNames(this);
            node.anchor =
                // eslint-disable-next-line @typescript-eslint/prefer-nullish-coalescing
                !name || prev.has(name) ? findNewAnchor(name || 'a', prev) : name;
        }
        return new Alias(node.anchor);
    }
    createNode(value, replacer, options) {
        let _replacer = undefined;
        if (typeof replacer === 'function') {
            value = replacer.call({ '': value }, '', value);
            _replacer = replacer;
        }
        else if (Array.isArray(replacer)) {
            const keyToStr = (v) => typeof v === 'number' || v instanceof String || v instanceof Number;
            const asStr = replacer.filter(keyToStr).map(String);
            if (asStr.length > 0)
                replacer = replacer.concat(asStr);
            _replacer = replacer;
        }
        else if (options === undefined && replacer) {
            options = replacer;
            replacer = undefined;
        }
        const { aliasDuplicateObjects, anchorPrefix, flow, keepUndefined, onTagObj, tag } = options ?? {};
        const { onAnchor, setAnchors, sourceObjects } = createNodeAnchors(this, 
        // eslint-disable-next-line @typescript-eslint/prefer-nullish-coalescing
        anchorPrefix || 'a');
        const ctx = {
            aliasDuplicateObjects: aliasDuplicateObjects ?? true,
            keepUndefined: keepUndefined ?? false,
            onAnchor,
            onTagObj,
            replacer: _replacer,
            schema: this.schema,
            sourceObjects
        };
        const node = createNode(value, tag, ctx);
        if (flow && isCollection(node))
            node.flow = true;
        setAnchors();
        return node;
    }
    /**
     * Convert a key and a value into a `Pair` using the current schema,
     * recursively wrapping all values as `Scalar` or `Collection` nodes.
     */
    createPair(key, value, options = {}) {
        const k = this.createNode(key, null, options);
        const v = this.createNode(value, null, options);
        return new Pair(k, v);
    }
    /**
     * Removes a value from the document.
     * @returns `true` if the item was found and removed.
     */
    delete(key) {
        return assertCollection(this.contents) ? this.contents.delete(key) : false;
    }
    /**
     * Removes a value from the document.
     * @returns `true` if the item was found and removed.
     */
    deleteIn(path) {
        if (isEmptyPath(path)) {
            if (this.contents == null)
                return false;
            // @ts-expect-error Presumed impossible if Strict extends false
            this.contents = null;
            return true;
        }
        return assertCollection(this.contents)
            ? this.contents.deleteIn(path)
            : false;
    }
    /**
     * Returns item at `key`, or `undefined` if not found. By default unwraps
     * scalar values from their surrounding node; to disable set `keepScalar` to
     * `true` (collections are always returned intact).
     */
    get(key, keepScalar) {
        return isCollection(this.contents)
            ? this.contents.get(key, keepScalar)
            : undefined;
    }
    /**
     * Returns item at `path`, or `undefined` if not found. By default unwraps
     * scalar values from their surrounding node; to disable set `keepScalar` to
     * `true` (collections are always returned intact).
     */
    getIn(path, keepScalar) {
        if (isEmptyPath(path))
            return !keepScalar && isScalar(this.contents)
                ? this.contents.value
                : this.contents;
        return isCollection(this.contents)
            ? this.contents.getIn(path, keepScalar)
            : undefined;
    }
    /**
     * Checks if the document includes a value with the key `key`.
     */
    has(key) {
        return isCollection(this.contents) ? this.contents.has(key) : false;
    }
    /**
     * Checks if the document includes a value at `path`.
     */
    hasIn(path) {
        if (isEmptyPath(path))
            return this.contents !== undefined;
        return isCollection(this.contents) ? this.contents.hasIn(path) : false;
    }
    /**
     * Sets a value in this document. For `!!set`, `value` needs to be a
     * boolean to add/remove the item from the set.
     */
    set(key, value) {
        if (this.contents == null) {
            // @ts-expect-error We can't really know that this matches Contents.
            this.contents = collectionFromPath(this.schema, [key], value);
        }
        else if (assertCollection(this.contents)) {
            this.contents.set(key, value);
        }
    }
    /**
     * Sets a value in this document. For `!!set`, `value` needs to be a
     * boolean to add/remove the item from the set.
     */
    setIn(path, value) {
        if (isEmptyPath(path)) {
            // @ts-expect-error We can't really know that this matches Contents.
            this.contents = value;
        }
        else if (this.contents == null) {
            // @ts-expect-error We can't really know that this matches Contents.
            this.contents = collectionFromPath(this.schema, Array.from(path), value);
        }
        else if (assertCollection(this.contents)) {
            this.contents.setIn(path, value);
        }
    }
    /**
     * Change the YAML version and schema used by the document.
     * A `null` version disables support for directives, explicit tags, anchors, and aliases.
     * It also requires the `schema` option to be given as a `Schema` instance value.
     *
     * Overrides all previously set schema options.
     */
    setSchema(version, options = {}) {
        if (typeof version === 'number')
            version = String(version);
        let opt;
        switch (version) {
            case '1.1':
                if (this.directives)
                    this.directives.yaml.version = '1.1';
                else
                    this.directives = new Directives({ version: '1.1' });
                opt = { merge: true, resolveKnownTags: false, schema: 'yaml-1.1' };
                break;
            case '1.2':
            case 'next':
                if (this.directives)
                    this.directives.yaml.version = version;
                else
                    this.directives = new Directives({ version });
                opt = { merge: false, resolveKnownTags: true, schema: 'core' };
                break;
            case null:
                if (this.directives)
                    delete this.directives;
                opt = null;
                break;
            default: {
                const sv = JSON.stringify(version);
                throw new Error(`Expected '1.1', '1.2' or null as first argument, but found: ${sv}`);
            }
        }
        // Not using `instanceof Schema` to allow for duck typing
        if (options.schema instanceof Object)
            this.schema = options.schema;
        else if (opt)
            this.schema = new Schema(Object.assign(opt, options));
        else
            throw new Error(`With a null YAML version, the { schema: Schema } option is required`);
    }
    // json & jsonArg are only used from toJSON()
    toJS({ json, jsonArg, mapAsMap, maxAliasCount, onAnchor, reviver } = {}) {
        const ctx = {
            anchors: new Map(),
            doc: this,
            keep: !json,
            mapAsMap: mapAsMap === true,
            mapKeyWarned: false,
            maxAliasCount: typeof maxAliasCount === 'number' ? maxAliasCount : 100
        };
        const res = toJS(this.contents, jsonArg ?? '', ctx);
        if (typeof onAnchor === 'function')
            for (const { count, res } of ctx.anchors.values())
                onAnchor(res, count);
        return typeof reviver === 'function'
            ? applyReviver(reviver, { '': res }, '', res)
            : res;
    }
    /**
     * A JSON representation of the document `contents`.
     *
     * @param jsonArg Used by `JSON.stringify` to indicate the array index or
     *   property name.
     */
    toJSON(jsonArg, onAnchor) {
        return this.toJS({ json: true, jsonArg, mapAsMap: false, onAnchor });
    }
    /** A YAML representation of the document. */
    toString(options = {}) {
        if (this.errors.length > 0)
            throw new Error('Document with errors cannot be stringified');
        if ('indent' in options &&
            (!Number.isInteger(options.indent) || Number(options.indent) <= 0)) {
            const s = JSON.stringify(options.indent);
            throw new Error(`"indent" option must be a positive integer, not ${s}`);
        }
        return stringifyDocument(this, options);
    }
}
function assertCollection(contents) {
    if (isCollection(contents))
        return true;
    throw new Error('Expected a YAML collection as document contents');
}

new Set('0123456789ABCDEFabcdef');
new Set("0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz-#;/?:@&=+$_.!~*'()");
new Set(',[]{}');
new Set(' ,[]{}\n\r\t');

function stringify(value, replacer, options) {
    let _replacer = null;
    if (typeof replacer === 'function' || Array.isArray(replacer)) {
        _replacer = replacer;
    }
    else if (options === undefined && replacer) {
        options = replacer;
    }
    if (typeof options === 'string')
        options = options.length;
    if (typeof options === 'number') {
        const indent = Math.round(options);
        options = indent < 1 ? undefined : indent > 8 ? { indent: 8 } : { indent };
    }
    if (value === undefined) {
        const { keepUndefined } = options ?? replacer ?? {};
        if (!keepUndefined)
            return undefined;
    }
    return new Document(value, _replacer, options).toString(options);
}

const defaultDecoratorsAllowList = [
    "TypeSpec\\.Xml\\..*",
    "Azure\\.Core\\.@useFinalStateVia",
    "Autorest\\.@example",
];

// these two vars are used to validate whether our SdkBuiltInKinds are exhaustive for all possible values from typespec
// if it is not, a typescript compilation error will be thrown here.
var SdkIntKindsEnum;
(function (SdkIntKindsEnum) {
    SdkIntKindsEnum["numeric"] = "numeric";
    SdkIntKindsEnum["integer"] = "integer";
    SdkIntKindsEnum["safeint"] = "safeint";
    SdkIntKindsEnum["int8"] = "int8";
    SdkIntKindsEnum["int16"] = "int16";
    SdkIntKindsEnum["int32"] = "int32";
    SdkIntKindsEnum["int64"] = "int64";
    SdkIntKindsEnum["uint8"] = "uint8";
    SdkIntKindsEnum["uint16"] = "uint16";
    SdkIntKindsEnum["uint32"] = "uint32";
    SdkIntKindsEnum["uint64"] = "uint64";
})(SdkIntKindsEnum || (SdkIntKindsEnum = {}));
var SdkFloatingPointKindsEnum;
(function (SdkFloatingPointKindsEnum) {
    SdkFloatingPointKindsEnum["float"] = "float";
    SdkFloatingPointKindsEnum["float32"] = "float32";
    SdkFloatingPointKindsEnum["float64"] = "float64";
})(SdkFloatingPointKindsEnum || (SdkFloatingPointKindsEnum = {}));
var SdkFixedPointKindsEnum;
(function (SdkFixedPointKindsEnum) {
    SdkFixedPointKindsEnum["decimal"] = "decimal";
    SdkFixedPointKindsEnum["decimal128"] = "decimal128";
})(SdkFixedPointKindsEnum || (SdkFixedPointKindsEnum = {}));
var SdkGenericBuiltInStringKindsEnum;
(function (SdkGenericBuiltInStringKindsEnum) {
    SdkGenericBuiltInStringKindsEnum["string"] = "string";
    SdkGenericBuiltInStringKindsEnum["url"] = "url";
})(SdkGenericBuiltInStringKindsEnum || (SdkGenericBuiltInStringKindsEnum = {}));
var SdkBuiltInKindsMiscellaneousEnum;
(function (SdkBuiltInKindsMiscellaneousEnum) {
    SdkBuiltInKindsMiscellaneousEnum["bytes"] = "bytes";
    SdkBuiltInKindsMiscellaneousEnum["boolean"] = "boolean";
    SdkBuiltInKindsMiscellaneousEnum["plainDate"] = "plainDate";
    SdkBuiltInKindsMiscellaneousEnum["plainTime"] = "plainTime";
    SdkBuiltInKindsMiscellaneousEnum["unknown"] = "unknown";
})(SdkBuiltInKindsMiscellaneousEnum || (SdkBuiltInKindsMiscellaneousEnum = {}));
function getKnownScalars() {
    const retval = {};
    const typespecNamespace = Object.keys(SdkBuiltInKindsMiscellaneousEnum)
        .concat(Object.keys(SdkIntKindsEnum))
        .concat(Object.keys(SdkFloatingPointKindsEnum))
        .concat(Object.keys(SdkFixedPointKindsEnum))
        .concat(Object.keys(SdkGenericBuiltInStringKindsEnum));
    for (const kind of typespecNamespace) {
        if (!isSdkBuiltInKind(kind))
            continue; // it will always be true
        retval[`TypeSpec.${kind}`] = kind;
    }
    return retval;
}
function isSdkBuiltInKind(kind) {
    return (kind in SdkBuiltInKindsMiscellaneousEnum ||
        isSdkIntKind(kind) ||
        isSdkFloatKind(kind) ||
        isSdkFixedPointKind(kind) ||
        kind in SdkGenericBuiltInStringKindsEnum);
}
function isSdkIntKind(kind) {
    return kind in SdkIntKindsEnum;
}
function isSdkFloatKind(kind) {
    return kind in SdkFloatingPointKindsEnum;
}
function isSdkFixedPointKind(kind) {
    return kind in SdkFixedPointKindsEnum;
}
const SdkDateTimeEncodingsConst = ["rfc3339", "rfc7231", "unixTimestamp"];
function isSdkDateTimeEncodings(encoding) {
    return SdkDateTimeEncodingsConst.includes(encoding);
}
/**
 * This enum represents the different ways a model can be used in a method.
 */
var UsageFlags;
(function (UsageFlags) {
    UsageFlags[UsageFlags["None"] = 0] = "None";
    UsageFlags[UsageFlags["Input"] = 2] = "Input";
    UsageFlags[UsageFlags["Output"] = 4] = "Output";
    UsageFlags[UsageFlags["ApiVersionEnum"] = 8] = "ApiVersionEnum";
    // Input and Json will also be set when JsonMergePatch is set.
    UsageFlags[UsageFlags["JsonMergePatch"] = 16] = "JsonMergePatch";
    // Input will also be set when MultipartFormData is set.
    UsageFlags[UsageFlags["MultipartFormData"] = 32] = "MultipartFormData";
    // Used in spread.
    UsageFlags[UsageFlags["Spread"] = 64] = "Spread";
    /**
     * @deprecated Use `Exception` instead.
     */
    // Output will also be set when Error is set.
    UsageFlags[UsageFlags["Error"] = 128] = "Error";
    // Set when type is used in conjunction with an application/json content type.
    UsageFlags[UsageFlags["Json"] = 256] = "Json";
    // Set when type is used in conjunction with an application/xml content type.
    UsageFlags[UsageFlags["Xml"] = 512] = "Xml";
    // Set when type is used for exception output.
    UsageFlags[UsageFlags["Exception"] = 1024] = "Exception";
    // Set when type is used as LRO initial response.
    UsageFlags[UsageFlags["LroInitial"] = 2048] = "LroInitial";
    // Set when type is used as LRO polling response.
    UsageFlags[UsageFlags["LroPolling"] = 4096] = "LroPolling";
    // Set when type is used as LRO final envelop response.
    UsageFlags[UsageFlags["LroFinalEnvelope"] = 8192] = "LroFinalEnvelope";
})(UsageFlags || (UsageFlags = {}));

const EmitterOptionsSchema = {
    type: "object",
    additionalProperties: true,
    properties: {
        "generate-protocol-methods": { type: "boolean", nullable: true },
        "generate-convenience-methods": { type: "boolean", nullable: true },
        "package-name": { type: "string", nullable: true },
        "flatten-union-as-enum": { type: "boolean", nullable: true },
        "api-version": { type: "string", nullable: true },
        "examples-directory": { type: "string", nullable: true },
        "examples-dir": { type: "string", nullable: true },
        "emitter-name": { type: "string", nullable: true },
    },
};
const $lib = createTypeSpecLibrary({
    name: "@azure-tools/typespec-client-generator-core",
    diagnostics: {
        "client-name": {
            severity: "warning",
            messages: {
                default: paramMessage `Client name "${"name"}" must end with Client. Use @client({name: "...Client"})`,
            },
        },
        "client-service": {
            severity: "warning",
            messages: {
                default: paramMessage `Client "${"name"}" is not inside a service namespace. Use @client({service: MyServiceNS})`,
            },
        },
        "unknown-client-format": {
            severity: "error",
            messages: {
                default: paramMessage `Client format "${"format"}" is unknown. Known values are "${"knownValues"}"`,
            },
        },
        "incorrect-client-format": {
            severity: "error",
            messages: {
                default: paramMessage `Format "${"format"}" can only apply to "${"expectedTargetTypes"}"`,
            },
        },
        "union-null": {
            severity: "error",
            messages: {
                default: "Cannot have a union containing only null types.",
            },
        },
        "union-unsupported": {
            severity: "error",
            messages: {
                default: "Unions cannot be emitted by our language generators unless all options are literals of the same type.",
                null: "Unions containing multiple model types cannot be emitted unless the union is between one model type and 'null'.",
            },
        },
        "use-enum-instead": {
            severity: "warning",
            messages: {
                default: "Use enum instead of union of string or number literals. Falling back to the literal type.",
            },
        },
        access: {
            severity: "error",
            messages: {
                default: `Access decorator value must be "public" or "internal".`,
            },
        },
        "invalid-usage": {
            severity: "error",
            messages: {
                default: `Usage decorator value must be 2 ("input") or 4 ("output").`,
            },
        },
        "invalid-encode": {
            severity: "error",
            messages: {
                default: "Invalid encoding",
                wrongType: paramMessage `Encoding '${"encoding"}' cannot be used on type '${"type"}'`,
            },
        },
        "conflicting-multipart-model-usage": {
            severity: "error",
            messages: {
                default: "Invalid encoding",
                wrongType: paramMessage `Model '${"modelName"}' cannot be used as both multipart/form-data input and regular body input. You can create a separate model with name 'model ${"modelName"}FormData' extends ${"modelName"} {}`,
            },
        },
        "discriminator-not-constant": {
            severity: "error",
            messages: {
                default: paramMessage `Discriminator ${"discriminator"} has to be constant`,
            },
        },
        "discriminator-not-string": {
            severity: "warning",
            messages: {
                default: paramMessage `Value of discriminator ${"discriminator"} has to be a string, not ${"discriminatorValue"}`,
            },
        },
        "wrong-client-decorator": {
            severity: "warning",
            messages: {
                default: "@client or @operationGroup should decorate namespace or interface in client.tsp",
            },
        },
        "encoding-multipart-bytes": {
            severity: "error",
            messages: {
                default: "Encoding should not be applied to bytes content in a multipart request. This is semi-incompatible with how multipart works in HTTP.",
            },
        },
        "unsupported-kind": {
            severity: "warning",
            messages: {
                default: paramMessage `Unsupported kind ${"kind"}`,
            },
        },
        "multiple-services": {
            severity: "warning",
            messages: {
                default: paramMessage `Multiple services found in definition. Only one service is supported, so we will choose the first one ${"service"}`,
            },
        },
        "server-param-not-path": {
            severity: "error",
            messages: {
                default: paramMessage `Template argument ${"templateArgumentName"} is not a path parameter, it is a ${"templateArgumentType"}. It has to be a path.`,
            },
        },
        "unexpected-http-param-type": {
            severity: "error",
            messages: {
                default: paramMessage `Expected parameter "${"paramName"}" to be of type "${"expectedType"}", but instead it is of type "${"actualType"}"`,
            },
        },
        "multiple-response-types": {
            severity: "warning",
            messages: {
                default: paramMessage `Multiple response types found in operation ${"operation"}. Only one response type is supported, so we will choose the first one ${"response"}`,
            },
        },
        "no-corresponding-method-param": {
            severity: "error",
            messages: {
                default: paramMessage `Missing "${"paramName"}" method parameter in method "${"methodName"}", when "${"paramName"}" must be sent to the service. Add a parameter named "${"paramName"}" to the method.`,
            },
        },
        "unsupported-protocol": {
            severity: "error",
            messages: {
                default: "Currently we only support HTTP and HTTPS protocols",
            },
        },
        "no-emitter-name": {
            severity: "warning",
            messages: {
                default: "Can not find name for your emitter, please check your emitter name.",
            },
        },
        "unsupported-generic-decorator-arg-type": {
            severity: "warning",
            messages: {
                default: paramMessage `Can not parse the arg type for decorator "${"decoratorName"}".`,
            },
        },
        "empty-client-name": {
            severity: "warning",
            messages: {
                default: `Cannot pass an empty value to the @clientName decorator`,
            },
        },
        "override-method-parameters-mismatch": {
            severity: "error",
            messages: {
                default: paramMessage `Method "${"methodName"}" is not directly referencing the same parameters as in the original operation. The original method has parameters "${"originalParameters"}", while the override method has parameters "${"overrideParameters"}".`,
            },
        },
        "duplicate-client-name": {
            severity: "error",
            messages: {
                default: paramMessage `Client name: "${"name"}" is duplicated in language scope: "${"scope"}"`,
                nonDecorator: paramMessage `Client name: "${"name"}" is defined somewhere causing nameing conflicts in language scope: "${"scope"}"`,
            },
        },
        "example-loading": {
            severity: "warning",
            messages: {
                default: paramMessage `Skipped loading invalid example file: ${"filename"}. Error: ${"error"}`,
                noDirectory: paramMessage `Skipping example loading from ${"directory"} because there was an error reading the directory.`,
                noOperationId: paramMessage `Skipping example file ${"filename"} because it does not contain an operationId and/or title.`,
            },
        },
        "duplicate-example-file": {
            severity: "error",
            messages: {
                default: paramMessage `Example file ${"filename"} uses duplicate title '${"title"}' for operationId '${"operationId"}'`,
            },
        },
        "example-value-no-mapping": {
            severity: "warning",
            messages: {
                default: paramMessage `Value in example file '${"relativePath"}' does not follow its definition:\n${"value"}`,
            },
        },
        "flatten-polymorphism": {
            severity: "error",
            messages: {
                default: `Cannot flatten property of polymorphic type.`,
            },
        },
        "conflict-access-override": {
            severity: "warning",
            messages: {
                default: `@access override conflicts with the access calculated from operation or other @access override.`,
            },
        },
        "conflict-usage-override": {
            severity: "warning",
            messages: {
                default: `@usage override conflicts with the usage calculated from operation or other @usage override.`,
            },
        },
        "duplicate-decorator": {
            severity: "warning",
            messages: {
                default: paramMessage `Decorator ${"decoratorName"} cannot be used twice on the same declaration with same scope.`,
            },
        },
        "empty-client-namespace": {
            severity: "warning",
            messages: {
                default: `Cannot pass an empty value to the @clientNamespace decorator`,
            },
        },
        "unexpected-pageable-operation-return-type": {
            severity: "error",
            messages: {
                default: `Operation is pageable but does not return a correct type.`,
            },
        },
        "invalid-alternate-source-type": {
            severity: "error",
            messages: {
                default: `@alternateType only supports scalar types. The source type is '${"typeName"}'.`,
            },
        },
    },
    emitter: {
        options: EmitterOptionsSchema,
    },
});
const { reportDiagnostic, createDiagnostic, createStateSymbol } = $lib;

// Regexps involved with splitting words in various case formats.
const SPLIT_LOWER_UPPER_RE = /([\p{Ll}\d])(\p{Lu})/gu;
const SPLIT_UPPER_UPPER_RE = /(\p{Lu})([\p{Lu}][\p{Ll}])/gu;
// Used to iterate over the initial split result and separate numbers.
const SPLIT_SEPARATE_NUMBER_RE = /(\d)\p{Ll}|(\p{L})\d/u;
// Regexp involved with stripping non-word characters from the result.
const DEFAULT_STRIP_REGEXP = /[^\p{L}\d]+/giu;
// The replacement value for splits.
const SPLIT_REPLACE_VALUE = "$1\0$2";
// The default characters to keep after transforming case.
const DEFAULT_PREFIX_SUFFIX_CHARACTERS = "";
/**
 * Split any cased input strings into an array of words.
 */
function split(value) {
    let result = value.trim();
    result = result
        .replace(SPLIT_LOWER_UPPER_RE, SPLIT_REPLACE_VALUE)
        .replace(SPLIT_UPPER_UPPER_RE, SPLIT_REPLACE_VALUE);
    result = result.replace(DEFAULT_STRIP_REGEXP, "\0");
    let start = 0;
    let end = result.length;
    // Trim the delimiter from around the output string.
    while (result.charAt(start) === "\0")
        start++;
    if (start === end)
        return [];
    while (result.charAt(end - 1) === "\0")
        end--;
    return result.slice(start, end).split(/\0/g);
}
/**
 * Split the input string into an array of words, separating numbers.
 */
function splitSeparateNumbers(value) {
    const words = split(value);
    for (let i = 0; i < words.length; i++) {
        const word = words[i];
        const match = SPLIT_SEPARATE_NUMBER_RE.exec(word);
        if (match) {
            const offset = match.index + (match[1] ?? match[2]).length;
            words.splice(i, 1, word.slice(0, offset), word.slice(offset));
        }
    }
    return words;
}
/**
 * Convert a string to camel case (`fooBar`).
 */
function camelCase(input, options) {
    const [prefix, words, suffix] = splitPrefixSuffix(input, options);
    const lower = lowerFactory(options?.locale);
    const upper = upperFactory(options?.locale);
    const transform = pascalCaseTransformFactory(lower, upper);
    return (prefix +
        words
            .map((word, index) => {
            if (index === 0)
                return lower(word);
            return transform(word, index);
        })
            .join("") +
        suffix);
}
/**
 * Convert a string to pascal case (`FooBar`).
 */
function pascalCase(input, options) {
    const [prefix, words, suffix] = splitPrefixSuffix(input, options);
    const lower = lowerFactory(options?.locale);
    const upper = upperFactory(options?.locale);
    const transform = pascalCaseTransformFactory(lower, upper);
    return prefix + words.map(transform).join("") + suffix;
}
function lowerFactory(locale) {
    return (input) => input.toLocaleLowerCase(locale);
}
function upperFactory(locale) {
    return (input) => input.toLocaleUpperCase(locale);
}
function pascalCaseTransformFactory(lower, upper) {
    return (word, index) => {
        const char0 = word[0];
        const initial = index > 0 && char0 >= "0" && char0 <= "9" ? "_" + char0 : upper(char0);
        return initial + lower(word.slice(1));
    };
}
function splitPrefixSuffix(input, options = {}) {
    const splitFn = options.split ?? (options.separateNumbers ? splitSeparateNumbers : split);
    const prefixCharacters = options.prefixCharacters ?? DEFAULT_PREFIX_SUFFIX_CHARACTERS;
    const suffixCharacters = options.suffixCharacters ?? DEFAULT_PREFIX_SUFFIX_CHARACTERS;
    let prefixIndex = 0;
    let suffixIndex = input.length;
    while (prefixIndex < input.length) {
        const char = input.charAt(prefixIndex);
        if (!prefixCharacters.includes(char))
            break;
        prefixIndex++;
    }
    while (suffixIndex > prefixIndex) {
        const index = suffixIndex - 1;
        const char = input.charAt(index);
        if (!suffixCharacters.includes(char))
            break;
        suffixIndex = index;
    }
    return [
        input.slice(0, prefixIndex),
        splitFn(input.slice(prefixIndex, suffixIndex)),
        input.slice(suffixIndex),
    ];
}

function getDefaultExportFromCjs (x) {
	return x && x.__esModule && Object.prototype.hasOwnProperty.call(x, 'default') ? x['default'] : x;
}

function commonjsRequire(path) {
	throw new Error('Could not dynamically require "' + path + '". Please configure the dynamicRequireTargets or/and ignoreDynamicRequires option of @rollup/plugin-commonjs appropriately for this require call to work.');
}

var pluralize$2 = {exports: {}};

/* global define */
var pluralize$1 = pluralize$2.exports;

var hasRequiredPluralize;

function requirePluralize () {
	if (hasRequiredPluralize) return pluralize$2.exports;
	hasRequiredPluralize = 1;
	(function (module, exports) {
		(function (root, pluralize) {
		  /* istanbul ignore else */
		  if (typeof commonjsRequire === 'function' && 'object' === 'object' && 'object' === 'object') {
		    // Node.
		    module.exports = pluralize();
		  } else {
		    // Browser global.
		    root.pluralize = pluralize();
		  }
		})(pluralize$1, function () {
		  // Rule storage - pluralize and singularize need to be run sequentially,
		  // while other rules can be optimized using an object for instant lookups.
		  var pluralRules = [];
		  var singularRules = [];
		  var uncountables = {};
		  var irregularPlurals = {};
		  var irregularSingles = {};

		  /**
		   * Sanitize a pluralization rule to a usable regular expression.
		   *
		   * @param  {(RegExp|string)} rule
		   * @return {RegExp}
		   */
		  function sanitizeRule (rule) {
		    if (typeof rule === 'string') {
		      return new RegExp('^' + rule + '$', 'i');
		    }

		    return rule;
		  }

		  /**
		   * Pass in a word token to produce a function that can replicate the case on
		   * another word.
		   *
		   * @param  {string}   word
		   * @param  {string}   token
		   * @return {Function}
		   */
		  function restoreCase (word, token) {
		    // Tokens are an exact match.
		    if (word === token) return token;

		    // Lower cased words. E.g. "hello".
		    if (word === word.toLowerCase()) return token.toLowerCase();

		    // Upper cased words. E.g. "WHISKY".
		    if (word === word.toUpperCase()) return token.toUpperCase();

		    // Title cased words. E.g. "Title".
		    if (word[0] === word[0].toUpperCase()) {
		      return token.charAt(0).toUpperCase() + token.substr(1).toLowerCase();
		    }

		    // Lower cased words. E.g. "test".
		    return token.toLowerCase();
		  }

		  /**
		   * Interpolate a regexp string.
		   *
		   * @param  {string} str
		   * @param  {Array}  args
		   * @return {string}
		   */
		  function interpolate (str, args) {
		    return str.replace(/\$(\d{1,2})/g, function (match, index) {
		      return args[index] || '';
		    });
		  }

		  /**
		   * Replace a word using a rule.
		   *
		   * @param  {string} word
		   * @param  {Array}  rule
		   * @return {string}
		   */
		  function replace (word, rule) {
		    return word.replace(rule[0], function (match, index) {
		      var result = interpolate(rule[1], arguments);

		      if (match === '') {
		        return restoreCase(word[index - 1], result);
		      }

		      return restoreCase(match, result);
		    });
		  }

		  /**
		   * Sanitize a word by passing in the word and sanitization rules.
		   *
		   * @param  {string}   token
		   * @param  {string}   word
		   * @param  {Array}    rules
		   * @return {string}
		   */
		  function sanitizeWord (token, word, rules) {
		    // Empty string or doesn't need fixing.
		    if (!token.length || uncountables.hasOwnProperty(token)) {
		      return word;
		    }

		    var len = rules.length;

		    // Iterate over the sanitization rules and use the first one to match.
		    while (len--) {
		      var rule = rules[len];

		      if (rule[0].test(word)) return replace(word, rule);
		    }

		    return word;
		  }

		  /**
		   * Replace a word with the updated word.
		   *
		   * @param  {Object}   replaceMap
		   * @param  {Object}   keepMap
		   * @param  {Array}    rules
		   * @return {Function}
		   */
		  function replaceWord (replaceMap, keepMap, rules) {
		    return function (word) {
		      // Get the correct token and case restoration functions.
		      var token = word.toLowerCase();

		      // Check against the keep object map.
		      if (keepMap.hasOwnProperty(token)) {
		        return restoreCase(word, token);
		      }

		      // Check against the replacement map for a direct word replacement.
		      if (replaceMap.hasOwnProperty(token)) {
		        return restoreCase(word, replaceMap[token]);
		      }

		      // Run all the rules against the word.
		      return sanitizeWord(token, word, rules);
		    };
		  }

		  /**
		   * Check if a word is part of the map.
		   */
		  function checkWord (replaceMap, keepMap, rules, bool) {
		    return function (word) {
		      var token = word.toLowerCase();

		      if (keepMap.hasOwnProperty(token)) return true;
		      if (replaceMap.hasOwnProperty(token)) return false;

		      return sanitizeWord(token, token, rules) === token;
		    };
		  }

		  /**
		   * Pluralize or singularize a word based on the passed in count.
		   *
		   * @param  {string}  word      The word to pluralize
		   * @param  {number}  count     How many of the word exist
		   * @param  {boolean} inclusive Whether to prefix with the number (e.g. 3 ducks)
		   * @return {string}
		   */
		  function pluralize (word, count, inclusive) {
		    var pluralized = count === 1
		      ? pluralize.singular(word) : pluralize.plural(word);

		    return (inclusive ? count + ' ' : '') + pluralized;
		  }

		  /**
		   * Pluralize a word.
		   *
		   * @type {Function}
		   */
		  pluralize.plural = replaceWord(
		    irregularSingles, irregularPlurals, pluralRules
		  );

		  /**
		   * Check if a word is plural.
		   *
		   * @type {Function}
		   */
		  pluralize.isPlural = checkWord(
		    irregularSingles, irregularPlurals, pluralRules
		  );

		  /**
		   * Singularize a word.
		   *
		   * @type {Function}
		   */
		  pluralize.singular = replaceWord(
		    irregularPlurals, irregularSingles, singularRules
		  );

		  /**
		   * Check if a word is singular.
		   *
		   * @type {Function}
		   */
		  pluralize.isSingular = checkWord(
		    irregularPlurals, irregularSingles, singularRules
		  );

		  /**
		   * Add a pluralization rule to the collection.
		   *
		   * @param {(string|RegExp)} rule
		   * @param {string}          replacement
		   */
		  pluralize.addPluralRule = function (rule, replacement) {
		    pluralRules.push([sanitizeRule(rule), replacement]);
		  };

		  /**
		   * Add a singularization rule to the collection.
		   *
		   * @param {(string|RegExp)} rule
		   * @param {string}          replacement
		   */
		  pluralize.addSingularRule = function (rule, replacement) {
		    singularRules.push([sanitizeRule(rule), replacement]);
		  };

		  /**
		   * Add an uncountable word rule.
		   *
		   * @param {(string|RegExp)} word
		   */
		  pluralize.addUncountableRule = function (word) {
		    if (typeof word === 'string') {
		      uncountables[word.toLowerCase()] = true;
		      return;
		    }

		    // Set singular and plural references for the word.
		    pluralize.addPluralRule(word, '$0');
		    pluralize.addSingularRule(word, '$0');
		  };

		  /**
		   * Add an irregular word definition.
		   *
		   * @param {string} single
		   * @param {string} plural
		   */
		  pluralize.addIrregularRule = function (single, plural) {
		    plural = plural.toLowerCase();
		    single = single.toLowerCase();

		    irregularSingles[single] = plural;
		    irregularPlurals[plural] = single;
		  };

		  /**
		   * Irregular rules.
		   */
		  [
		    // Pronouns.
		    ['I', 'we'],
		    ['me', 'us'],
		    ['he', 'they'],
		    ['she', 'they'],
		    ['them', 'them'],
		    ['myself', 'ourselves'],
		    ['yourself', 'yourselves'],
		    ['itself', 'themselves'],
		    ['herself', 'themselves'],
		    ['himself', 'themselves'],
		    ['themself', 'themselves'],
		    ['is', 'are'],
		    ['was', 'were'],
		    ['has', 'have'],
		    ['this', 'these'],
		    ['that', 'those'],
		    // Words ending in with a consonant and `o`.
		    ['echo', 'echoes'],
		    ['dingo', 'dingoes'],
		    ['volcano', 'volcanoes'],
		    ['tornado', 'tornadoes'],
		    ['torpedo', 'torpedoes'],
		    // Ends with `us`.
		    ['genus', 'genera'],
		    ['viscus', 'viscera'],
		    // Ends with `ma`.
		    ['stigma', 'stigmata'],
		    ['stoma', 'stomata'],
		    ['dogma', 'dogmata'],
		    ['lemma', 'lemmata'],
		    ['schema', 'schemata'],
		    ['anathema', 'anathemata'],
		    // Other irregular rules.
		    ['ox', 'oxen'],
		    ['axe', 'axes'],
		    ['die', 'dice'],
		    ['yes', 'yeses'],
		    ['foot', 'feet'],
		    ['eave', 'eaves'],
		    ['goose', 'geese'],
		    ['tooth', 'teeth'],
		    ['quiz', 'quizzes'],
		    ['human', 'humans'],
		    ['proof', 'proofs'],
		    ['carve', 'carves'],
		    ['valve', 'valves'],
		    ['looey', 'looies'],
		    ['thief', 'thieves'],
		    ['groove', 'grooves'],
		    ['pickaxe', 'pickaxes'],
		    ['passerby', 'passersby']
		  ].forEach(function (rule) {
		    return pluralize.addIrregularRule(rule[0], rule[1]);
		  });

		  /**
		   * Pluralization rules.
		   */
		  [
		    [/s?$/i, 's'],
		    [/[^\u0000-\u007F]$/i, '$0'],
		    [/([^aeiou]ese)$/i, '$1'],
		    [/(ax|test)is$/i, '$1es'],
		    [/(alias|[^aou]us|t[lm]as|gas|ris)$/i, '$1es'],
		    [/(e[mn]u)s?$/i, '$1s'],
		    [/([^l]ias|[aeiou]las|[ejzr]as|[iu]am)$/i, '$1'],
		    [/(alumn|syllab|vir|radi|nucle|fung|cact|stimul|termin|bacill|foc|uter|loc|strat)(?:us|i)$/i, '$1i'],
		    [/(alumn|alg|vertebr)(?:a|ae)$/i, '$1ae'],
		    [/(seraph|cherub)(?:im)?$/i, '$1im'],
		    [/(her|at|gr)o$/i, '$1oes'],
		    [/(agend|addend|millenni|dat|extrem|bacteri|desiderat|strat|candelabr|errat|ov|symposi|curricul|automat|quor)(?:a|um)$/i, '$1a'],
		    [/(apheli|hyperbat|periheli|asyndet|noumen|phenomen|criteri|organ|prolegomen|hedr|automat)(?:a|on)$/i, '$1a'],
		    [/sis$/i, 'ses'],
		    [/(?:(kni|wi|li)fe|(ar|l|ea|eo|oa|hoo)f)$/i, '$1$2ves'],
		    [/([^aeiouy]|qu)y$/i, '$1ies'],
		    [/([^ch][ieo][ln])ey$/i, '$1ies'],
		    [/(x|ch|ss|sh|zz)$/i, '$1es'],
		    [/(matr|cod|mur|sil|vert|ind|append)(?:ix|ex)$/i, '$1ices'],
		    [/\b((?:tit)?m|l)(?:ice|ouse)$/i, '$1ice'],
		    [/(pe)(?:rson|ople)$/i, '$1ople'],
		    [/(child)(?:ren)?$/i, '$1ren'],
		    [/eaux$/i, '$0'],
		    [/m[ae]n$/i, 'men'],
		    ['thou', 'you']
		  ].forEach(function (rule) {
		    return pluralize.addPluralRule(rule[0], rule[1]);
		  });

		  /**
		   * Singularization rules.
		   */
		  [
		    [/s$/i, ''],
		    [/(ss)$/i, '$1'],
		    [/(wi|kni|(?:after|half|high|low|mid|non|night|[^\w]|^)li)ves$/i, '$1fe'],
		    [/(ar|(?:wo|[ae])l|[eo][ao])ves$/i, '$1f'],
		    [/ies$/i, 'y'],
		    [/\b([pl]|zomb|(?:neck|cross)?t|coll|faer|food|gen|goon|group|lass|talk|goal|cut)ies$/i, '$1ie'],
		    [/\b(mon|smil)ies$/i, '$1ey'],
		    [/\b((?:tit)?m|l)ice$/i, '$1ouse'],
		    [/(seraph|cherub)im$/i, '$1'],
		    [/(x|ch|ss|sh|zz|tto|go|cho|alias|[^aou]us|t[lm]as|gas|(?:her|at|gr)o|[aeiou]ris)(?:es)?$/i, '$1'],
		    [/(analy|diagno|parenthe|progno|synop|the|empha|cri|ne)(?:sis|ses)$/i, '$1sis'],
		    [/(movie|twelve|abuse|e[mn]u)s$/i, '$1'],
		    [/(test)(?:is|es)$/i, '$1is'],
		    [/(alumn|syllab|vir|radi|nucle|fung|cact|stimul|termin|bacill|foc|uter|loc|strat)(?:us|i)$/i, '$1us'],
		    [/(agend|addend|millenni|dat|extrem|bacteri|desiderat|strat|candelabr|errat|ov|symposi|curricul|quor)a$/i, '$1um'],
		    [/(apheli|hyperbat|periheli|asyndet|noumen|phenomen|criteri|organ|prolegomen|hedr|automat)a$/i, '$1on'],
		    [/(alumn|alg|vertebr)ae$/i, '$1a'],
		    [/(cod|mur|sil|vert|ind)ices$/i, '$1ex'],
		    [/(matr|append)ices$/i, '$1ix'],
		    [/(pe)(rson|ople)$/i, '$1rson'],
		    [/(child)ren$/i, '$1'],
		    [/(eau)x?$/i, '$1'],
		    [/men$/i, 'man']
		  ].forEach(function (rule) {
		    return pluralize.addSingularRule(rule[0], rule[1]);
		  });

		  /**
		   * Uncountable rules.
		   */
		  [
		    // Singular words with no plurals.
		    'adulthood',
		    'advice',
		    'agenda',
		    'aid',
		    'aircraft',
		    'alcohol',
		    'ammo',
		    'analytics',
		    'anime',
		    'athletics',
		    'audio',
		    'bison',
		    'blood',
		    'bream',
		    'buffalo',
		    'butter',
		    'carp',
		    'cash',
		    'chassis',
		    'chess',
		    'clothing',
		    'cod',
		    'commerce',
		    'cooperation',
		    'corps',
		    'debris',
		    'diabetes',
		    'digestion',
		    'elk',
		    'energy',
		    'equipment',
		    'excretion',
		    'expertise',
		    'firmware',
		    'flounder',
		    'fun',
		    'gallows',
		    'garbage',
		    'graffiti',
		    'hardware',
		    'headquarters',
		    'health',
		    'herpes',
		    'highjinks',
		    'homework',
		    'housework',
		    'information',
		    'jeans',
		    'justice',
		    'kudos',
		    'labour',
		    'literature',
		    'machinery',
		    'mackerel',
		    'mail',
		    'media',
		    'mews',
		    'moose',
		    'music',
		    'mud',
		    'manga',
		    'news',
		    'only',
		    'personnel',
		    'pike',
		    'plankton',
		    'pliers',
		    'police',
		    'pollution',
		    'premises',
		    'rain',
		    'research',
		    'rice',
		    'salmon',
		    'scissors',
		    'series',
		    'sewage',
		    'shambles',
		    'shrimp',
		    'software',
		    'species',
		    'staff',
		    'swine',
		    'tennis',
		    'traffic',
		    'transportation',
		    'trout',
		    'tuna',
		    'wealth',
		    'welfare',
		    'whiting',
		    'wildebeest',
		    'wildlife',
		    'you',
		    /pok[eé]mon$/i,
		    // Regexes.
		    /[^aeiou]ese$/i, // "chinese", "japanese"
		    /deer$/i, // "deer", "reindeer"
		    /fish$/i, // "fish", "blowfish", "angelfish"
		    /measles$/i,
		    /o[iu]s$/i, // "carnivorous"
		    /pox$/i, // "chickpox", "smallpox"
		    /sheep$/i
		  ].forEach(pluralize.addUncountableRule);

		  return pluralize;
		}); 
	} (pluralize$2));
	return pluralize$2.exports;
}

var pluralizeExports = requirePluralize();
var pluralize = /*@__PURE__*/getDefaultExportFromCjs(pluralizeExports);

/**
 * Return the default api version for a versioned service. Will return undefined if one does not exist
 * @param program
 * @param serviceNamespace
 * @returns
 */
function getDefaultApiVersion(context, serviceNamespace) {
    try {
        const versions = getVersions(context.program, serviceNamespace)[1].getVersions();
        removeVersionsLargerThanExplicitlySpecified(context, versions);
        // follow versioning principals of the versioning library and return last in list
        return versions[versions.length - 1];
    }
    catch (e) {
        return undefined;
    }
}
function isModelProperty(type) {
    return type && typeof type === "object" && "kind" in type && type.kind === "ModelProperty";
}
/**
 * Return whether a parameter is the Api Version parameter of a client
 * @param program
 * @param parameter
 * @returns
 */
function isApiVersion(context, type) {
    if (isModelProperty(type)) {
        const override = getIsApiVersion(context, type);
        if (override !== undefined) {
            return override;
        }
    }
    return (type.name.toLowerCase().includes("apiversion") ||
        type.name.toLowerCase().includes("api-version"));
}
/**
 * Get the client's namespace for generation. If package-name is passed in config, we return
 * that value as our namespace. Otherwise, we default to the TypeSpec service namespace.
 * @param program
 * @param context
 * @returns
 */
function getClientNamespaceString(context) {
    return getClientNamespaceStringHelper(context, listServices(context.program)[0]?.type);
}
/**
 * If the given type is an anonymous model and all of its properties excluding
 * header/query/path/status-code are sourced from a named model, returns that original named model.
 * Otherwise the given type is returned unchanged.
 *
 * @param context
 * @param type
 * @returns
 */
function getEffectivePayloadType(context, type) {
    const program = context.program;
    // if a type has name, we should resolve the name
    // this logic is for template cases, for e.g.,
    // model Catalog is TrackedResource<CatalogProperties>{}
    // model Deployment is TrackedResource<DeploymentProperties>{}
    // when pass them to getEffectiveModelType, we will get two different types
    // with the same name "TrackedResource" which will loose original infomation
    if (type.name) {
        return type;
    }
    const effective = getEffectiveModelType(program, type, (t) => !isMetadata(context.program, t) && !getVisibility(context.program, t)?.includes("none"));
    if (effective.name) {
        return effective;
    }
    return type;
}
/**
 * Get the library and wire name of a model property. Takes `@clientName` and `@encodedName` into account
 * @param context
 * @param property
 * @returns a tuple of the library and wire name for a model property
 */
function getPropertyNames(context, property) {
    return [getLibraryName(context, property), getWireName(context, property)];
}
/**
 * Get the library name of a property / parameter / operation / model / enum. Takes projections into account
 *
 * Returns name in the following order of priority
 * 1. language emitter name, i.e. @clientName("csharpSpecificName", "csharp") => "csharpSpecificName"
 * 2. client name, i.e. @clientName(""clientName") => "clientName"
 * 3. deprecated projected name
 * 4. friendly name, i.e. @friendlyName("friendlyName") => "friendlyName"
 * 5. name in typespec
 *
 * @param context
 * @param type
 * @returns the library name for a typespec type
 */
function getLibraryName(context, type) {
    // 1. check if there's a client name
    let emitterSpecificName = getClientNameOverride(context, type);
    if (emitterSpecificName && emitterSpecificName !== type.name)
        return emitterSpecificName;
    // 2. check if there's a specific name for our language with deprecated @projectedName
    emitterSpecificName = getProjectedName(context.program, type, context.emitterName);
    if (emitterSpecificName && emitterSpecificName !== type.name)
        return emitterSpecificName;
    // 3. check if there's a client name with deprecated @projectedName
    const clientSpecificName = getProjectedName(context.program, type, "client");
    if (clientSpecificName && emitterSpecificName !== type.name)
        return clientSpecificName;
    // 4. check if there's a friendly name, if so return friendly name
    const friendlyName = getFriendlyName(context.program, type);
    if (friendlyName)
        return friendlyName;
    // 5. if type is derived from template and name is the same as template, add template parameters' name as suffix
    if (typeof type.name === "string" &&
        type.name !== "" &&
        type.kind === "Model" &&
        type.templateMapper?.args) {
        return (type.name +
            type.templateMapper.args
                .filter((arg) => "kind" in arg &&
                (arg.kind === "Model" || arg.kind === "Enum" || arg.kind === "Union") &&
                arg.name !== undefined &&
                arg.name.length > 0)
                .map((arg) => pascalCase(arg.name))
                .join(""));
    }
    return typeof type.name === "string" ? type.name : "";
}
/**
 * Get the serialized name of a type.
 * @param context
 * @param type
 * @returns
 */
function getWireName(context, type) {
    // 1. Check if there's an encoded name
    const encodedName = resolveEncodedName(context.program, type, "application/json");
    if (encodedName !== type.name)
        return encodedName;
    // 2. Check if there's deprecated language projection
    return getProjectedName(context.program, type, "json") ?? type.name;
}
/**
 * Helper function to return cross language definition id for a type
 * @param type
 * @returns
 */
function getCrossLanguageDefinitionId(context, type, operation, appendNamespace = true) {
    let retval = type.name || "anonymous";
    let namespace = type.kind === "ModelProperty" ? type.model?.namespace : type.namespace;
    switch (type.kind) {
        // Enum and Scalar will always have a name
        case "Union":
        case "Model":
            if (type.name) {
                break;
            }
            const contextPath = operation
                ? getContextPath(context, operation, type)
                : findContextPath(context, type);
            const namingPart = contextPath.slice(findLastNonAnonymousNode(contextPath));
            if (namingPart[0]?.type?.kind === "Model" ||
                namingPart[0]?.type?.kind === "Union" ||
                namingPart[0]?.type?.kind === "Operation") {
                namespace = namingPart[0]?.type?.namespace;
            }
            retval =
                namingPart
                    .map((x) => x.type?.kind === "Model" || x.type?.kind === "Union"
                    ? x.type.name || x.name
                    : x.name || "anonymous")
                    .join(".") +
                    "." +
                    retval;
            break;
        case "ModelProperty":
            if (type.model) {
                // operation parameter case
                if (type.model === operation?.parameters) {
                    retval = `${getCrossLanguageDefinitionId(context, operation, undefined, false)}.${retval}`;
                }
                else {
                    retval = `${getCrossLanguageDefinitionId(context, type.model, operation, false)}.${retval}`;
                }
            }
            break;
        case "Operation":
            if (type.interface) {
                retval = `${getCrossLanguageDefinitionId(context, type.interface, undefined, false)}.${retval}`;
            }
            break;
    }
    if (appendNamespace && namespace && getNamespaceFullName(namespace)) {
        retval = `${getNamespaceFullName(namespace)}.${retval}`;
    }
    return retval;
}
/**
 * Helper function return the cross langauge package id for a package
 */
function getCrossLanguagePackageId(context) {
    const diagnostics = createDiagnosticCollector();
    const services = listServices(context.program);
    if (services.length === 0)
        return diagnostics.wrap("");
    const serviceNamespace = getNamespaceFullName(services[0].type);
    if (services.length > 1) {
        diagnostics.add(createDiagnostic({
            code: "multiple-services",
            target: services[0].type,
            format: {
                service: serviceNamespace,
            },
        }));
    }
    return diagnostics.wrap(serviceNamespace);
}
/**
 * Create a name for anonymous model
 * @param context
 * @param type
 */
function getGeneratedName(context, type, operation) {
    if (!context.generatedNames) {
        context.generatedNames = new Map();
    }
    const generatedName = context.generatedNames.get(type);
    if (generatedName)
        return generatedName;
    const contextPath = operation
        ? getContextPath(context, operation, type)
        : findContextPath(context, type);
    const createdName = buildNameFromContextPaths(context, type, contextPath);
    return createdName;
}
/**
 * Traverse each operation and model to find one possible context path for the given type.
 * @param context
 * @param type
 * @returns
 */
function findContextPath(context, type) {
    for (const client of listClients(context)) {
        // orphan models
        if (client.service) {
            for (const model of client.service.models.values()) {
                if ([...model.properties.values()].filter((p) => !isMetadata(context.program, p)).length === 0)
                    continue;
                const result = getContextPath(context, model, type);
                if (result.length > 0) {
                    return result;
                }
            }
        }
        for (const operation of listOperationsInOperationGroup(context, client)) {
            const result = getContextPath(context, operation, type);
            if (result.length > 0) {
                return result;
            }
        }
        const ogs = listOperationGroups(context, client);
        while (ogs.length) {
            const operationGroup = ogs.pop();
            for (const operation of listOperationsInOperationGroup(context, operationGroup)) {
                const result = getContextPath(context, operation, type);
                if (result.length > 0) {
                    return result;
                }
            }
            if (operationGroup?.subOperationGroups) {
                ogs.push(...operationGroup.subOperationGroups);
            }
        }
    }
    return [];
}
/**
 * Find one possible context path for the given type in the given operation or model.
 * @param context
 * @param root
 * @param typeToFind
 * @returns
 */
function getContextPath(context, root, typeToFind) {
    // use visited set to avoid cycle model reference
    const visited = new Set();
    let result;
    if (root.kind === "Operation") {
        const httpOperation = getHttpOperationWithCache(context, root);
        if (httpOperation.parameters.body) {
            visited.clear();
            result = [{ name: root.name, type: root }];
            let bodyType;
            if (isHttpBodySpread(httpOperation.parameters.body)) {
                bodyType = getHttpBodySpreadModel(httpOperation.parameters.body.type);
            }
            else {
                bodyType = httpOperation.parameters.body.type;
            }
            if (dfsModelProperties(typeToFind, bodyType, "Request")) {
                return result;
            }
        }
        for (const parameter of Object.values(httpOperation.parameters.parameters)) {
            visited.clear();
            result = [{ name: root.name, type: root }];
            if (dfsModelProperties(typeToFind, parameter.param.type, `Request${pascalCase(parameter.name)}`)) {
                return result;
            }
        }
        for (const response of httpOperation.responses) {
            for (const innerResponse of response.responses) {
                if (innerResponse.body?.type) {
                    visited.clear();
                    result = [{ name: root.name, type: root }];
                    if (dfsModelProperties(typeToFind, innerResponse.body.type, "Response", true)) {
                        return result;
                    }
                }
                const headers = getHttpOperationResponseHeaders(innerResponse);
                if (headers) {
                    for (const header of Object.values(headers)) {
                        visited.clear();
                        result = [{ name: root.name, type: root }];
                        if (dfsModelProperties(typeToFind, header.type, `Response${pascalCase(header.name)}`)) {
                            return result;
                        }
                    }
                }
            }
        }
    }
    else {
        visited.clear();
        result = [];
        if (dfsModelProperties(typeToFind, root, root.name)) {
            return result;
        }
    }
    return [];
    /**
     * Traverse each node, if it is not model or union, no need to traverse anymore.
     * If it is the expected type just return.
     * If it is array or dict, traverse the array/dict item node. e.g. {name: string}[] case.
     * If it is model, add the current node to the path and traverse each property node.
     * If it is model, traverse the base and derived model node if existed.
     * @param expectedType
     * @param currentType
     * @param displayName
     * @param currentContextPath
     * @param contextPaths
     * @param visited
     * @returns
     */
    function dfsModelProperties(expectedType, currentType, displayName, needFindEffectiveType = false) {
        if (currentType == null || visited.has(currentType)) {
            // cycle reference detected
            return false;
        }
        if (!["Model", "Union", "String", "Number", "Boolean"].includes(currentType.kind)) {
            return false;
        }
        visited.add(currentType);
        if (currentType === expectedType) {
            result.push({ name: displayName, type: currentType });
            return true;
        }
        else if (currentType.kind === "Model") {
            // Peel off HttpPart<MyRealType> to get "MyRealType"
            const typeWrappedByHttpPart = getHttpPart(context.program, currentType);
            if (typeWrappedByHttpPart) {
                return dfsModelProperties(expectedType, typeWrappedByHttpPart.type, displayName);
            }
            if (currentType.indexer &&
                currentType.properties.size === 0 &&
                ((currentType.indexer.key.name === "string" && currentType.name === "Record") ||
                    currentType.indexer.key.name === "integer")) {
                // handle array or dict
                const dictOrArrayItemType = currentType.indexer.value;
                return dfsModelProperties(expectedType, dictOrArrayItemType, pluralize.singular(displayName));
            }
            // handle model
            if (needFindEffectiveType) {
                currentType = getEffectiveModelType(context.program, currentType);
            }
            result.push({ name: displayName, type: currentType });
            for (const property of currentType.properties.values()) {
                // traverse model property
                // use property.name as displayName
                const result = dfsModelProperties(expectedType, property.type, property.name);
                if (result)
                    return true;
            }
            // handle additional properties type: model MyModel is Record<> {}
            if (currentType.sourceModel?.kind === "Model" && currentType.sourceModel?.name === "Record") {
                const result = dfsModelProperties(expectedType, currentType.sourceModel.indexer.value, "AdditionalProperty");
                if (result)
                    return true;
            }
            // handle additional properties type: model MyModel { ...Record<>}
            if (currentType.indexer) {
                const result = dfsModelProperties(expectedType, currentType.indexer.value, "AdditionalProperty");
                if (result)
                    return true;
            }
            // handle additional properties type: model MyModel extends Record<> {}
            if (currentType.baseModel) {
                if (currentType.baseModel.name === "Record") {
                    const result = dfsModelProperties(expectedType, currentType.baseModel.indexer.value, "AdditionalProperty");
                    if (result)
                        return true;
                }
            }
            result.pop();
            if (currentType.baseModel) {
                const result = dfsModelProperties(expectedType, currentType.baseModel, currentType.baseModel.name);
                if (result)
                    return true;
            }
            for (const derivedModel of currentType.derivedModels) {
                const result = dfsModelProperties(expectedType, derivedModel, derivedModel.name);
                if (result)
                    return true;
            }
            return false;
        }
        else if (currentType.kind === "Union") {
            // handle union
            for (const unionType of currentType.variants.values()) {
                // traverse union type
                // use unionType.name as displayName
                const result = dfsModelProperties(expectedType, unionType.type, displayName);
                if (result)
                    return true;
            }
            return false;
        }
        else {
            return false;
        }
    }
}
function findLastNonAnonymousNode(contextPath) {
    let lastNonAnonymousModelNodeIndex = contextPath.length - 1;
    while (lastNonAnonymousModelNodeIndex >= 0) {
        const currType = contextPath[lastNonAnonymousModelNodeIndex].type;
        if ((currType.kind === "Model" || currType.kind === "Union" || currType.kind === "Operation") &&
            currType.name) {
            // it's non anonymous node
            break;
        }
        else {
            --lastNonAnonymousModelNodeIndex;
        }
    }
    return lastNonAnonymousModelNodeIndex;
}
/**
 * The logic is basically three steps:
 * 1. find the last nonanonymous model node, this node can be operation node or model node which is not anonymous
 * 2. build the name from the last nonanonymous model node to the end of the path
 * 3. simplely handle duplication with adding number suffix
 * @param contextPaths
 * @returns
 */
function buildNameFromContextPaths(context, type, contextPath) {
    // fallback to empty name for corner case
    if (contextPath.length === 0) {
        return "";
    }
    // 1. find the last non-anonymous model node
    const lastNonAnonymousNodeIndex = findLastNonAnonymousNode(contextPath);
    // 2. build name
    let createName = "";
    for (let j = lastNonAnonymousNodeIndex; j < contextPath.length; j++) {
        const currContextPathType = contextPath[j]?.type;
        if (currContextPathType?.kind === "String" ||
            currContextPathType?.kind === "Number" ||
            currContextPathType?.kind === "Boolean") {
            // constant type
            createName = `${createName}${pascalCase(contextPath[j].name)}`;
        }
        else if (!currContextPathType?.name || currContextPathType.kind === "Operation") {
            // is anonymous node or operation node
            createName = `${createName}${pascalCase(contextPath[j].name)}`;
        }
        else {
            // is non-anonymous node, use type name
            createName = `${createName}${currContextPathType.name}`;
        }
    }
    // 3. simplely handle duplication
    let duplicateCount = 1;
    const rawCreateName = createName;
    const generatedNames = [...(context.generatedNames?.values() ?? [])];
    while (generatedNames.includes(createName)) {
        createName = `${rawCreateName}${duplicateCount++}`;
    }
    if (context.generatedNames) {
        context.generatedNames.set(type, createName);
    }
    else {
        context.generatedNames = new Map([[type, createName]]);
    }
    return createName;
}
function getHttpOperationWithCache(context, operation) {
    if (context.httpOperationCache === undefined) {
        context.httpOperationCache = new Map();
    }
    if (context.httpOperationCache.has(operation)) {
        return context.httpOperationCache.get(operation);
    }
    const httpOperation = ignoreDiagnostics(getHttpOperation(context.program, operation));
    context.httpOperationCache.set(operation, httpOperation);
    return httpOperation;
}
/**
 * Get the examples for a given http operation.
 */
function getHttpOperationExamples(context, operation) {
    return context.__httpOperationExamples?.get(operation) ?? [];
}
/**
 * Get all the sub clients from current client.
 *
 * @param client
 * @param listNestedClients determine if nested clients should be listed
 * @returns
 */
function listSubClients(client, listNestedClients = false) {
    const subClients = client.methods
        .filter((c) => c.kind === "clientaccessor")
        .map((c) => c.response);
    if (listNestedClients) {
        for (const subClient of [...subClients]) {
            subClients.push(...listSubClients(subClient, listNestedClients));
        }
    }
    return subClients;
}
function isAzureCoreModel(t) {
    return t.__raw !== undefined && isAzureCoreTspModel(t.__raw);
}
/**
 * Judge whether a type is a paged result model.
 *
 * @param context TCGC context
 * @param t Any TCGC types
 * @returns
 */
function isPagedResultModel(context, t) {
    return context.__pagedResultSet.has(t);
}
/**
 * Find corresponding http parameter list for a client initialization parameter, a service method parameter or a property of a service method parameter.
 *
 * @param method
 * @param param
 * @returns
 */
function getHttpOperationParameter(method, param) {
    const operation = method.operation;
    // BFS to find the corresponding http parameter.
    // An http parameter will be mapped to a method/client parameter, several method/client parameters (body spread case), or one property of a method property (metadata on property case).
    // So, when we try to find which http parameter a parameter or property corresponds to, we compare the `correspondingMethodParams` list directly.
    // If a method parameter is spread case, then we need to find the cooresponding http body parameter's property.
    for (const p of operation.parameters) {
        for (const cp of p.correspondingMethodParams) {
            if (cp === param) {
                return p;
            }
        }
    }
    if (operation.bodyParam) {
        for (const cp of operation.bodyParam.correspondingMethodParams) {
            if (cp === param) {
                if (operation.bodyParam.type.kind === "model" && operation.bodyParam.type !== param.type) {
                    return operation.bodyParam.type.properties.find((p) => p.kind === "property" && p.name === param.name);
                }
                return operation.bodyParam;
            }
        }
    }
    return undefined;
}

function getSdkHttpOperation(context, httpOperation, methodParameters) {
    const diagnostics = createDiagnosticCollector();
    const { responses, exceptions } = diagnostics.pipe(getSdkHttpResponseAndExceptions(context, httpOperation));
    const responsesWithBodies = [...responses.values(), ...exceptions.values()].filter((r) => r.type);
    const parameters = diagnostics.pipe(getSdkHttpParameters(context, httpOperation, methodParameters, responsesWithBodies[0]));
    filterOutUselessPathParameters(context, httpOperation, methodParameters);
    return diagnostics.wrap({
        __raw: httpOperation,
        kind: "http",
        path: httpOperation.path,
        uriTemplate: httpOperation.uriTemplate,
        verb: httpOperation.verb,
        ...parameters,
        responses,
        exceptions,
    });
}
function isSdkHttpParameter(context, type) {
    const program = context.program;
    return (isPathParam(program, type) ||
        isQueryParam(program, type) ||
        isHeader(program, type) ||
        isBody(program, type) ||
        isCookieParam(program, type));
}
function getSdkHttpParameters(context, httpOperation, methodParameters, responseBody) {
    const diagnostics = createDiagnosticCollector();
    const retval = {
        parameters: [],
        bodyParam: undefined,
    };
    retval.parameters = httpOperation.parameters.parameters
        .filter((x) => !isNeverOrVoidType(x.param.type))
        .map((x) => diagnostics.pipe(getSdkHttpParameter(context, x.param, httpOperation.operation, x, x.type)));
    const headerParams = retval.parameters.filter((x) => x.kind === "header");
    // add operation info onto body param
    const tspBody = httpOperation.parameters.body;
    // we add correspondingMethodParams after we create the type, since we need the info on the type
    const correspondingMethodParams = [];
    if (tspBody) {
        // explicit @body and @bodyRoot
        if (tspBody.property && !isNeverOrVoidType(tspBody.property.type)) {
            const bodyParam = diagnostics.pipe(getSdkHttpParameter(context, tspBody.property, httpOperation.operation, undefined, "body"));
            if (bodyParam.kind !== "body") {
                diagnostics.add(createDiagnostic({
                    code: "unexpected-http-param-type",
                    target: tspBody.property,
                    format: {
                        paramName: tspBody.property.name,
                        expectedType: "body",
                        actualType: bodyParam.kind,
                    },
                }));
                return diagnostics.wrap(retval);
            }
            retval.bodyParam = bodyParam;
        }
        else if (!isNeverOrVoidType(tspBody.type)) {
            const spread = isHttpBodySpread(tspBody);
            let type;
            if (spread) {
                type = diagnostics.pipe(getClientTypeWithDiagnostics(context, getHttpBodySpreadModel(tspBody.type), httpOperation.operation));
            }
            else {
                type = diagnostics.pipe(getClientTypeWithDiagnostics(context, tspBody.type, httpOperation.operation));
            }
            const name = camelCase(type.name ?? "body");
            retval.bodyParam = {
                kind: "body",
                name,
                isGeneratedName: true,
                serializedName: "body",
                doc: getDoc(context.program, tspBody.type),
                summary: getSummary(context.program, tspBody.type),
                onClient: false,
                contentTypes: [],
                defaultContentType: "application/json", // actual content type info is added later
                isApiVersionParam: false,
                apiVersions: getAvailableApiVersions(context, tspBody.type, httpOperation.operation),
                type,
                optional: false,
                correspondingMethodParams,
                crossLanguageDefinitionId: `${getCrossLanguageDefinitionId(context, httpOperation.operation)}.body`,
                decorators: diagnostics.pipe(getTypeDecorators(context, tspBody.type)),
            };
        }
        if (retval.bodyParam) {
            addContentTypeInfoToBodyParam(context, httpOperation, retval.bodyParam);
            retval.bodyParam.correspondingMethodParams = diagnostics.pipe(getCorrespondingMethodParams(context, httpOperation.operation, methodParameters, retval.bodyParam));
        }
    }
    if (retval.bodyParam && !headerParams.some((h) => isContentTypeHeader(h))) {
        // if we have a body param and no content type header, we add one
        const contentTypeBase = {
            ...createContentTypeOrAcceptHeader(context, httpOperation, retval.bodyParam),
            doc: `Body parameter's content type. Known values are ${retval.bodyParam.contentTypes}`,
        };
        if (!methodParameters.some((m) => m.name === "contentType")) {
            methodParameters.push({
                ...contentTypeBase,
                kind: "method",
            });
        }
        retval.parameters.push({
            ...contentTypeBase,
            kind: "header",
            serializedName: "Content-Type",
            correspondingMethodParams,
        });
    }
    if (responseBody && !headerParams.some((h) => isAcceptHeader(h))) {
        // If our operation returns a body, we add an accept header if none exist
        const acceptBase = {
            ...createContentTypeOrAcceptHeader(context, httpOperation, responseBody),
        };
        if (!methodParameters.some((m) => m.name === "accept")) {
            methodParameters.push({
                ...acceptBase,
                kind: "method",
            });
        }
        retval.parameters.push({
            ...acceptBase,
            kind: "header",
            serializedName: "Accept",
            correspondingMethodParams,
        });
    }
    for (const param of retval.parameters) {
        param.correspondingMethodParams = diagnostics.pipe(getCorrespondingMethodParams(context, httpOperation.operation, methodParameters, param));
    }
    return diagnostics.wrap(retval);
}
function createContentTypeOrAcceptHeader(context, httpOperation, bodyObject) {
    const name = bodyObject.kind === "body" ? "contentType" : "accept";
    let type = getTypeSpecBuiltInType(context, "string");
    // for contentType, we treat it as a constant IFF there's one value and it's application/json.
    // this is to prevent a breaking change when a service adds more content types in the future.
    // e.g. the service accepting image/png then later image/jpeg should _not_ be a breaking change.
    //
    // for accept, we treat it as a constant IFF there's a single value. adding more content types
    // for this case is considered a breaking change for SDKs so we want to surface it as such.
    // e.g. the service returns image/png then later provides the option to return image/jpeg.
    if (bodyObject.contentTypes &&
        bodyObject.contentTypes.length === 1 &&
        (/json/.test(bodyObject.contentTypes[0]) || name === "accept")) {
        // in this case, we just want a content type of application/json
        type = {
            kind: "constant",
            value: bodyObject.contentTypes[0],
            valueType: type,
            name: `${httpOperation.operation.name}ContentType`,
            isGeneratedName: true,
            decorators: [],
        };
    }
    const optional = bodyObject.kind === "body" ? bodyObject.optional : false;
    // No need for clientDefaultValue because it's a constant, it only has one value
    return {
        type,
        name,
        isGeneratedName: true,
        apiVersions: bodyObject.apiVersions,
        isApiVersionParam: false,
        onClient: false,
        optional: optional,
        crossLanguageDefinitionId: `${getCrossLanguageDefinitionId(context, httpOperation.operation)}.${name}`,
        decorators: [],
    };
}
function addContentTypeInfoToBodyParam(context, httpOperation, bodyParam) {
    const diagnostics = createDiagnosticCollector();
    const tspBody = httpOperation.parameters.body;
    if (!tspBody)
        return diagnostics.diagnostics;
    let contentTypes = tspBody.contentTypes;
    if (contentTypes.length === 0) {
        contentTypes = ["application/json"];
    }
    const defaultContentType = contentTypes.includes("application/json")
        ? "application/json"
        : contentTypes[0];
    bodyParam.contentTypes = contentTypes;
    bodyParam.defaultContentType = defaultContentType;
    diagnostics.pipe(addEncodeInfo(context, bodyParam.__raw, bodyParam.type, defaultContentType));
    return diagnostics.diagnostics;
}
/**
 * Generate TCGC Http parameter type, `httpParam` or `location` should be provided at least one
 * @param context
 * @param param TypeSpec param for the http parameter
 * @param operation
 * @param httpParam TypeSpec Http parameter type
 * @param location Location of the http parameter
 * @returns
 */
function getSdkHttpParameter(context, param, operation, httpParam, location) {
    const diagnostics = createDiagnosticCollector();
    const base = diagnostics.pipe(getSdkModelPropertyTypeBase(context, param, operation));
    const program = context.program;
    const correspondingMethodParams = []; // we set it later in the operation
    if (isPathParam(context.program, param) || location === "path") {
        // we don't url encode if the type can be assigned to url
        const urlEncode = !ignoreDiagnostics(program.checker.isTypeAssignableTo(param.type.projectionBase ?? param.type, program.checker.getStdType("url"), param.type));
        return diagnostics.wrap({
            ...base,
            kind: "path",
            urlEncode,
            explode: httpParam?.explode ?? false,
            style: httpParam?.style ?? "simple",
            allowReserved: httpParam?.allowReserved ?? false,
            serializedName: getPathParamName(program, param) ?? base.name,
            correspondingMethodParams,
            optional: false,
        });
    }
    if (isCookieParam(context.program, param) || location === "cookie") {
        return diagnostics.wrap({
            ...base,
            kind: "cookie",
            serializedName: getCookieParamOptions(program, param)?.name ?? base.name,
            correspondingMethodParams,
            optional: param.optional,
        });
    }
    if (isBody(context.program, param) || location === "body") {
        return diagnostics.wrap({
            ...base,
            kind: "body",
            serializedName: param.name === "" ? "body" : getWireName(context, param),
            contentTypes: ["application/json"], // will update when we get to the operation level
            defaultContentType: "application/json", // will update when we get to the operation level
            optional: param.optional,
            correspondingMethodParams,
        });
    }
    const headerQueryBase = {
        ...base,
        optional: param.optional,
        collectionFormat: getCollectionFormat(context, param),
        correspondingMethodParams,
    };
    if (isQueryParam(context.program, param) || location === "query") {
        return diagnostics.wrap({
            ...headerQueryBase,
            kind: "query",
            serializedName: getQueryParamName(program, param) ?? base.name,
            explode: httpParam?.explode,
        });
    }
    if (!(isHeader(context.program, param) || location === "header")) {
        diagnostics.add(createDiagnostic({
            code: "unexpected-http-param-type",
            target: param,
            format: {
                paramName: param.name,
                expectedType: "path, query, header, or body",
                actualType: param.kind,
            },
        }));
    }
    return diagnostics.wrap({
        ...headerQueryBase,
        kind: "header",
        serializedName: getHeaderFieldName(program, param) ?? base.name,
    });
}
function getSdkHttpResponseAndExceptions(context, httpOperation) {
    const diagnostics = createDiagnosticCollector();
    const responses = [];
    const exceptions = [];
    for (const response of httpOperation.responses) {
        const headers = [];
        let body;
        let type;
        let contentTypes = [];
        for (const innerResponse of response.responses) {
            const defaultContentType = innerResponse.body?.contentTypes.includes("application/json")
                ? "application/json"
                : innerResponse.body?.contentTypes[0];
            for (const header of getHttpOperationResponseHeaders(innerResponse)) {
                if (isNeverOrVoidType(header.type))
                    continue;
                const clientType = diagnostics.pipe(getClientTypeWithDiagnostics(context, header.type));
                addEncodeInfo(context, header, clientType, defaultContentType);
                headers.push({
                    __raw: header,
                    doc: getDoc(context.program, header),
                    summary: getSummary(context.program, header),
                    serializedName: getHeaderFieldName(context.program, header),
                    type: clientType,
                });
            }
            if (innerResponse.body && !isNeverOrVoidType(innerResponse.body.type)) {
                if (body && body !== innerResponse.body.type) {
                    diagnostics.add(createDiagnostic({
                        code: "multiple-response-types",
                        target: innerResponse.body.type,
                        format: {
                            operation: httpOperation.operation.name,
                            response: innerResponse.body.type.kind === "Model"
                                ? innerResponse.body.type.name
                                : innerResponse.body.type.kind,
                        },
                    }));
                }
                contentTypes = contentTypes.concat(innerResponse.body.contentTypes);
                body =
                    innerResponse.body.type.kind === "Model"
                        ? getEffectivePayloadType(context, innerResponse.body.type)
                        : innerResponse.body.type;
                type = diagnostics.pipe(getClientTypeWithDiagnostics(context, body, httpOperation.operation));
                if (innerResponse.body.property) {
                    addEncodeInfo(context, innerResponse.body.property, type, defaultContentType);
                }
            }
        }
        const sdkResponse = {
            __raw: response,
            type,
            headers,
            contentTypes: contentTypes.length > 0 ? contentTypes : undefined,
            defaultContentType: contentTypes.includes("application/json")
                ? "application/json"
                : contentTypes[0],
            apiVersions: getAvailableApiVersions(context, httpOperation.operation, httpOperation.operation),
            description: response.description,
        };
        if (response.statusCodes === "*" ||
            isErrorModel(context.program, response.type) ||
            (body && isErrorModel(context.program, body))) {
            exceptions.push({
                ...sdkResponse,
                kind: "http",
                statusCodes: response.statusCodes,
            });
        }
        else {
            responses.push({
                ...sdkResponse,
                kind: "http",
                statusCodes: response.statusCodes,
            });
        }
    }
    return diagnostics.wrap({ responses, exceptions });
}
function getCorrespondingMethodParams(context, operation, methodParameters, serviceParam) {
    const diagnostics = createDiagnosticCollector();
    // 1. To see if the service parameter is a client parameter.
    const operationLocation = getLocationOfOperation(operation);
    let clientParams = context.__clientToParameters.get(operationLocation);
    if (!clientParams) {
        clientParams = [];
        context.__clientToParameters.set(operationLocation, clientParams);
    }
    const correspondingClientParams = clientParams.filter((x) => twoParamsEquivalent(context, x.__raw, serviceParam.__raw) ||
        (x.__raw?.kind === "ModelProperty" && getParamAlias(context, x.__raw) === serviceParam.name));
    if (correspondingClientParams.length > 0) {
        return diagnostics.wrap(correspondingClientParams);
    }
    // 2. To see if the service parameter is api version parameter that has been elevated to client.
    if (serviceParam.isApiVersionParam && serviceParam.onClient) {
        const existingApiVersion = clientParams?.find((x) => isApiVersion(context, x));
        if (!existingApiVersion) {
            diagnostics.add(createDiagnostic({
                code: "no-corresponding-method-param",
                target: serviceParam.__raw,
                format: {
                    paramName: "apiVersion",
                    methodName: operation.name,
                },
            }));
            return diagnostics.wrap([]);
        }
        return diagnostics.wrap(existingApiVersion ? [existingApiVersion] : []);
    }
    // 3. To see if the service parameter is subscription parameter that has been elevated to client (only for arm service).
    if (isSubscriptionId(context, serviceParam)) {
        const subId = clientParams.find((x) => isSubscriptionId(context, x));
        if (!subId) {
            diagnostics.add(createDiagnostic({
                code: "no-corresponding-method-param",
                target: serviceParam.__raw,
                format: {
                    paramName: "subscriptionId",
                    methodName: operation.name,
                },
            }));
            return diagnostics.wrap([]);
        }
        return diagnostics.wrap(subId ? [subId] : []);
    }
    // 4. To see if the service parameter is a method parameter or a property of a method parameter.
    const directMapping = findMapping(methodParameters, serviceParam);
    if (directMapping) {
        return diagnostics.wrap([directMapping]);
    }
    // 5. To see if all the property of the service parameter could be mapped to a method parameter or a property of a method parameter.
    if (serviceParam.kind === "body" && serviceParam.type.kind === "model") {
        const retVal = [];
        for (const serviceParamProp of serviceParam.type.properties) {
            const propertyMapping = findMapping(methodParameters, serviceParamProp);
            if (propertyMapping) {
                retVal.push(propertyMapping);
            }
        }
        if (retVal.length === serviceParam.type.properties.length) {
            return diagnostics.wrap(retVal);
        }
    }
    // If mapping could not be found,  TCGC will report error since we can't generate the client code without this mapping.
    diagnostics.add(createDiagnostic({
        code: "no-corresponding-method-param",
        target: serviceParam.__raw,
        format: {
            paramName: serviceParam.name,
            methodName: operation.name,
        },
    }));
    return diagnostics.wrap([]);
}
/**
 * Try to find the mapping of a service paramete or a property of a service parameter to a method parameter or a property of a method parameter.
 * @param methodParameters
 * @param serviceParam
 * @returns
 */
function findMapping(methodParameters, serviceParam) {
    const queue = [...methodParameters];
    const visited = new Set();
    while (queue.length > 0) {
        const methodParam = queue.shift();
        // HTTP operation parameter/body parameter/property of body parameter could either from an operation parameter directly or from a property of an operation parameter.
        if (methodParam.__raw &&
            serviceParam.__raw &&
            findRootSourceProperty(methodParam.__raw) === findRootSourceProperty(serviceParam.__raw)) {
            return methodParam;
        }
        // Two following two hard coded mapping is for the case that TCGC help to add content type and accept header when not exists.
        if (serviceParam.kind === "header" &&
            serviceParam.serializedName === "Content-Type" &&
            methodParam.name === "contentType") {
            return methodParam;
        }
        if (serviceParam.kind === "header" &&
            serviceParam.serializedName === "Accept" &&
            methodParam.name === "accept") {
            return methodParam;
        }
        // BFS to find the mapping.
        if (methodParam.type.kind === "model" && !visited.has(methodParam.type)) {
            visited.add(methodParam.type);
            let current = methodParam.type;
            while (current) {
                for (const prop of methodParam.type.properties) {
                    queue.push(prop);
                }
                current = current.baseModel;
            }
        }
    }
    return undefined;
}
function filterOutUselessPathParameters(context, httpOperation, methodParameters) {
    // there are some cases that method path parameter is not in operation:
    // 1. autoroute with constant parameter
    // 2. singleton arm resource name
    // 3. visibility mis-match
    // so we will remove the method parameter for consistent
    for (let i = 0; i < methodParameters.length; i++) {
        const param = methodParameters[i];
        if (param.__raw &&
            isPathParam(context.program, param.__raw) &&
            httpOperation.parameters.parameters.filter((p) => p.type === "path" &&
                p.name === (getPathParamName(context.program, param.__raw) ?? param.name)).length === 0) {
            methodParameters.splice(i, 1);
            i--;
        }
    }
}
function findRootSourceProperty(property) {
    while (property.sourceProperty) {
        property = property.sourceProperty;
    }
    return property;
}
function getCollectionFormat(context, type) {
    const program = context.program;
    /* eslint-disable @typescript-eslint/no-deprecated */
    const tspCollectionFormat = (isQueryParam(program, type)
        ? getQueryParamOptions(program, type)
        : isHeader(program, type)
            ? getHeaderFieldOptions(program, type)
            : undefined)?.format;
    return tspCollectionFormat;
}

function getTypeSpecBuiltInType(context, kind) {
    const global = context.program.getGlobalNamespaceType();
    const typeSpecNamespace = global.namespaces.get("TypeSpec");
    const type = typeSpecNamespace.scalars.get(kind);
    return getSdkBuiltInType(context, type);
}
function getUnknownType(context, type) {
    const diagnostics = createDiagnosticCollector();
    const unknownType = {
        ...diagnostics.pipe(getSdkTypeBaseHelper(context, type, "unknown")),
        name: getLibraryName(context, type),
        encode: getEncodeHelper(context, type),
        crossLanguageDefinitionId: "",
    };
    return diagnostics.wrap(unknownType);
}
function getEncodeHelper(context, type) {
    if (type.kind === "ModelProperty" || type.kind === "Scalar") {
        return getEncode(context.program, type)?.encoding;
    }
    return undefined;
}
/**
 * Add encoding info onto an sdk type. Since the encoding decorator
 * decorates the ModelProperty, we add the encoding info onto the property's internal
 * type.
 * @param context sdk context
 * @param type the original typespec type. Used to grab the encoding decorator off of
 * @param propertyType the type of the property, i.e. the internal type that we add the encoding info onto
 */
function addEncodeInfo(context, type, propertyType, defaultContentType) {
    const diagnostics = createDiagnosticCollector();
    const innerType = propertyType.kind === "nullable" ? propertyType.type : propertyType;
    const encodeData = getEncode(context.program, type);
    if (innerType.kind === "duration") {
        if (!encodeData)
            return diagnostics.wrap(undefined);
        innerType.encode = encodeData.encoding;
        innerType.wireType = diagnostics.pipe(getClientTypeWithDiagnostics(context, encodeData.type));
    }
    if (innerType.kind === "utcDateTime" || innerType.kind === "offsetDateTime") {
        if (encodeData) {
            innerType.encode = encodeData.encoding;
            innerType.wireType = diagnostics.pipe(getClientTypeWithDiagnostics(context, encodeData.type));
        }
        else if (type.kind === "ModelProperty" && isHeader(context.program, type)) {
            innerType.encode = "rfc7231";
        }
    }
    if (innerType.kind === "bytes") {
        if (encodeData) {
            innerType.encode = encodeData.encoding;
        }
        else if (!defaultContentType || defaultContentType === "application/json") {
            innerType.encode = "base64";
        }
        else {
            innerType.encode = "bytes";
        }
    }
    if (isSdkIntKind(innerType.kind)) {
        // only integer type is allowed to be encoded as string
        if (encodeData && "encode" in innerType) {
            if (encodeData?.encoding) {
                innerType.encode = encodeData.encoding;
            }
            if (encodeData?.type) {
                // if we specify the encoding type in the decorator, we set the `.encode` string
                // to the kind of the encoding type
                innerType.encode = getSdkBuiltInType(context, encodeData.type).kind;
            }
        }
    }
    return diagnostics.wrap(undefined);
}
/**
 * Mapping of typespec scalar kinds to the built in kinds exposed in the SDK
 * @param context the TCGC context
 * @param scalar the original typespec scalar
 * @returns the corresponding sdk built in kind
 */
function getScalarKind(context, scalar) {
    if (context.program.checker.isStdType(scalar)) {
        return scalar.name;
    }
    // for those scalar defined as `scalar newThing;`,
    // the best we could do here is return as a `any` type with a name and namespace and let the generator figure what this is
    if (scalar.baseScalar === undefined) {
        return "unknown";
    }
    return getScalarKind(context, scalar.baseScalar);
}
/**
 * This function converts a Scalar into SdkBuiltInType.
 * @param context
 * @param type
 * @param kind
 * @returns
 */
function getSdkBuiltInTypeWithDiagnostics(context, type, kind) {
    const diagnostics = createDiagnosticCollector();
    const stdType = {
        ...diagnostics.pipe(getSdkTypeBaseHelper(context, type, kind)),
        name: getLibraryName(context, type),
        encode: getEncodeHelper(context, type),
        doc: getDoc(context.program, type),
        summary: getSummary(context.program, type),
        baseType: type.baseScalar && !context.program.checker.isStdType(type) // we only calculate the base type when this type has a base type and this type is not a std type because for std types there is no point of calculating its base type.
            ? diagnostics.pipe(getSdkBuiltInTypeWithDiagnostics(context, type.baseScalar, kind))
            : undefined,
        crossLanguageDefinitionId: getCrossLanguageDefinitionId(context, type),
    };
    addEncodeInfo(context, type, stdType);
    return diagnostics.wrap(stdType);
}
/**
 * This function calculates the encode and wireType for a datetime or duration type.
 * We always first try to get the `@encode` decorator on this type and returns it if any.
 * If we did not get anything from the encode, we try to get the baseType's encode and wireType.
 * @param context
 * @param encodeData
 * @param baseType
 * @returns
 */
function getEncodeInfoForDateTimeOrDuration(context, encodeData, baseType) {
    const encode = encodeData?.encoding;
    const wireType = encodeData?.type
        ? getClientType(context, encodeData.type)
        : undefined;
    // if we get something from the encode
    if (encode || wireType) {
        return [encode, wireType];
    }
    // if we did not get anything from the encode, try the baseType
    return [baseType?.encode, baseType?.wireType];
}
/**
 * This function converts a Scalar into SdkDateTimeType.
 * @param context
 * @param type
 * @param kind
 * @returns
 */
function getSdkDateTimeType(context, type, kind) {
    const diagnostics = createDiagnosticCollector();
    const baseType = type.baseScalar && !context.program.checker.isStdType(type) // we only calculate the base type when this type has a base type and this type is not a std type because for std types there is no point of calculating its base type.
        ? diagnostics.pipe(getSdkDateTimeType(context, type.baseScalar, kind))
        : undefined;
    const [encode, wireType] = getEncodeInfoForDateTimeOrDuration(context, getEncode(context.program, type), baseType);
    return diagnostics.wrap({
        ...diagnostics.pipe(getSdkTypeBaseHelper(context, type, kind)),
        name: getLibraryName(context, type),
        encode: (encode ?? "rfc3339"),
        wireType: wireType ?? getTypeSpecBuiltInType(context, "string"),
        baseType: baseType,
        doc: getDoc(context.program, type),
        summary: getSummary(context.program, type),
        crossLanguageDefinitionId: getCrossLanguageDefinitionId(context, type),
    });
}
function getSdkDateTimeOrDurationOrBuiltInType(context, type) {
    // follow the extends hierarchy to determine the final kind of this type
    const kind = getScalarKind(context, type);
    if (kind === "utcDateTime" || kind === "offsetDateTime") {
        return getSdkDateTimeType(context, type, kind);
    }
    if (kind === "duration") {
        return getSdkDurationTypeWithDiagnostics(context, type, kind);
    }
    // handle the std types of typespec
    return getSdkBuiltInTypeWithDiagnostics(context, type, kind);
}
function getSdkTypeForLiteral(context, type) {
    let kind;
    if (type.kind === "String") {
        kind = "string";
    }
    else if (type.kind === "Boolean") {
        kind = "boolean";
    }
    else {
        kind = intOrFloat(type.value);
    }
    return getTypeSpecBuiltInType(context, kind);
}
function getSdkTypeForIntrinsic(context, type) {
    const kind = "unknown";
    const diagnostics = createDiagnosticCollector();
    return {
        ...diagnostics.pipe(getSdkTypeBaseHelper(context, type, kind)),
        name: getLibraryName(context, type),
        crossLanguageDefinitionId: "",
        encode: kind,
    };
}
function getSdkBuiltInType(context, type) {
    switch (type.kind) {
        case "Scalar":
            return ignoreDiagnostics(getSdkDateTimeOrDurationOrBuiltInType(context, type));
        case "Intrinsic":
            return getSdkTypeForIntrinsic(context, type);
        case "String":
        case "Number":
        case "Boolean":
            return getSdkTypeForLiteral(context, type);
    }
}
function getSdkDurationType(context, type) {
    return ignoreDiagnostics(getSdkDurationTypeWithDiagnostics(context, type, "duration"));
}
/**
 * This function converts a Scalar into SdkDurationType.
 * @param context
 * @param type
 * @param kind
 * @returns
 */
function getSdkDurationTypeWithDiagnostics(context, type, kind) {
    const diagnostics = createDiagnosticCollector();
    const baseType = type.baseScalar && !context.program.checker.isStdType(type) // we only calculate the base type when this type has a base type and this type is not a std type because for std types there is no point of calculating its base type.
        ? diagnostics.pipe(getSdkDurationTypeWithDiagnostics(context, type.baseScalar, kind))
        : undefined;
    const [encode, wireType] = getEncodeInfoForDateTimeOrDuration(context, getEncode(context.program, type), baseType);
    return diagnostics.wrap({
        ...diagnostics.pipe(getSdkTypeBaseHelper(context, type, kind)),
        name: getLibraryName(context, type),
        encode: (encode ?? "ISO8601"),
        wireType: wireType ?? getTypeSpecBuiltInType(context, "string"),
        baseType: baseType,
        doc: getDoc(context.program, type),
        summary: getSummary(context.program, type),
        crossLanguageDefinitionId: getCrossLanguageDefinitionId(context, type),
    });
}
function getSdkArrayOrDict(context, type, operation) {
    return ignoreDiagnostics(getSdkArrayOrDictWithDiagnostics(context, type, operation));
}
function getSdkArrayOrDictWithDiagnostics(context, type, operation) {
    const diagnostics = createDiagnosticCollector();
    // if model with both indexer and properties or name should be a model with additional properties
    if (type.indexer !== undefined && type.properties.size === 0) {
        if (!isNeverType(type.indexer.key)) {
            const valueType = diagnostics.pipe(getClientTypeWithDiagnostics(context, type.indexer.value, operation));
            const name = type.indexer.key.name;
            if (name === "string" && type.name === "Record") {
                // model MyModel is Record<> {} should be model with additional properties
                if (type.sourceModel?.kind === "Model" && type.sourceModel?.name === "Record") {
                    return diagnostics.wrap(undefined);
                }
                // other cases are dict
                return diagnostics.wrap({
                    ...diagnostics.pipe(getSdkTypeBaseHelper(context, type, "dict")),
                    keyType: diagnostics.pipe(getClientTypeWithDiagnostics(context, type.indexer.key, operation)),
                    valueType: valueType,
                });
            }
            else if (name === "integer") {
                // only array's index key name is integer
                return diagnostics.wrap({
                    ...diagnostics.pipe(getSdkTypeBaseHelper(context, type, "array")),
                    name: getLibraryName(context, type),
                    valueType: valueType,
                    crossLanguageDefinitionId: getCrossLanguageDefinitionId(context, type, operation),
                });
            }
        }
    }
    return diagnostics.wrap(undefined);
}
function getSdkTuple(context, type, operation) {
    return ignoreDiagnostics(getSdkTupleWithDiagnostics(context, type, operation));
}
function getSdkTupleWithDiagnostics(context, type, operation) {
    const diagnostics = createDiagnosticCollector();
    return diagnostics.wrap({
        ...diagnostics.pipe(getSdkTypeBaseHelper(context, type, "tuple")),
        valueTypes: type.values.map((x) => diagnostics.pipe(getClientTypeWithDiagnostics(context, x, operation))),
    });
}
function getSdkUnion(context, type, operation) {
    return ignoreDiagnostics(getSdkUnionWithDiagnostics(context, type, operation));
}
function getSdkUnionWithDiagnostics(context, type, operation) {
    let retval = context.referencedTypeMap?.get(type);
    const diagnostics = createDiagnosticCollector();
    if (!retval) {
        const nonNullOptions = getNonNullOptions(type);
        const nullOption = getNullOption(type);
        if (nonNullOptions.length === 0) {
            diagnostics.add(createDiagnostic({ code: "union-null", target: type }));
            return diagnostics.wrap(diagnostics.pipe(getUnknownType(context, type)));
        }
        // if a union is `type | null`, then we will return a nullable wrapper type of the type
        if (nonNullOptions.length === 1 && nullOption !== undefined) {
            retval = diagnostics.pipe(getClientTypeWithDiagnostics(context, nonNullOptions[0], operation));
        }
        else if (
        // judge if the union can be converted to enum
        // if language does not need flatten union as enum
        // filter the case that union is composed of union or enum
        context.flattenUnionAsEnum ||
            ![...type.variants.values()].some((variant) => {
                return variant.type.kind === "Union" || variant.type.kind === "Enum";
            })) {
            const unionAsEnum = diagnostics.pipe(getUnionAsEnum(type));
            if (unionAsEnum) {
                retval = diagnostics.pipe(getSdkUnionEnumWithDiagnostics(context, unionAsEnum, operation));
            }
        }
        // other cases
        if (retval === undefined) {
            retval = {
                ...diagnostics.pipe(getSdkTypeBaseHelper(context, type, "union")),
                name: getLibraryName(context, type) || getGeneratedName(context, type, operation),
                isGeneratedName: !type.name,
                clientNamespace: getClientNamespace(context, type),
                variantTypes: nonNullOptions.map((x) => diagnostics.pipe(getClientTypeWithDiagnostics(context, x, operation))),
                crossLanguageDefinitionId: getCrossLanguageDefinitionId(context, type, operation),
                access: "public",
                usage: UsageFlags.None,
            };
        }
        if (nullOption !== undefined) {
            retval = {
                ...diagnostics.pipe(getSdkTypeBaseHelper(context, type, "nullable")),
                type: retval,
                access: "public",
                usage: UsageFlags.None,
                clientNamespace: getClientNamespace(context, type),
            };
        }
        updateReferencedTypeMap(context, type, retval);
    }
    return diagnostics.wrap(retval);
}
function getSdkConstantWithDiagnostics(context, type, operation) {
    const diagnostics = createDiagnosticCollector();
    switch (type.kind) {
        case "Number":
        case "String":
        case "Boolean":
            const valueType = getSdkTypeForLiteral(context, type);
            return diagnostics.wrap({
                ...diagnostics.pipe(getSdkTypeBaseHelper(context, type, "constant")),
                value: type.value,
                valueType,
                name: getGeneratedName(context, type, operation),
                isGeneratedName: true,
            });
    }
}
function getSdkConstant(context, type, operation) {
    return ignoreDiagnostics(getSdkConstantWithDiagnostics(context, type, operation));
}
function addDiscriminatorToModelType(context, type, model) {
    const discriminator = getDiscriminator(context.program, type);
    const diagnostics = createDiagnosticCollector();
    if (discriminator) {
        let discriminatorType = undefined;
        for (let i = 0; i < model.properties.length; i++) {
            const property = model.properties[i];
            if (property.kind === "property" && property.__raw?.name === discriminator.propertyName) {
                discriminatorType = property.type;
            }
        }
        let discriminatorProperty;
        for (const childModel of type.derivedModels) {
            const childModelSdkType = diagnostics.pipe(getSdkModelWithDiagnostics(context, childModel));
            for (const property of childModelSdkType.properties) {
                if (property.kind === "property") {
                    if (property.__raw?.name === discriminator?.propertyName) {
                        if (property.type.kind !== "constant" && property.type.kind !== "enumvalue") {
                            diagnostics.add(createDiagnostic({
                                code: "discriminator-not-constant",
                                target: type,
                                format: { discriminator: property.name },
                            }));
                        }
                        else if (typeof property.type.value !== "string") {
                            diagnostics.add(createDiagnostic({
                                code: "discriminator-not-string",
                                target: type,
                                format: {
                                    discriminator: property.name,
                                    discriminatorValue: String(property.type.value),
                                },
                            }));
                        }
                        else {
                            // map string value type to enum value type
                            if (property.type.kind === "constant" && discriminatorType?.kind === "enum") {
                                for (const value of discriminatorType.values) {
                                    if (value.value === property.type.value) {
                                        property.type = value;
                                    }
                                }
                            }
                            childModelSdkType.discriminatorValue = property.type.value;
                            property.discriminator = true;
                            if (model.discriminatedSubtypes === undefined) {
                                model.discriminatedSubtypes = {};
                            }
                            model.discriminatedSubtypes[property.type.value] = childModelSdkType;
                            discriminatorProperty = property;
                        }
                    }
                }
            }
        }
        for (let i = 0; i < model.properties.length; i++) {
            const property = model.properties[i];
            if (property.kind === "property" && property.__raw?.name === discriminator.propertyName) {
                property.discriminator = true;
                model.discriminatorProperty = property;
                return diagnostics.wrap(undefined);
            }
        }
        if (discriminatorProperty) {
            if (discriminatorProperty.type.kind === "constant") {
                discriminatorType = { ...discriminatorProperty.type.valueType };
            }
            else if (discriminatorProperty.type.kind === "enumvalue") {
                discriminatorType = discriminatorProperty.type.enumType;
            }
        }
        else {
            discriminatorType = getTypeSpecBuiltInType(context, "string");
        }
        const name = discriminatorProperty ? discriminatorProperty.name : discriminator.propertyName;
        model.properties.splice(0, 0, {
            kind: "property",
            doc: `Discriminator property for ${model.name}.`,
            optional: false,
            discriminator: true,
            serializedName: discriminatorProperty
                ? discriminatorProperty.serializedName
                : discriminator.propertyName,
            type: discriminatorType,
            name,
            isGeneratedName: false,
            onClient: false,
            apiVersions: discriminatorProperty
                ? getAvailableApiVersions(context, discriminatorProperty.__raw, type)
                : getAvailableApiVersions(context, type, type),
            isApiVersionParam: false,
            isMultipartFileInput: false, // discriminator property cannot be a file
            flatten: false, // discriminator properties can not be flattened
            crossLanguageDefinitionId: `${model.crossLanguageDefinitionId}.${name}`,
            decorators: [],
        });
        model.discriminatorProperty = model.properties[0];
    }
    return diagnostics.wrap(undefined);
}
function getSdkModel(context, type, operation) {
    return ignoreDiagnostics(getSdkModelWithDiagnostics(context, type, operation));
}
function getInitializationType(context, type, operation) {
    const model = ignoreDiagnostics(getSdkModelWithDiagnostics(context, type, operation));
    for (const property of model.properties) {
        property.kind = "method";
    }
    return model;
}
function getSdkModelWithDiagnostics(context, type, operation) {
    const diagnostics = createDiagnosticCollector();
    let sdkType = context.referencedTypeMap?.get(type);
    if (!sdkType) {
        const name = getLibraryName(context, type) || getGeneratedName(context, type, operation);
        const usage = isErrorModel(context.program, type) ? UsageFlags.Error : UsageFlags.None; // eslint-disable-line @typescript-eslint/no-deprecated
        sdkType = {
            ...diagnostics.pipe(getSdkTypeBaseHelper(context, type, "model")),
            name: name,
            isGeneratedName: !type.name,
            clientNamespace: getClientNamespace(context, type),
            doc: getDoc(context.program, type),
            summary: getSummary(context.program, type),
            properties: [],
            additionalProperties: undefined, // going to set additional properties in the next few lines when we look at base model
            access: "public",
            usage,
            crossLanguageDefinitionId: getCrossLanguageDefinitionId(context, type, operation),
            apiVersions: getAvailableApiVersions(context, type, type.namespace),
        };
        updateReferencedTypeMap(context, type, sdkType);
        // model MyModel is Record<> {} should be model with additional properties
        if (type.sourceModel?.kind === "Model" && type.sourceModel?.name === "Record") {
            sdkType.additionalProperties = diagnostics.pipe(getClientTypeWithDiagnostics(context, type.sourceModel.indexer.value, operation));
        }
        // model MyModel { ...Record<>} should be model with additional properties
        if (type.indexer) {
            sdkType.additionalProperties = diagnostics.pipe(getClientTypeWithDiagnostics(context, type.indexer.value, operation));
        }
        // propreties should be generated first since base model'sdiscriminator handling is depend on derived model's properties
        diagnostics.pipe(addPropertiesToModelType(context, type, sdkType, operation));
        if (type.baseModel) {
            sdkType.baseModel = context.referencedTypeMap?.get(type.baseModel);
            if (sdkType.baseModel === undefined) {
                const baseModel = diagnostics.pipe(getClientTypeWithDiagnostics(context, type.baseModel, operation));
                if (baseModel.kind === "dict") {
                    // model MyModel extends Record<> {} should be model with additional properties
                    sdkType.additionalProperties = baseModel.valueType;
                }
                else {
                    sdkType.baseModel = baseModel;
                }
            }
        }
        diagnostics.pipe(addDiscriminatorToModelType(context, type, sdkType));
        updateReferencedTypeMap(context, type, sdkType);
    }
    return diagnostics.wrap(sdkType);
}
function getSdkEnumValueType(context, values) {
    const diagnostics = createDiagnosticCollector();
    let kind = "string";
    for (const value of values) {
        if (typeof value === "number") {
            kind = intOrFloat(value);
            if (kind === "float32") {
                break;
            }
        }
        else if (typeof value === "string") {
            kind = "string";
            break;
        }
    }
    return diagnostics.wrap(getTypeSpecBuiltInType(context, kind));
}
function getUnionAsEnumValueType(context, union) {
    const diagnostics = createDiagnosticCollector();
    const nonNullOptions = getNonNullOptions(union);
    for (const option of nonNullOptions) {
        if (option.kind === "Union") {
            const ret = diagnostics.pipe(getUnionAsEnumValueType(context, option));
            if (ret)
                return diagnostics.wrap(ret);
        }
        else if (option.kind === "Scalar") {
            const ret = diagnostics.pipe(getClientTypeWithDiagnostics(context, option));
            return diagnostics.wrap(ret);
        }
    }
    return diagnostics.wrap(undefined);
}
function getSdkEnumValue(context, enumType, type) {
    return ignoreDiagnostics(getSdkEnumValueWithDiagnostics(context, enumType, type));
}
function getSdkEnumValueWithDiagnostics(context, enumType, type) {
    const diagnostics = createDiagnosticCollector();
    return diagnostics.wrap({
        ...diagnostics.pipe(getSdkTypeBaseHelper(context, type, "enumvalue")),
        name: getLibraryName(context, type),
        value: type.value ?? type.name,
        doc: getDoc(context.program, type),
        summary: getSummary(context.program, type),
        enumType,
        valueType: enumType.valueType,
    });
}
function getSdkEnum(context, type, operation) {
    return ignoreDiagnostics(getSdkEnumWithDiagnostics(context, type, operation));
}
function getSdkEnumWithDiagnostics(context, type, operation) {
    const diagnostics = createDiagnosticCollector();
    let sdkType = context.referencedTypeMap?.get(type);
    if (!sdkType) {
        sdkType = {
            ...diagnostics.pipe(getSdkTypeBaseHelper(context, type, "enum")),
            name: getLibraryName(context, type),
            isGeneratedName: false,
            clientNamespace: getClientNamespace(context, type),
            doc: getDoc(context.program, type),
            summary: getSummary(context.program, type),
            valueType: diagnostics.pipe(getSdkEnumValueType(context, [...type.members.values()].map((v) => v.value))),
            values: [],
            isFixed: true, // enums are always fixed after we switch to use union to represent extensible enum
            isFlags: false,
            usage: UsageFlags.None, // We will add usage as we loop through the operations
            access: "public", // Dummy value until we update models map
            crossLanguageDefinitionId: getCrossLanguageDefinitionId(context, type, operation),
            apiVersions: getAvailableApiVersions(context, type, type.namespace),
            isUnionAsEnum: false,
        };
        for (const member of type.members.values()) {
            sdkType.values.push(diagnostics.pipe(getSdkEnumValueWithDiagnostics(context, sdkType, member)));
        }
    }
    updateReferencedTypeMap(context, type, sdkType);
    return diagnostics.wrap(sdkType);
}
function getSdkUnionEnumValues(context, type, enumType) {
    const diagnostics = createDiagnosticCollector();
    const values = [];
    for (const member of type.flattenedMembers.values()) {
        const name = getLibraryName(context, member.type);
        values.push({
            ...diagnostics.pipe(getSdkTypeBaseHelper(context, member.type, "enumvalue")),
            name: name ? name : `${member.value}`,
            doc: getDoc(context.program, member.type),
            summary: getSummary(context.program, member.type),
            value: member.value,
            valueType: enumType.valueType,
            enumType,
        });
    }
    return diagnostics.wrap(values);
}
function getSdkUnionEnum(context, type, operation) {
    return ignoreDiagnostics(getSdkUnionEnumWithDiagnostics(context, type, operation));
}
function getSdkUnionEnumWithDiagnostics(context, type, operation) {
    const diagnostics = createDiagnosticCollector();
    const union = type.union;
    let sdkType = context.referencedTypeMap?.get(union);
    if (!sdkType) {
        const name = getLibraryName(context, type.union) || getGeneratedName(context, union, operation);
        sdkType = {
            ...diagnostics.pipe(getSdkTypeBaseHelper(context, type.union, "enum")),
            name,
            isGeneratedName: !type.union.name,
            clientNamespace: getClientNamespace(context, type.union),
            doc: getDoc(context.program, union),
            summary: getSummary(context.program, union),
            valueType: diagnostics.pipe(getUnionAsEnumValueType(context, type.union)) ??
                diagnostics.pipe(getSdkEnumValueType(context, [...type.flattenedMembers.values()].map((v) => v.value))),
            values: [],
            isFixed: !type.open,
            isFlags: false,
            usage: UsageFlags.None, // We will add usage as we loop through the operations
            access: "public", // Dummy value until we update models map
            crossLanguageDefinitionId: getCrossLanguageDefinitionId(context, union, operation),
            apiVersions: getAvailableApiVersions(context, type.union, type.union.namespace),
            isUnionAsEnum: true,
        };
        sdkType.values = diagnostics.pipe(getSdkUnionEnumValues(context, type, sdkType));
    }
    return diagnostics.wrap(sdkType);
}
function getClientTypeWithDiagnostics(context, type, operation) {
    if (!context.knownScalars) {
        context.knownScalars = getKnownScalars();
    }
    const diagnostics = createDiagnosticCollector();
    let retval = undefined;
    switch (type.kind) {
        case "String":
        case "Number":
        case "Boolean":
            retval = diagnostics.pipe(getSdkConstantWithDiagnostics(context, type));
            break;
        case "Tuple":
            retval = diagnostics.pipe(getSdkTupleWithDiagnostics(context, type, operation));
            break;
        case "Model":
            retval = diagnostics.pipe(getSdkArrayOrDictWithDiagnostics(context, type, operation));
            if (retval === undefined) {
                const httpPart = getHttpPart(context.program, type);
                if (httpPart === undefined) {
                    retval = diagnostics.pipe(getSdkModelWithDiagnostics(context, type, operation));
                }
                else {
                    retval = diagnostics.pipe(getClientTypeWithDiagnostics(context, httpPart.type, operation));
                }
            }
            break;
        case "Intrinsic":
            retval = getSdkTypeForIntrinsic(context, type);
            break;
        case "Scalar":
            retval = diagnostics.pipe(getSdkDateTimeOrDurationOrBuiltInType(context, getAlternateType(context, type) ?? type));
            break;
        case "Enum":
            retval = diagnostics.pipe(getSdkEnumWithDiagnostics(context, type, operation));
            break;
        case "Union":
            retval = diagnostics.pipe(getSdkUnionWithDiagnostics(context, type, operation));
            break;
        case "ModelProperty":
            const alternateType = getAlternateType(context, type);
            retval = diagnostics.pipe(getClientTypeWithDiagnostics(context, alternateType ?? type.type, operation));
            diagnostics.pipe(addEncodeInfo(context, alternateType ?? type, retval));
            break;
        case "UnionVariant":
            const unionType = diagnostics.pipe(getClientTypeWithDiagnostics(context, type.union, operation));
            if (unionType.kind === "enum") {
                retval = unionType.values.find((x) => x.name === getLibraryName(context, type));
            }
            else {
                retval = diagnostics.pipe(getClientTypeWithDiagnostics(context, type.type, operation));
            }
            break;
        case "EnumMember":
            const enumType = diagnostics.pipe(getSdkEnumWithDiagnostics(context, type.enum, operation));
            retval = diagnostics.pipe(getSdkEnumValueWithDiagnostics(context, enumType, type));
            break;
        default:
            retval = diagnostics.pipe(getUnknownType(context, type));
            diagnostics.add(createDiagnostic({ code: "unsupported-kind", target: type, format: { kind: type.kind } }));
    }
    return diagnostics.wrap(retval);
}
function getClientType(context, type, operation) {
    return ignoreDiagnostics(getClientTypeWithDiagnostics(context, type, operation));
}
function isReadOnly(property) {
    if (property.visibility &&
        property.visibility.includes(Visibility.Read) &&
        property.visibility.length === 1) {
        return true;
    }
    return false;
}
function getSdkVisibility(context, type) {
    // eslint-disable-next-line @typescript-eslint/no-deprecated
    const visibility = getVisibility(context.program, type);
    if (visibility) {
        const result = [];
        if (visibility.includes("read")) {
            result.push(Visibility.Read);
        }
        if (visibility.includes("create")) {
            result.push(Visibility.Create);
        }
        if (visibility.includes("update")) {
            result.push(Visibility.Update);
        }
        if (visibility.includes("delete")) {
            result.push(Visibility.Delete);
        }
        if (visibility.includes("query")) {
            result.push(Visibility.Query);
        }
        return result;
    }
    return undefined;
}
function getSdkCredentialType(context, client, authentication) {
    const credentialTypes = [];
    for (const option of authentication.options) {
        for (const scheme of option.schemes) {
            credentialTypes.push({
                __raw: client.service,
                kind: "credential",
                scheme: scheme,
                decorators: [],
            });
        }
    }
    if (credentialTypes.length > 1) {
        return {
            __raw: client.service,
            kind: "union",
            variantTypes: credentialTypes,
            name: createGeneratedName(context, client.service, "CredentialUnion"),
            isGeneratedName: true,
            clientNamespace: getClientNamespace(context, client.service),
            crossLanguageDefinitionId: `${getCrossLanguageDefinitionId(context, client.service)}.CredentialUnion`,
            decorators: [],
            access: "public",
            usage: UsageFlags.None,
        };
    }
    return credentialTypes[0];
}
function getSdkCredentialParameter(context, client) {
    const auth = getAuthentication(context.program, client.service);
    if (!auth)
        return undefined;
    return {
        type: getSdkCredentialType(context, client, auth),
        kind: "credential",
        name: "credential",
        isGeneratedName: true,
        doc: "Credential used to authenticate requests to the service.",
        apiVersions: getAvailableApiVersions(context, client.service, client.type),
        onClient: true,
        optional: false,
        isApiVersionParam: false,
        crossLanguageDefinitionId: `${getCrossLanguageDefinitionId(context, client.service)}.credential`,
        decorators: [],
    };
}
function getSdkModelPropertyTypeBase(context, type, operation) {
    const diagnostics = createDiagnosticCollector();
    // get api version info so we can cache info about its api versions before we get to property type level
    const apiVersions = getAvailableApiVersions(context, type, operation || type.model);
    const alternateType = getAlternateType(context, type);
    let propertyType = diagnostics.pipe(getClientTypeWithDiagnostics(context, alternateType ?? type.type, operation));
    diagnostics.pipe(addEncodeInfo(context, alternateType ?? type, propertyType));
    const knownValues = getKnownValues(context.program, type);
    if (knownValues) {
        propertyType = diagnostics.pipe(getSdkEnumWithDiagnostics(context, knownValues, operation));
    }
    const name = getPropertyNames(context, type)[0];
    const onClient = isOnClient(context, type, operation, apiVersions.length > 0);
    return diagnostics.wrap({
        __raw: type,
        doc: getDoc(context.program, type),
        summary: getSummary(context.program, type),
        apiVersions,
        type: propertyType,
        name,
        isGeneratedName: false,
        optional: type.optional,
        ...updateWithApiVersionInformation(context, type, operation ? getLocationOfOperation(operation) : undefined),
        onClient,
        crossLanguageDefinitionId: getCrossLanguageDefinitionId(context, type, operation),
        decorators: diagnostics.pipe(getTypeDecorators(context, type)),
    });
}
function isFilePart(context, type) {
    if (type.kind === "array") {
        // HttpFile<T>[]
        return isFilePart(context, type.valueType);
    }
    else if (type.kind === "bytes") {
        // Http<bytes>
        return true;
    }
    else if (type.kind === "model") {
        if (type.__raw && isOrExtendsHttpFile(context.program, type.__raw)) {
            // Http<File>
            return true;
        }
        // HttpPart<{@body body: bytes}> or HttpPart<{@body body: File}>
        const body = type.properties.find((x) => x.kind === "body");
        if (body) {
            return isFilePart(context, body.type);
        }
    }
    return false;
}
function getHttpOperationParts(context, operation) {
    const body = getHttpOperationWithCache(context, operation).parameters.body;
    if (body?.bodyKind === "multipart") {
        return body.parts;
    }
    return [];
}
function hasHttpPart(context, type) {
    if (type.kind === "Model") {
        if (type.indexer) {
            // HttpPart<T>[]
            return (type.indexer.key.name === "integer" &&
                getHttpPart(context.program, type.indexer.value) !== undefined);
        }
        else {
            // HttpPart<T>
            return getHttpPart(context.program, type) !== undefined;
        }
    }
    return false;
}
function getHttpOperationPart(context, type, operation) {
    if (hasHttpPart(context, type.type)) {
        const httpOperationParts = getHttpOperationParts(context, operation);
        if (type.model &&
            httpOperationParts.length > 0 &&
            httpOperationParts.length === type.model.properties.size) {
            const index = Array.from(type.model.properties.values()).findIndex((p) => p === type);
            if (index !== -1) {
                return httpOperationParts[index];
            }
        }
    }
    return undefined;
}
function updateMultiPartInfo(context, type, base, operation) {
    const httpOperationPart = getHttpOperationPart(context, type, operation);
    const diagnostics = createDiagnosticCollector();
    if (httpOperationPart) {
        // body decorated with @multipartBody
        base.multipartOptions = {
            isFilePart: isFilePart(context, base.type),
            isMulti: httpOperationPart.multi,
            filename: httpOperationPart.filename
                ? diagnostics.pipe(getSdkModelPropertyType(context, httpOperationPart.filename, operation))
                : undefined,
            contentType: httpOperationPart.body.contentTypeProperty
                ? diagnostics.pipe(getSdkModelPropertyType(context, httpOperationPart.body.contentTypeProperty, operation))
                : undefined,
            defaultContentTypes: httpOperationPart.body.contentTypes,
        };
        // after https://github.com/microsoft/typespec/issues/3779 fixed, could use httpOperationPart.name directly
        const httpPart = getHttpPart(context.program, type.type);
        if (httpPart?.options?.name) {
            base.serializedName = httpPart?.options?.name;
        }
    }
    else {
        // common body
        const httpOperation = getHttpOperationWithCache(context, operation);
        const operationIsMultipart = Boolean(httpOperation && httpOperation.parameters.body?.contentTypes.includes("multipart/form-data"));
        if (operationIsMultipart) {
            const isBytesInput = base.type.kind === "bytes" ||
                (base.type.kind === "array" && base.type.valueType.kind === "bytes");
            // Currently we only recognize bytes and list of bytes as potential file inputs
            if (isBytesInput && getEncode(context.program, type)) {
                diagnostics.add(createDiagnostic({
                    code: "encoding-multipart-bytes",
                    target: type,
                }));
            }
            base.multipartOptions = {
                isFilePart: isBytesInput,
                isMulti: base.type.kind === "array",
                defaultContentTypes: [],
            };
        }
    }
    if (base.multipartOptions !== undefined) {
        base.isMultipartFileInput = base.multipartOptions.isFilePart;
    }
    return diagnostics.wrap(undefined);
}
function getSdkModelPropertyType(context, type, operation) {
    const diagnostics = createDiagnosticCollector();
    const clientParams = operation
        ? context.__clientToParameters.get(getLocationOfOperation(operation))
        : undefined;
    const correspondingClientParams = clientParams?.find((x) => twoParamsEquivalent(context, x.__raw, type));
    if (correspondingClientParams)
        return diagnostics.wrap(correspondingClientParams);
    const base = diagnostics.pipe(getSdkModelPropertyTypeBase(context, type, operation));
    if (isSdkHttpParameter(context, type))
        return getSdkHttpParameter(context, type, operation);
    const result = {
        ...base,
        kind: "property",
        optional: type.optional,
        visibility: getSdkVisibility(context, type),
        discriminator: false,
        serializedName: getPropertyNames(context, type)[1],
        isMultipartFileInput: false,
        flatten: shouldFlattenProperty(context, type),
    };
    if (operation && type.model) {
        const httpOperation = getHttpOperationWithCache(context, operation);
        const httpBody = httpOperation.parameters.body;
        if (httpBody) {
            const httpBodyType = isHttpBodySpread(httpBody)
                ? getHttpBodySpreadModel(httpBody.type)
                : httpBody.type;
            if (type.model === httpBodyType) {
                // only try to add multipartOptions for property of body
                diagnostics.pipe(updateMultiPartInfo(context, type, result, operation));
            }
        }
    }
    return diagnostics.wrap(result);
}
function addPropertiesToModelType(context, type, sdkType, operation) {
    const diagnostics = createDiagnosticCollector();
    for (const property of type.properties.values()) {
        if (isStatusCode(context.program, property) ||
            isNeverOrVoidType(property.type) ||
            getVisibility(context.program, property)?.includes("none") || // eslint-disable-line @typescript-eslint/no-deprecated
            sdkType.kind !== "model") {
            continue;
        }
        const clientProperty = diagnostics.pipe(getSdkModelPropertyType(context, property, operation));
        if (sdkType.properties) {
            sdkType.properties.push(clientProperty);
        }
        else {
            sdkType.properties = [clientProperty];
        }
    }
    return diagnostics.wrap(undefined);
}
function updateReferencedTypeMap(context, type, sdkType) {
    if (sdkType.kind !== "model" &&
        sdkType.kind !== "enum" &&
        sdkType.kind !== "union" &&
        sdkType.kind !== "nullable") {
        return;
    }
    if (context.referencedTypeMap === undefined) {
        context.referencedTypeMap = new Map();
    }
    context.referencedTypeMap.set(type, sdkType);
}
function updateUsageOrAccess(context, value, type, options) {
    const diagnostics = createDiagnosticCollector();
    options = options ?? {};
    options.propagation = options?.propagation ?? true;
    options.ignoreSubTypeStack = options.ignoreSubTypeStack ?? [];
    if (!type)
        return diagnostics.wrap(undefined);
    if (options?.seenTypes === undefined) {
        options.seenTypes = new Set();
    }
    if (options.seenTypes.has(type)) {
        options.skipFirst = false;
        return diagnostics.wrap(undefined); // avoid circular references
    }
    if (type.kind === "array" || type.kind === "dict") {
        diagnostics.pipe(updateUsageOrAccess(context, value, type.valueType, options));
        return diagnostics.wrap(undefined);
    }
    if (type.kind === "enumvalue") {
        diagnostics.pipe(updateUsageOrAccess(context, value, type.enumType, options));
        return diagnostics.wrap(undefined);
    }
    if (type.kind !== "model" &&
        type.kind !== "enum" &&
        type.kind !== "union" &&
        type.kind !== "nullable")
        return diagnostics.wrap(undefined);
    options.seenTypes.add(type);
    if (!options.skipFirst) {
        if (typeof value === "number") {
            // usage set
            if (options.isOverride) {
                // when a type has usage, it could not be override to narrow usage
                if (((type.usage & UsageFlags.Input) > 0 && (value & UsageFlags.Input) === 0) ||
                    ((type.usage & UsageFlags.Output) > 0 && (value & UsageFlags.Output) === 0)) {
                    diagnostics.add(createDiagnostic({
                        code: "conflict-usage-override",
                        target: type.__raw,
                    }));
                    return diagnostics.wrap(undefined);
                }
            }
            type.usage |= value;
        }
        else {
            // access set
            if (options.isOverride) {
                // when a type has access set to public, it could not be override to internal
                if (value === "internal" && type.access === "public" && type.__accessSet) {
                    diagnostics.add(createDiagnostic({
                        code: "conflict-access-override",
                        target: type.__raw,
                    }));
                }
                else {
                    type.access = value;
                }
            }
            else {
                if (!type.__accessSet || type.access !== "public") {
                    type.access = value;
                }
            }
            type.__accessSet = true;
        }
    }
    else {
        options.skipFirst = false;
        if (typeof value !== "number") {
            type.__accessSet = true;
        }
    }
    if (type.kind === "enum")
        return diagnostics.wrap(undefined);
    if (type.kind === "union") {
        for (const unionType of type.variantTypes) {
            diagnostics.pipe(updateUsageOrAccess(context, value, unionType, options));
        }
        return diagnostics.wrap(undefined);
    }
    if (type.kind === "nullable") {
        diagnostics.pipe(updateUsageOrAccess(context, value, type.type, options));
        return diagnostics.wrap(undefined);
    }
    if (!options.propagation)
        return diagnostics.wrap(undefined);
    if (type.baseModel) {
        options.ignoreSubTypeStack.push(true);
        if (context.disableUsageAccessPropagationToBase) {
            options.skipFirst = true;
        }
        diagnostics.pipe(updateUsageOrAccess(context, value, type.baseModel, options));
        options.ignoreSubTypeStack.pop();
    }
    if (type.discriminatedSubtypes &&
        (options.ignoreSubTypeStack.length === 0 || !options.ignoreSubTypeStack.at(-1))) {
        for (const discriminatedSubtype of Object.values(type.discriminatedSubtypes)) {
            options.ignoreSubTypeStack.push(false);
            diagnostics.pipe(updateUsageOrAccess(context, value, discriminatedSubtype, options));
            options.ignoreSubTypeStack.pop();
        }
    }
    if (type.additionalProperties) {
        options.ignoreSubTypeStack.push(false);
        diagnostics.pipe(updateUsageOrAccess(context, value, type.additionalProperties, options));
        options.ignoreSubTypeStack.pop();
    }
    for (const property of type.properties) {
        options.ignoreSubTypeStack.push(false);
        if (property.kind === "property" && isReadOnly(property) && value === UsageFlags.Input) {
            continue;
        }
        diagnostics.pipe(updateUsageOrAccess(context, value, property.type, options));
        options.ignoreSubTypeStack.pop();
    }
    return diagnostics.wrap(undefined);
}
function updateTypesFromOperation(context, operation) {
    const diagnostics = createDiagnosticCollector();
    const program = context.program;
    const httpOperation = getHttpOperationWithCache(context, operation);
    const generateConvenient = shouldGenerateConvenient(context, operation);
    const overriddenClientMethod = getOverriddenClientMethod(context, operation);
    for (const param of (overriddenClientMethod ?? operation).parameters.properties.values()) {
        if (isNeverOrVoidType(param.type))
            continue;
        // if it is a body model, skip
        if (httpOperation.parameters.body?.property === param)
            continue;
        const sdkType = diagnostics.pipe(getClientTypeWithDiagnostics(context, param.type, operation));
        if (generateConvenient) {
            diagnostics.pipe(updateUsageOrAccess(context, UsageFlags.Input, sdkType));
        }
        const access = getAccessOverride(context, operation) ?? "public";
        diagnostics.pipe(updateUsageOrAccess(context, access, sdkType));
    }
    for (const param of httpOperation.parameters.parameters) {
        if (isNeverOrVoidType(param.param.type))
            continue;
        const sdkType = diagnostics.pipe(getClientTypeWithDiagnostics(context, param.param.type, operation));
        if (generateConvenient) {
            diagnostics.pipe(updateUsageOrAccess(context, UsageFlags.Input, sdkType));
        }
        const access = getAccessOverride(context, operation) ?? "public";
        diagnostics.pipe(updateUsageOrAccess(context, access, sdkType));
    }
    const httpBody = httpOperation.parameters.body;
    if (httpBody && !isNeverOrVoidType(httpBody.type)) {
        const spread = isHttpBodySpread(httpBody);
        let sdkType;
        if (spread) {
            sdkType = diagnostics.pipe(getClientTypeWithDiagnostics(context, getHttpBodySpreadModel(httpBody.type), operation));
        }
        else {
            sdkType = diagnostics.pipe(getClientTypeWithDiagnostics(context, httpBody.type, operation));
        }
        const multipartOperation = isMultipartOperation(context, operation);
        if (generateConvenient) {
            if (spread) {
                updateUsageOrAccess(context, UsageFlags.Spread, sdkType, { propagation: false });
                updateUsageOrAccess(context, UsageFlags.Input, sdkType, { skipFirst: true });
            }
            else {
                updateUsageOrAccess(context, UsageFlags.Input, sdkType);
            }
            if (httpBody.contentTypes.some((x) => isJsonContentType(x))) {
                diagnostics.pipe(updateUsageOrAccess(context, UsageFlags.Json, sdkType));
            }
            if (httpBody.contentTypes.some((x) => isXmlContentType(x))) {
                diagnostics.pipe(updateUsageOrAccess(context, UsageFlags.Xml, sdkType));
            }
            if (httpBody.contentTypes.includes("application/merge-patch+json")) {
                // will also have Json type
                diagnostics.pipe(updateUsageOrAccess(context, UsageFlags.JsonMergePatch, sdkType));
            }
            if (multipartOperation) {
                diagnostics.pipe(updateUsageOrAccess(context, UsageFlags.MultipartFormData, sdkType, {
                    propagation: false,
                }));
            }
            // after completion of usage calculation for httpBody, check whether it has
            // conflicting usage between multipart and regular body
            if (sdkType.kind === "model") {
                const isUsedInMultipart = (sdkType.usage & UsageFlags.MultipartFormData) > 0;
                const isUsedInOthers = ((sdkType.usage & UsageFlags.Json) | (sdkType.usage & UsageFlags.Xml)) > 0;
                if ((!multipartOperation && isUsedInMultipart) || (multipartOperation && isUsedInOthers)) {
                    // This means we have a model that is used both for formdata input and for regular body input
                    diagnostics.add(createDiagnostic({
                        code: "conflicting-multipart-model-usage",
                        target: httpBody.type,
                        format: {
                            modelName: sdkType.name,
                        },
                    }));
                }
            }
        }
        const access = getAccessOverride(context, operation) ?? "public";
        diagnostics.pipe(updateUsageOrAccess(context, access, sdkType));
    }
    const lroMetaData = getLroMetadata(program, operation);
    for (const response of httpOperation.responses) {
        for (const innerResponse of response.responses) {
            if (innerResponse.body?.type && !isNeverOrVoidType(innerResponse.body.type)) {
                const body = innerResponse.body.type.kind === "Model"
                    ? getEffectivePayloadType(context, innerResponse.body.type)
                    : innerResponse.body.type;
                const sdkType = diagnostics.pipe(getClientTypeWithDiagnostics(context, body, operation));
                if (generateConvenient) {
                    if (response.statusCodes === "*" || isErrorModel(context.program, body)) {
                        diagnostics.pipe(updateUsageOrAccess(context, UsageFlags.Exception, sdkType));
                    }
                    else if (lroMetaData !== undefined) {
                        // when the operation is an lro, the response should be its initial response.
                        diagnostics.pipe(updateUsageOrAccess(context, UsageFlags.LroInitial, sdkType));
                    }
                    else {
                        diagnostics.pipe(updateUsageOrAccess(context, UsageFlags.Output, sdkType));
                    }
                    if (innerResponse.body.contentTypes.some((x) => isJsonContentType(x))) {
                        diagnostics.pipe(updateUsageOrAccess(context, UsageFlags.Json, sdkType));
                    }
                }
                const access = getAccessOverride(context, operation) ?? "public";
                diagnostics.pipe(updateUsageOrAccess(context, access, sdkType));
            }
            const headers = getHttpOperationResponseHeaders(innerResponse);
            if (headers) {
                for (const header of Object.values(headers)) {
                    if (isNeverOrVoidType(header.type))
                        continue;
                    const sdkType = diagnostics.pipe(getClientTypeWithDiagnostics(context, header.type, operation));
                    if (generateConvenient) {
                        diagnostics.pipe(updateUsageOrAccess(context, UsageFlags.Output, sdkType));
                    }
                    const access = getAccessOverride(context, operation) ?? "public";
                    diagnostics.pipe(updateUsageOrAccess(context, access, sdkType));
                }
            }
        }
    }
    if (lroMetaData && generateConvenient) {
        // the final result will be normal output usage.
        updateUsageOrAccessForLroComponent(lroMetaData.finalResult, UsageFlags.Output);
        // the final envelope result will have LroFinalEnvelope.
        updateUsageOrAccessForLroComponent(lroMetaData.finalEnvelopeResult, UsageFlags.LroFinalEnvelope);
        // the polling model will have LroPolling.
        updateUsageOrAccessForLroComponent(lroMetaData.pollingInfo.responseModel, UsageFlags.LroPolling);
    }
    return diagnostics.wrap(undefined);
    function updateUsageOrAccessForLroComponent(model, usage) {
        if (model === undefined || model === "void")
            return;
        const sdkType = diagnostics.pipe(getClientTypeWithDiagnostics(context, model, operation));
        diagnostics.pipe(updateUsageOrAccess(context, usage, sdkType));
        const access = getAccessOverride(context, operation) ?? "public";
        diagnostics.pipe(updateUsageOrAccess(context, access, sdkType));
    }
}
function updateAccessOverride(context) {
    const diagnostics = createDiagnosticCollector();
    // set access for all orphan model without override
    for (const sdkType of context.referencedTypeMap?.values() ?? []) {
        const accessOverride = getAccessOverride(context, sdkType.__raw);
        if (!sdkType.__accessSet && accessOverride === undefined) {
            diagnostics.pipe(updateUsageOrAccess(context, "public", sdkType));
        }
    }
    for (const sdkType of context.referencedTypeMap?.values() ?? []) {
        const accessOverride = getAccessOverride(context, sdkType.__raw);
        if (accessOverride) {
            diagnostics.pipe(updateUsageOrAccess(context, accessOverride, sdkType, { isOverride: true }));
        }
    }
    return diagnostics.wrap(undefined);
}
function updateUsageOverride(context) {
    const diagnostics = createDiagnosticCollector();
    for (const sdkType of context.referencedTypeMap?.values() ?? []) {
        const usageOverride = getUsageOverride(context, sdkType.__raw);
        if (usageOverride) {
            diagnostics.pipe(updateUsageOrAccess(context, usageOverride, sdkType, { isOverride: true }));
        }
    }
    return diagnostics.wrap(undefined);
}
function updateSpreadModelUsageAndAccess(context) {
    for (const [_, sdkType] of context.referencedTypeMap?.entries() ?? []) {
        if (sdkType.kind === "model" &&
            (sdkType.usage & UsageFlags.Spread) > 0 &&
            (sdkType.usage & (UsageFlags.Input | UsageFlags.Output)) === 0) {
            // if a type has spread usage, but not used in any other operation, then set it to be internal
            sdkType.access = "internal";
        }
    }
}
function handleServiceOrphanType(context, type) {
    const diagnostics = createDiagnosticCollector();
    const sdkType = diagnostics.pipe(getClientTypeWithDiagnostics(context, type));
    diagnostics.pipe(updateUsageOrAccess(context, UsageFlags.None, sdkType));
    return diagnostics.wrap(undefined);
}
function filterOutTypes(context, filter) {
    const result = new Array();
    for (const sdkType of context.referencedTypeMap?.values() ?? []) {
        // filter models with unexpected usage
        if ((sdkType.usage & filter) === 0) {
            continue;
        }
        if (!result.includes(sdkType)) {
            result.push(sdkType);
        }
    }
    return result;
}
function getAllModelsWithDiagnostics(context, options = {}) {
    const diagnostics = createDiagnosticCollector();
    return diagnostics.wrap(filterOutTypes(context, getFilterNumber(options)).filter((x) => x.kind === "model" || x.kind === "enum"));
}
function getAllModels(context, options = {}) {
    // we currently don't return diagnostics even though we keep track of them
    // when we move to the new sdk type ecosystem completely, we'll expose
    // diagnostics as a separate property on the TCGCContext
    return ignoreDiagnostics(getAllModelsWithDiagnostics(context, options));
}
function getFilterNumber(options = {}) {
    options = { input: true, output: true, ...options };
    let filter = 0;
    if (options.input && options.output) {
        filter = Number.MAX_SAFE_INTEGER;
    }
    else if (options.input) {
        filter += UsageFlags.Input;
    }
    else if (options.output) {
        filter += UsageFlags.Output;
    }
    return filter;
}
function getAllReferencedTypes(context, options = {}) {
    return filterOutTypes(context, getFilterNumber(options));
}
function handleAllTypes(context) {
    const diagnostics = createDiagnosticCollector();
    if (context.referencedTypeMap === undefined) {
        context.referencedTypeMap = new Map();
    }
    for (const client of listClients(context)) {
        for (const operation of listOperationsInOperationGroup(context, client)) {
            // operations on a client
            diagnostics.pipe(updateTypesFromOperation(context, operation));
        }
        const ogs = listOperationGroups(context, client);
        while (ogs.length) {
            const operationGroup = ogs.pop();
            for (const operation of listOperationsInOperationGroup(context, operationGroup)) {
                // operations on operation groups
                diagnostics.pipe(updateTypesFromOperation(context, operation));
            }
            if (operationGroup?.subOperationGroups) {
                ogs.push(...operationGroup.subOperationGroups);
            }
        }
        // server parameters
        const servers = getServers(context.program, client.service);
        if (servers !== undefined && servers[0].parameters !== undefined) {
            for (const param of servers[0].parameters.values()) {
                const sdkType = diagnostics.pipe(getClientTypeWithDiagnostics(context, param));
                diagnostics.pipe(updateUsageOrAccess(context, UsageFlags.Input, sdkType));
            }
        }
        // versioned enums
        const [_, versionMap] = getVersions(context.program, client.service);
        if (versionMap && versionMap.getVersions()[0]) {
            // create sdk enum for versions enum
            const sdkVersionsEnum = diagnostics.pipe(getSdkEnumWithDiagnostics(context, versionMap.getVersions()[0].enumMember.enum));
            filterApiVersionsInEnum(context, client, sdkVersionsEnum);
            diagnostics.pipe(updateUsageOrAccess(context, UsageFlags.ApiVersionEnum, sdkVersionsEnum));
        }
    }
    // update for orphan models/enums/unions
    for (const client of listClients(context)) {
        const namespaces = [client.service];
        while (namespaces.length) {
            const namespace = namespaces.pop();
            // orphan models
            for (const model of namespace.models.values()) {
                diagnostics.pipe(handleServiceOrphanType(context, model));
            }
            // orphan enums
            for (const enumType of namespace.enums.values()) {
                diagnostics.pipe(handleServiceOrphanType(context, enumType));
            }
            // orphan unions
            for (const unionType of namespace.unions.values()) {
                diagnostics.pipe(handleServiceOrphanType(context, unionType));
            }
            namespaces.push(...namespace.namespaces.values());
        }
    }
    // update access
    diagnostics.pipe(updateAccessOverride(context));
    // update usage
    diagnostics.pipe(updateUsageOverride(context));
    // update spread model
    updateSpreadModelUsageAndAccess(context);
    return diagnostics.wrap(undefined);
}

const namespace = "Azure.ClientGenerator.Core";
function getScopedDecoratorData(context, key, target, languageScope) {
    const retval = context.program.stateMap(key).get(target);
    if (retval === undefined)
        return retval;
    if (languageScope === AllScopes) {
        return retval[languageScope];
    }
    if (languageScope === undefined || typeof languageScope === "string") {
        const scope = languageScope ?? context.emitterName;
        if (Object.keys(retval).includes(scope))
            return retval[scope];
        // if the scope is negated, we should return undefined
        // if the scope is not negated, we should return the value for AllScopes
        const negationScopes = retval[negationScopesKey];
        if (negationScopes !== undefined && negationScopes.includes(scope)) {
            return undefined;
        }
    }
    return retval[AllScopes]; // in this case it applies to all languages
}
function listScopedDecoratorData(context, key) {
    const retval = [...context.program.stateMap(key).values()];
    return retval
        .filter((targetEntry) => {
        return targetEntry[context.emitterName] || targetEntry[AllScopes];
    })
        .flatMap((targetEntry) => targetEntry[context.emitterName] ?? targetEntry[AllScopes]);
}
function setScopedDecoratorData(context, decorator, key, target, value, scope) {
    // if no scope specified, then set with the new value
    if (!scope) {
        context.program.stateMap(key).set(target, Object.fromEntries([[AllScopes, value]]));
        return;
    }
    const targetEntry = context.program.stateMap(key).get(target);
    const [negationScopes, scopes] = parseScopes(context, scope);
    if (negationScopes !== undefined && negationScopes.length > 0) {
        // override the previous value for negation scopes
        const newObject = scopes !== undefined && scopes.length > 0
            ? Object.fromEntries([AllScopes, ...scopes].map((scope) => [scope, value]))
            : Object.fromEntries([[AllScopes, value]]);
        newObject[negationScopesKey] = negationScopes;
        context.program.stateMap(key).set(target, newObject);
        // if a scope exists in the target entry and it overlaps with the negation scope, it means negation scope doesn't override it
        if (targetEntry !== undefined) {
            const existingScopes = Object.getOwnPropertyNames(targetEntry);
            const intersections = existingScopes.filter((x) => negationScopes.includes(x));
            if (intersections !== undefined && intersections.length > 0) {
                for (const scopeToKeep of intersections) {
                    newObject[scopeToKeep] = targetEntry[scopeToKeep];
                }
            }
        }
    }
    else if (scopes !== undefined && scopes.length > 0) {
        // for normal scopes, add them incrementally
        const newObject = Object.fromEntries(scopes.map((scope) => [scope, value]));
        context.program
            .stateMap(key)
            .set(target, !targetEntry ? newObject : { ...targetEntry, ...newObject });
    }
}
function parseScopes(context, scope) {
    if (scope === undefined) {
        return [undefined, undefined];
    }
    // handle !(scope1, scope2,...) syntax
    const negationScopeRegex = new RegExp(/!\((.*?)\)/);
    const negationScopeMatch = scope.match(negationScopeRegex);
    if (negationScopeMatch) {
        return [negationScopeMatch[1].split(",").map((s) => s.trim()), undefined];
    }
    // handle !scope1, !scope2, scope3, ... syntax
    const splitScopes = scope.split(",").map((s) => s.trim());
    const negationScopes = [];
    const scopes = [];
    for (const s of splitScopes) {
        if (s.startsWith("!")) {
            negationScopes.push(s.slice(1));
        }
        else {
            scopes.push(s);
        }
    }
    return [negationScopes, scopes];
}
const clientKey = createStateSymbol("client");
function isArm(service) {
    return service.decorators.some((decorator) => decorator.decorator.name === "$armProviderNamespace");
}
const $client = (context, target, options, scope) => {
    if (context.decoratorTarget.kind === SyntaxKind.AugmentDecoratorStatement) {
        reportDiagnostic(context.program, {
            code: "wrong-client-decorator",
            target: context.decoratorTarget,
        });
        return;
    }
    const explicitName = options?.properties.get("name")?.type;
    const name = explicitName?.kind === "String" ? explicitName.value : target.name;
    const explicitService = options?.properties.get("service")?.type;
    const service = explicitService?.kind === "Namespace"
        ? explicitService
        : (findClientService(context.program, target) ?? target);
    if (!name.endsWith("Client")) {
        reportDiagnostic(context.program, {
            code: "client-name",
            format: { name },
            target: context.decoratorTarget,
        });
    }
    if (!isService(context.program, service)) {
        reportDiagnostic(context.program, {
            code: "client-service",
            format: { name },
            target: context.decoratorTarget,
        });
    }
    const client = {
        kind: "SdkClient",
        name,
        service,
        type: target,
        crossLanguageDefinitionId: `${getNamespaceFullName(service)}.${name}`,
    };
    setScopedDecoratorData(context, $client, clientKey, target, client, scope);
};
function findClientService(program, client) {
    let current = client;
    while (current) {
        if (isService(program, current)) {
            return current;
        }
        current = current.namespace;
    }
    return undefined;
}
function findOperationGroupService(program, client, scope) {
    let current = client;
    while (current) {
        if (isService(program, current)) {
            // we don't check scoped clients here, because we want to find the service for the client
            return current;
        }
        const client = program.stateMap(clientKey).get(current);
        if (client && (client[scope] || client[AllScopes])) {
            return (client[scope] ?? client[AllScopes]).service;
        }
        current = current.namespace;
    }
    return undefined;
}
/**
 * Return the client object for the given namespace or interface, or undefined if the given namespace or interface is not a client.
 *
 * @param context TCGCContext
 * @param type Type to check
 * @returns Client or undefined
 */
function getClient(context, type) {
    for (const client of listClients(context)) {
        if (client.type === type) {
            return client;
        }
    }
    return undefined;
}
function hasExplicitClientOrOperationGroup(context) {
    return (listScopedDecoratorData(context, clientKey).length > 0 ||
        listScopedDecoratorData(context, operationGroupKey).length > 0);
}
function serviceVersioningProjection(context, client) {
    if (!context.__service_projection) {
        context.__service_projection = new Map();
    }
    let projectedService;
    let projectedProgram;
    if (context.__service_projection.has(client.service)) {
        [projectedService, projectedProgram] = context.__service_projection.get(client.service);
    }
    else {
        const allApiVersions = getVersions(context.program, client.service)[1]
            ?.getVersions()
            .map((x) => x.value);
        if (!allApiVersions)
            return;
        const apiVersion = getValidApiVersion(context, allApiVersions);
        if (apiVersion === undefined)
            return;
        const versionProjections = buildVersionProjections(context.program, client.service).filter((v) => apiVersion === v.version);
        if (versionProjections.length !== 1)
            throw new Error("Version projects should only contain one element");
        const projectedVersion = versionProjections[0];
        if (projectedVersion.projections.length > 0) {
            projectedProgram = context.program = projectProgram(context.originalProgram, projectedVersion.projections);
        }
        projectedService = projectedProgram
            ? projectedProgram.projector.projectedTypes.get(client.service)
            : client.service;
        context.__service_projection.set(client.service, [projectedService, projectedProgram]);
    }
    if (client.service !== client.type) {
        client.type = projectedProgram
            ? projectedProgram.projector.projectedTypes.get(client.type)
            : client.type;
    }
    else {
        client.type = projectedService;
    }
    client.service = projectedService;
}
function getClientsWithVersioning(context, clients) {
    if (context.apiVersion !== "all") {
        const projectedClients = [];
        for (const client of clients) {
            const projectedClient = { ...client };
            serviceVersioningProjection(context, projectedClient);
            // filter client not existed in the current version
            if (projectedClient.type.kind !== "Intrinsic") {
                projectedClients.push(projectedClient);
            }
        }
        return projectedClients;
    }
    return clients;
}
/**
 * List all the clients.
 *
 * @param context TCGCContext
 * @returns Array of clients
 */
function listClients(context) {
    if (context.__rawClients)
        return context.__rawClients;
    const explicitClients = [...listScopedDecoratorData(context, clientKey)];
    if (explicitClients.length > 0) {
        context.__rawClients = getClientsWithVersioning(context, explicitClients);
        if (context.__rawClients.some((client) => isArm(client.service))) {
            context.arm = true;
        }
        return context.__rawClients;
    }
    // if there is no explicit client, we will treat namespaces with service decorator as clients
    const services = listServices(context.program);
    const clients = services.map((service) => {
        let originalName = service.type.name;
        const clientNameOverride = getClientNameOverride(context, service.type);
        if (clientNameOverride) {
            originalName = clientNameOverride;
        }
        else {
            originalName = getProjectedName(context.program, service.type, "client") ?? service.type.name;
        }
        const clientName = originalName.endsWith("Client") ? originalName : `${originalName}Client`;
        context.arm = isArm(service.type);
        return {
            kind: "SdkClient",
            name: clientName,
            service: service.type,
            type: service.type,
            crossLanguageDefinitionId: getNamespaceFullName(service.type),
        };
    });
    context.__rawClients = getClientsWithVersioning(context, clients);
    return context.__rawClients;
}
const operationGroupKey = createStateSymbol("operationGroup");
const $operationGroup = (context, target, scope) => {
    if (context.decoratorTarget.kind === SyntaxKind.AugmentDecoratorStatement) {
        reportDiagnostic(context.program, {
            code: "wrong-client-decorator",
            target: context.decoratorTarget,
        });
        return;
    }
    setScopedDecoratorData(context, $operationGroup, operationGroupKey, target, {
        kind: "SdkOperationGroup",
        type: target,
    }, scope);
};
/**
 * Check a namespace or interface is an operation group.
 * @param context TCGCContext
 * @param type Type to check
 * @returns boolean
 */
function isOperationGroup(context, type) {
    if (hasExplicitClientOrOperationGroup(context)) {
        return getScopedDecoratorData(context, operationGroupKey, type) !== undefined;
    }
    // if there is no explicit client, we will treat non-client namespaces and all interfaces as operation group
    if (type.kind === "Interface" && !isTemplateDeclaration(type)) {
        return true;
    }
    if (type.kind === "Namespace" && !type.decorators.some((t) => t.decorator.name === "$service")) {
        return true;
    }
    return false;
}
/**
 * Check an operation is in an operation group.
 * @param context TCGCContext
 * @param type Type to check
 * @returns boolean
 */
function isInOperationGroup(context, type) {
    switch (type.kind) {
        case "Operation":
            return type.interface
                ? isInOperationGroup(context, type.interface)
                : type.namespace
                    ? isInOperationGroup(context, type.namespace)
                    : false;
        case "Interface":
        case "Namespace":
            return (isOperationGroup(context, type) ||
                (type.namespace ? isInOperationGroup(context, type.namespace) : false));
    }
}
function buildOperationGroupPath(context, type) {
    const path = [];
    while (true) {
        const client = getClient(context, type);
        if (client) {
            path.push(client.name);
            break;
        }
        if (isOperationGroup(context, type)) {
            path.push(getLibraryName(context, type));
        }
        if (type.namespace) {
            type = type.namespace;
        }
        else {
            break;
        }
    }
    return path.reverse().join(".");
}
/**
 * Return the operation group object for the given namespace or interface or undefined is not an operation group.
 * @param context TCGCContext
 * @param type Type to check
 * @returns Operation group or undefined.
 */
function getOperationGroup(context, type) {
    let operationGroup;
    const service = findOperationGroupService(context.program, type, context.emitterName) ?? type;
    if (!isService(context.program, service)) {
        reportDiagnostic(context.program, {
            code: "client-service",
            format: { name: type.name },
            target: type,
        });
    }
    if (hasExplicitClientOrOperationGroup(context)) {
        operationGroup = getScopedDecoratorData(context, operationGroupKey, type);
        if (operationGroup) {
            operationGroup.groupPath = buildOperationGroupPath(context, type);
            operationGroup.service = service;
        }
    }
    else {
        // if there is no explicit client, we will treat non-client namespaces and all interfaces as operation group
        if (type.kind === "Interface" && !isTemplateDeclaration(type)) {
            operationGroup = {
                kind: "SdkOperationGroup",
                type,
                groupPath: buildOperationGroupPath(context, type),
                service,
            };
        }
        if (type.kind === "Namespace" &&
            !type.decorators.some((t) => t.decorator.name === "$service")) {
            operationGroup = {
                kind: "SdkOperationGroup",
                type,
                groupPath: buildOperationGroupPath(context, type),
                service,
            };
        }
    }
    // build hierarchy of operation group
    if (operationGroup && type.kind === "Namespace") {
        const subOperationGroups = [];
        type.namespaces.forEach((ns) => {
            const subOperationGroup = getOperationGroup(context, ns);
            if (subOperationGroup) {
                subOperationGroups.push(subOperationGroup);
            }
        });
        type.interfaces.forEach((i) => {
            const subOperationGroup = getOperationGroup(context, i);
            if (subOperationGroup) {
                subOperationGroups.push(subOperationGroup);
            }
        });
        if (subOperationGroups.length > 0) {
            operationGroup.subOperationGroups = subOperationGroups;
        }
    }
    return operationGroup;
}
/**
 * List all the operation groups inside a client or an operation group. If ignoreHierarchy is true, the result will include all nested operation groups.
 *
 * @param context TCGCContext
 * @param group Client or operation group to list operation groups
 * @param ignoreHierarchy Whether to get all nested operation groups
 * @returns
 */
function listOperationGroups(context, group, ignoreHierarchy = false) {
    const groups = [];
    if (group.type.kind === "Interface") {
        return groups;
    }
    for (const subItem of group.type.namespaces.values()) {
        track(getOperationGroup(context, subItem));
    }
    for (const subItem of group.type.interfaces.values()) {
        track(getOperationGroup(context, subItem));
    }
    function track(item) {
        if (!item) {
            return;
        }
        groups.push(item);
        if (ignoreHierarchy) {
            for (const subItem of item.subOperationGroups ?? []) {
                track(subItem);
            }
        }
    }
    return groups;
}
/**
 * List operations inside a client or an operation group. If ignoreHierarchy is true, the result will include all nested operations.
 * @param program TCGCContext
 * @param group Client or operation group to list operations
 * @param ignoreHierarchy Whether to get all nested operations
 * @returns
 */
function listOperationsInOperationGroup(context, group, ignoreHierarchy = false) {
    const operations = [];
    function addOperations(current) {
        if (current !== group.type &&
            !ignoreHierarchy &&
            (getClient(context, current) || isOperationGroup(context, current))) {
            return;
        }
        if (current.kind === "Interface" && isTemplateDeclaration(current)) {
            // Skip template interface operations
            return;
        }
        for (const op of current.operations.values()) {
            if (!IsInScope(context, op)) {
                continue;
            }
            // Skip templated operations and omit operations
            if (!isTemplateDeclarationOrInstance(op) &&
                !context.program.stateMap(omitOperation).get(op)) {
                operations.push(op);
            }
        }
        if (current.kind === "Namespace") {
            for (const subItem of current.namespaces.values()) {
                addOperations(subItem);
            }
            for (const subItem of current.interfaces.values()) {
                addOperations(subItem);
            }
        }
    }
    addOperations(group.type);
    return operations;
}
const protocolAPIKey = createStateSymbol("protocolAPI");
const $protocolAPI = (context, entity, value, scope) => {
    setScopedDecoratorData(context, $protocolAPI, protocolAPIKey, entity, value, scope);
};
const convenientAPIKey = createStateSymbol("convenientAPI");
const $convenientAPI = (context, entity, value, scope) => {
    setScopedDecoratorData(context, $convenientAPI, convenientAPIKey, entity, value, scope);
};
function shouldGenerateProtocol(context, entity) {
    const value = getScopedDecoratorData(context, protocolAPIKey, entity);
    return value ?? Boolean(context.generateProtocolMethods);
}
function shouldGenerateConvenient(context, entity) {
    const value = getScopedDecoratorData(context, convenientAPIKey, entity);
    return value ?? Boolean(context.generateConvenienceMethods);
}
const usageKey = createStateSymbol("usage");
const $usage = (context, entity, value, scope) => {
    const isValidValue = (value) => value === 2 || value === 4;
    if (value.kind === "EnumMember") {
        if (typeof value.value === "number" && isValidValue(value.value)) {
            setScopedDecoratorData(context, $usage, usageKey, entity, value.value, scope);
            return;
        }
    }
    else {
        let usage = 0;
        for (const variant of value.variants.values()) {
            if (variant.type.kind === "EnumMember" && typeof variant.type.value === "number") {
                if (isValidValue(variant.type.value)) {
                    usage |= variant.type.value;
                }
            }
            else {
                break;
            }
        }
        if (usage !== 0) {
            setScopedDecoratorData(context, $usage, usageKey, entity, usage, scope);
            return;
        }
    }
    reportDiagnostic(context.program, {
        code: "invalid-usage",
        format: {},
        target: entity,
    });
};
function getUsageOverride(context, entity) {
    const usageFlags = getScopedDecoratorData(context, usageKey, entity);
    if (usageFlags || entity.namespace === undefined)
        return usageFlags;
    return getScopedDecoratorData(context, usageKey, entity.namespace);
}
function getUsage(context, entity) {
    switch (entity.kind) {
        case "Union":
            const type = getSdkUnion(context, entity);
            if (type.kind === "enum" || type.kind === "union" || type.kind === "nullable") {
                return type.usage;
            }
            return UsageFlags.None;
        case "Model":
            return getSdkModel(context, entity).usage;
        case "Enum":
            return getSdkEnum(context, entity).usage;
    }
}
const accessKey = createStateSymbol("access");
const $access = (context, entity, value, scope) => {
    if (typeof value.value !== "string" || (value.value !== "public" && value.value !== "internal")) {
        reportDiagnostic(context.program, {
            code: "access",
            format: {},
            target: entity,
        });
    }
    setScopedDecoratorData(context, $access, accessKey, entity, value.value, scope);
};
function getAccessOverride(context, entity) {
    const accessOverride = getScopedDecoratorData(context, accessKey, entity);
    if (!accessOverride && entity.namespace) {
        return getAccessOverride(context, entity.namespace);
    }
    return accessOverride;
}
function getAccess(context, entity) {
    const override = getAccessOverride(context, entity);
    if (override || entity.kind === "Operation") {
        return override || "public";
    }
    switch (entity.kind) {
        case "Model":
            return getSdkModel(context, entity).access;
        case "Enum":
            return getSdkEnum(context, entity).access;
        case "Union": {
            const type = getSdkUnion(context, entity);
            if (type.kind === "enum" || type.kind === "union" || type.kind === "nullable") {
                return type.access;
            }
            return "public";
        }
    }
}
const flattenPropertyKey = createStateSymbol("flattenPropertyKey");
/**
 * Whether a model property should be flattened.
 *
 * @param context DecoratorContext
 * @param target ModelProperty to mark as flattened
 * @param scope Names of the projection (e.g. "python", "csharp", "java", "javascript")
 * @deprecated This decorator is not recommended to use.
 */
const $flattenProperty = (context, target, scope) => {
    if (getDiscriminator(context.program, target.type)) {
        reportDiagnostic(context.program, {
            code: "flatten-polymorphism",
            format: {},
            target: target,
        });
    }
    setScopedDecoratorData(context, $flattenProperty, flattenPropertyKey, target, true, scope); // eslint-disable-line @typescript-eslint/no-deprecated
};
/**
 * Whether a model property should be flattened or not.
 *
 * @param context TCGCContext
 * @param target ModelProperty that we want to check whether it should be flattened or not
 * @returns whether the model property should be flattened or not
 */
function shouldFlattenProperty(context, target) {
    return getScopedDecoratorData(context, flattenPropertyKey, target) ?? false;
}
const $clientName = (context, entity, value, scope) => {
    // workaround for current lack of functionality in compiler
    // https://github.com/microsoft/typespec/issues/2717
    if (entity.kind === "Model" || entity.kind === "Operation") {
        if (context.decoratorTarget.kind === SyntaxKind.AugmentDecoratorStatement) {
            if (ignoreDiagnostics(context.program.checker.resolveTypeReference(context.decoratorTarget.targetType))?.node !== entity.node) {
                return;
            }
        }
        if (context.decoratorTarget.kind === SyntaxKind.DecoratorExpression) {
            if (context.decoratorTarget.parent !== entity.node) {
                return;
            }
        }
    }
    if (value.trim() === "") {
        reportDiagnostic(context.program, {
            code: "empty-client-name",
            format: {},
            target: entity,
        });
    }
    setScopedDecoratorData(context, $clientName, clientNameKey, entity, value, scope);
};
function getClientNameOverride(context, entity, languageScope) {
    return getScopedDecoratorData(context, clientNameKey, entity, languageScope);
}
const overrideKey = createStateSymbol("override");
const omitOperation = createStateSymbol("omitOperation");
// Recursive function to collect parameter names
function collectParams(properties, params = []) {
    properties.forEach((value, key) => {
        // If the property is of type 'model', recurse into its properties
        if (params.filter((x) => compareModelProperties(x, value)).length === 0) {
            if (value.type.kind === "Model") {
                collectParams(value.type.properties, params);
            }
            else {
                let sourceProp = value;
                while (sourceProp.sourceProperty) {
                    sourceProp = sourceProp.sourceProperty;
                }
                if (sourceProp.model && !isAzureCoreTspModel(sourceProp.model)) {
                    params.push(value);
                }
                else if (!sourceProp.model) {
                    params.push(value);
                }
                else {
                    // eslint-disable-next-line no-console
                    console.log(`We are not counting "${sourceProp.name}" as part of a method parameter because it's been added by Azure.Core templates`);
                }
            }
        }
    });
    return params;
}
function compareModelProperties(modelPropA, modelPropB) {
    // can't rely fully on equals because the `.model` property may be different
    return (modelPropA.name === modelPropB.name &&
        modelPropA.type === modelPropB.type &&
        modelPropA.node === modelPropB.node);
}
const $override = (context, original, override, scope) => {
    // omit all override operation
    context.program.stateMap(omitOperation).set(override, true);
    // Extract and sort parameter names
    const originalParams = collectParams(original.parameters.properties).sort((a, b) => a.name.localeCompare(b.name));
    const overrideParams = collectParams(override.parameters.properties).sort((a, b) => a.name.localeCompare(b.name));
    // Check if the sorted parameter names arrays are equal
    const parametersMatch = originalParams.length === overrideParams.length &&
        originalParams.every((value, index) => compareModelProperties(value, overrideParams[index]));
    if (!parametersMatch) {
        reportDiagnostic(context.program, {
            code: "override-method-parameters-mismatch",
            target: context.decoratorTarget,
            format: {
                methodName: original.name,
                originalParameters: originalParams.map((x) => x.name).join(`", "`),
                overrideParameters: overrideParams.map((x) => x.name).join(`", "`),
            },
        });
    }
    setScopedDecoratorData(context, $override, overrideKey, original, override, scope);
};
/**
 * Gets additional information on how to serialize / deserialize TYPESPEC standard types depending
 * on whether additional serialization information is provided or needed
 *
 * @param context the Sdk Context
 * @param entity the entity whose client format we are going to get
 * @returns the format in which to serialize the typespec type or undefined
 */
function getOverriddenClientMethod(context, entity) {
    return getScopedDecoratorData(context, overrideKey, entity);
}
const alternateTypeKey = createStateSymbol("alternateType");
/**
 * Replace a source type with an alternate type in a specific scope.
 *
 * @param context the decorator context
 * @param source source type to be replaced
 * @param alternate target type to replace the source type
 * @param scope Names of the projection (e.g. "python", "csharp", "java", "javascript")
 */
const $alternateType = (context, source, alternate, scope) => {
    if (source.kind === "ModelProperty" && source.type.kind !== "Scalar") {
        reportDiagnostic(context.program, {
            code: "invalid-alternate-source-type",
            format: {
                typeName: source.type.kind,
            },
            target: source,
        });
    }
    setScopedDecoratorData(context, $alternateType, alternateTypeKey, source, alternate, scope);
};
/**
 * Get the alternate type for a source type in a specific scope.
 *
 * @param context the Sdk Context
 * @param source source type to be replaced
 * @returns alternate type to replace the source type, or undefined if no alternate type is found
 */
function getAlternateType(context, source) {
    return getScopedDecoratorData(context, alternateTypeKey, source);
}
const $useSystemTextJsonConverter = (context, entity, scope) => { };
const clientInitializationKey = createStateSymbol("clientInitialization");
const $clientInitialization = (context, target, options, scope) => {
    setScopedDecoratorData(context, $clientInitialization, clientInitializationKey, target, options, scope);
};
function getClientInitialization(context, entity) {
    const model = getScopedDecoratorData(context, clientInitializationKey, entity);
    if (!model)
        return model;
    const sdkModel = getSdkModel(context, model);
    const initializationProps = sdkModel.properties.map((property) => {
        property.onClient = true;
        property.kind = "method";
        return property;
    });
    return {
        ...sdkModel,
        properties: initializationProps,
    };
}
const paramAliasKey = createStateSymbol("paramAlias");
const $paramAlias = (context, original, paramAlias, scope) => {
    setScopedDecoratorData(context, $paramAlias, paramAliasKey, original, paramAlias, scope);
};
function getParamAlias(context, original) {
    return getScopedDecoratorData(context, paramAliasKey, original);
}
const apiVersionKey = createStateSymbol("apiVersion");
const $apiVersion = (context, target, value, scope) => {
    setScopedDecoratorData(context, $apiVersion, apiVersionKey, target, value ?? true, scope);
};
function getIsApiVersion(context, param) {
    return getScopedDecoratorData(context, apiVersionKey, param);
}
const $clientNamespace = (context, entity, value, scope) => {
    if (value.trim() === "") {
        reportDiagnostic(context.program, {
            code: "empty-client-namespace",
            format: {},
            target: entity,
        });
    }
    setScopedDecoratorData(context, $clientNamespace, clientNamespaceKey, entity, value, scope);
};
function getClientNamespace(context, entity) {
    const override = getScopedDecoratorData(context, clientNamespaceKey, entity);
    if (override)
        return override;
    if (!entity.namespace) {
        return "";
    }
    if (entity.kind === "Namespace") {
        return getNamespaceFullNameWithOverride(context, entity);
    }
    return getNamespaceFullNameWithOverride(context, entity.namespace);
}
function getNamespaceFullNameWithOverride(context, namespace) {
    const segments = [];
    let current = namespace;
    while (current && current.name !== "") {
        const override = getScopedDecoratorData(context, clientNamespaceKey, current);
        if (override) {
            segments.unshift(override);
            break;
        }
        segments.unshift(current.name);
        current = current.namespace;
    }
    return segments.join(".");
}
const $scope = (context, entity, scope) => {
    const [negationScopes, scopes] = parseScopes(context, scope);
    if (negationScopes !== undefined && negationScopes.length > 0) {
        // for negation scope, override the previous value
        setScopedDecoratorData(context, $scope, negationScopesKey, entity, negationScopes);
    }
    if (scopes !== undefined && scopes.length > 0) {
        // for normal scope, add them incrementally
        const targetEntry = context.program.stateMap(scopeKey).get(entity);
        setScopedDecoratorData(context, $scope, scopeKey, entity, !targetEntry ? scopes : [...Object.values(targetEntry), ...scopes]);
    }
};
function IsInScope(context, entity) {
    const scopes = getScopedDecoratorData(context, scopeKey, entity);
    if (scopes !== undefined && scopes.includes(context.emitterName)) {
        return true;
    }
    const negationScopes = getScopedDecoratorData(context, negationScopesKey, entity);
    if (negationScopes !== undefined && negationScopes.includes(context.emitterName)) {
        return false;
    }
    return true;
}

const AllScopes = Symbol.for("@azure-core/typespec-client-generator-core/all-scopes");
const clientNameKey = createStateSymbol("clientName");
const clientNamespaceKey = createStateSymbol("clientNamespace");
const negationScopesKey = createStateSymbol("negationScopes");
const scopeKey = createStateSymbol("scope");
/**
 *
 * @param emitterName Full emitter name
 * @returns The language of the emitter. I.e. "@azure-tools/typespec-csharp" will return "csharp"
 */
function parseEmitterName(program, emitterName) {
    const diagnostics = createDiagnosticCollector();
    if (!emitterName) {
        diagnostics.add(createDiagnostic({
            code: "no-emitter-name",
            format: {},
            target: program.getGlobalNamespaceType(),
        }));
        return diagnostics.wrap("none");
    }
    const regex = /(?:cadl|typespec|client|server)-([^\\/-]*)/;
    const match = emitterName.match(regex);
    if (!match || match.length < 2)
        return diagnostics.wrap("none");
    const language = match[1];
    if (["typescript", "ts"].includes(language))
        return diagnostics.wrap("javascript");
    return diagnostics.wrap(language);
}
/**
 *
 * @param context
 * @param namespace If we know explicitly the namespace of the client, pass this in
 * @returns The name of the namespace
 */
function getClientNamespaceStringHelper(context, namespace) {
    let packageName = context.packageName;
    if (packageName) {
        packageName = packageName
            .replace(/-/g, ".")
            .replace(/\.([a-z])?/g, (match) => match.toUpperCase());
        return packageName.charAt(0).toUpperCase() + packageName.slice(1);
    }
    if (namespace) {
        return getNamespaceFullName(namespace);
    }
    return undefined;
}
/**
 *
 * @param context
 * @param type The type that we are adding api version information onto
 * @returns Whether the type is the api version parameter and the default value for the client
 */
function updateWithApiVersionInformation(context, type, namespace) {
    const isApiVersionParam = isApiVersion(context, type);
    return {
        isApiVersionParam,
        clientDefaultValue: isApiVersionParam && namespace
            ? context.__clientToApiVersionClientDefaultValue.get(namespace)
            : undefined,
    };
}
function filterApiVersionsWithDecorators(context, type, apiVersions) {
    const addedOnVersions = getAddedOnVersions(context.program, type)?.map((x) => x.value) ?? [];
    const removedOnVersions = getRemovedOnVersions(context.program, type)?.map((x) => x.value) ?? [];
    let added = addedOnVersions.length ? false : true;
    let addedCounter = 0;
    let removeCounter = 0;
    const retval = [];
    for (let i = 0; i < apiVersions.length; i++) {
        const version = apiVersions[i];
        if (addedCounter < addedOnVersions.length && version === addedOnVersions[addedCounter]) {
            added = true;
            addedCounter++;
        }
        if (removeCounter < removedOnVersions.length && version === removedOnVersions[removeCounter]) {
            added = false;
            removeCounter++;
        }
        if (added) {
            // only add version smaller than config
            if (context.apiVersion === undefined ||
                context.apiVersion === "latest" ||
                context.apiVersion === "all" ||
                apiVersions.indexOf(context.apiVersion) >= i) {
                retval.push(version);
            }
        }
    }
    return retval;
}
function sortAndRemoveDuplicates(a, b, apiVersions) {
    const union = Array.from(new Set([...a, ...b]));
    return apiVersions.filter((item) => union.includes(item));
}
/**
 *
 * @param context
 * @param type
 * @param client If it's associated with a client, meaning it's a param etc, we can see if it's available on that client
 * @returns All api versions the type is available on
 */
function getAvailableApiVersions(context, type, wrapper) {
    let wrapperApiVersions = [];
    if (wrapper) {
        wrapperApiVersions = context.__tspTypeToApiVersions.get(wrapper) || [];
    }
    const allApiVersions = getVersions(context.program, type)[1]
        ?.getVersions()
        .map((x) => x.value) || [];
    const apiVersions = wrapperApiVersions.length ? wrapperApiVersions : allApiVersions;
    if (!apiVersions)
        return [];
    const explicitlyDecorated = filterApiVersionsWithDecorators(context, type, apiVersions);
    if (explicitlyDecorated.length) {
        context.__tspTypeToApiVersions.set(type, explicitlyDecorated);
        return explicitlyDecorated;
    }
    // we take the union of all of the api versions that the type is available on
    // if it's called multiple times with diff wrappers, we want to make sure we have
    // all of the possible api versions listed
    const existing = context.__tspTypeToApiVersions.get(type) || [];
    const retval = sortAndRemoveDuplicates(wrapperApiVersions, existing, allApiVersions);
    context.__tspTypeToApiVersions.set(type, retval);
    return retval;
}
/**
 *
 * @param type
 * @returns A unique id for each type so we can do set comparisons
 */
function getHashForType(type) {
    if (type.kind === "array" || type.kind === "dict") {
        return `${type.kind}[${getHashForType(type.valueType)}]`;
    }
    if (type.kind === "enum" || type.kind === "model" || type.kind === "enumvalue")
        return type.name;
    if (type.kind === "union") {
        return type.variantTypes.map((x) => getHashForType(x)).join("|");
    }
    return type.kind;
}
/**
 * Helper function to return default values for encode etc
 * @param type
 */
function getSdkTypeBaseHelper(context, type, kind) {
    const diagnostics = createDiagnosticCollector();
    return diagnostics.wrap({
        __raw: type,
        deprecation: getDeprecationDetails(context.program, type)?.message,
        kind,
        decorators: diagnostics.pipe(getTypeDecorators(context, type)),
    });
}
function getNamespacePrefix(namespace) {
    return namespace ? getNamespaceFullName(namespace) + "." : "";
}
function getTypeDecorators(context, type) {
    const diagnostics = createDiagnosticCollector();
    const retval = [];
    if ("decorators" in type) {
        for (const decorator of type.decorators) {
            // only process explicitly defined decorators
            if (decorator.definition) {
                const decoratorName = `${getNamespacePrefix(decorator.definition?.namespace)}${decorator.definition?.name}`;
                // white list filtering
                if (!context.decoratorsAllowList ||
                    !context.decoratorsAllowList.some((x) => new RegExp(x).test(decoratorName))) {
                    continue;
                }
                const decoratorInfo = {
                    name: decoratorName,
                    arguments: {},
                };
                for (let i = 0; i < decorator.args.length; i++) {
                    decoratorInfo.arguments[decorator.definition.parameters[i].name] = diagnostics.pipe(getDecoratorArgValue(context, decorator.args[i].jsValue, type, decoratorName));
                }
                retval.push(decoratorInfo);
            }
        }
    }
    return diagnostics.wrap(retval);
}
function getDecoratorArgValue(context, arg, type, decoratorName) {
    const diagnostics = createDiagnosticCollector();
    if (typeof arg === "object" && arg !== null && "kind" in arg) {
        if (arg.kind === "EnumMember") {
            return diagnostics.wrap(diagnostics.pipe(getClientTypeWithDiagnostics(context, arg)));
        }
        if (arg.kind === "String" || arg.kind === "Number" || arg.kind === "Boolean") {
            return diagnostics.wrap(arg.value);
        }
        diagnostics.add(createDiagnostic({
            code: "unsupported-generic-decorator-arg-type",
            target: type,
            format: { decoratorName },
        }));
        return diagnostics.wrap(undefined);
    }
    return diagnostics.wrap(arg);
}
function intOrFloat(value) {
    return value.toString().indexOf(".") === -1 ? "int32" : "float32";
}
/**
 * Whether a model or enum or union as enum is in Azure.Core[.Foundations] namespace
 * @param t
 * @returns
 */
function isAzureCoreTspModel(t) {
    return ((t.kind === "Model" || t.kind === "Enum" || t.kind === "Union") &&
        t.namespace !== undefined &&
        ["Azure.Core", "Azure.Core.Foundations"].includes(getNamespaceFullName(t.namespace)));
}
function isAcceptHeader(param) {
    return param.kind === "header" && param.serializedName.toLowerCase() === "accept";
}
function isContentTypeHeader(param) {
    return param.kind === "header" && param.serializedName.toLowerCase() === "content-type";
}
function isMultipartOperation(context, operation) {
    if (!operation)
        return false;
    const httpOperation = getHttpOperationWithCache(context, operation);
    const httpBody = httpOperation.parameters.body;
    if (httpBody && httpBody.type.kind === "Model") {
        return httpBody.contentTypes.some((x) => x.startsWith("multipart/"));
    }
    return false;
}
function getNonNullOptions(type) {
    return [...type.variants.values()].map((x) => x.type).filter((t) => !isNullType(t));
}
function getNullOption(type) {
    return [...type.variants.values()].map((x) => x.type).filter((t) => isNullType(t))[0];
}
function getAllResponseBodiesAndNonBodyExists(responses) {
    const allResponseBodies = [];
    let nonBodyExists = false;
    for (const response of responses) {
        if (response.type) {
            if (response.type.kind === "nullable") {
                nonBodyExists = true;
            }
            allResponseBodies.push(response.type);
        }
        else {
            nonBodyExists = true;
        }
    }
    return { allResponseBodies, nonBodyExists };
}
/**
 * Use this if you are trying to create a generated name for something without an original TypeSpec type.
 *
 * Otherwise, you should use the `getGeneratedName` function.
 * @param context
 */
function createGeneratedName(context, type, suffix) {
    return `${getCrossLanguageDefinitionId(context, type).split(".").at(-1)}${suffix}`;
}
function isSubscriptionId(context, parameter) {
    return Boolean(context.arm) && parameter.name === "subscriptionId";
}
function getLocationOfOperation(operation) {
    // have to check interface first, because interfaces are more granular than namespaces
    return (operation.interface || operation.namespace);
}
function isNeverOrVoidType(type) {
    return isNeverType(type) || isVoidType(type);
}
function getValidApiVersion(context, versions) {
    let apiVersion = context.apiVersion;
    if (apiVersion === "all") {
        return apiVersion;
    }
    if (apiVersion === "latest" || apiVersion === undefined || !versions.includes(apiVersion)) {
        apiVersion = versions[versions.length - 1];
    }
    return apiVersion;
}
function getHttpOperationResponseHeaders(response) {
    const headers = response.headers ? Object.values(response.headers) : [];
    if (response.body?.contentTypeProperty) {
        headers.push(response.body.contentTypeProperty);
    }
    return headers;
}
function removeVersionsLargerThanExplicitlySpecified(context, versions) {
    // filter with specific api version
    if (context.apiVersion !== undefined &&
        context.apiVersion !== "latest" &&
        context.apiVersion !== "all") {
        const index = versions.findIndex((version) => version.value === context.apiVersion);
        if (index >= 0) {
            versions.splice(index + 1, versions.length - index - 1);
        }
    }
}
function filterApiVersionsInEnum(context, client, sdkVersionsEnum) {
    // if they explicitly set an api version, remove larger versions
    removeVersionsLargerThanExplicitlySpecified(context, sdkVersionsEnum.values);
    const defaultApiVersion = getDefaultApiVersion(context, client.service);
    if (!context.previewStringRegex.test(defaultApiVersion?.value || "")) {
        sdkVersionsEnum.values = sdkVersionsEnum.values.filter((v) => typeof v.value === "string" && !context.previewStringRegex.test(v.value));
    }
}
function isJsonContentType(contentType) {
    const regex = new RegExp(/^(application|text)\/(.+\+)?json$/);
    return regex.test(contentType);
}
function isXmlContentType(contentType) {
    const regex = new RegExp(/^(application|text)\/(.+\+)?xml$/);
    return regex.test(contentType);
}
function twoParamsEquivalent(context, param1, param2) {
    if (!param1 || !param2) {
        return false;
    }
    return (param1.name === param2.name ||
        getParamAlias(context, param1) === param2.name ||
        param1.name === getParamAlias(context, param2));
}
/**
 * If body is from spread, then it does not directly from a model property.
 * @param httpBody
 * @param parameters
 * @returns
 */
function isHttpBodySpread(httpBody) {
    return httpBody.property === undefined;
}
/**
 * If body is from simple spread, then we use the original model as body model.
 * @param type
 * @returns
 */
function getHttpBodySpreadModel(type) {
    if (type.sourceModels.length === 1 && type.sourceModels[0].usage === "spread") {
        const innerModel = type.sourceModels[0].model;
        // for case: `op test(...Model):void;`
        if (innerModel.name !== "" && innerModel.properties.size === type.properties.size) {
            return innerModel;
        }
        // for case: `op test(@header h: string, @query q: string, ...Model): void;`
        if (innerModel.sourceModels.length === 1 &&
            innerModel.sourceModels[0].usage === "spread" &&
            innerModel.sourceModels[0].model.name !== "" &&
            innerModel.sourceModels[0].model.properties.size === type.properties.size) {
            return innerModel.sourceModels[0].model;
        }
    }
    return type;
}
function isOnClient(context, type, operation, versioning) {
    const namespace = operation ? getLocationOfOperation(operation) : type.model?.namespace;
    return (isSubscriptionId(context, type) ||
        (isApiVersion(context, type) && versioning) ||
        Boolean(namespace &&
            context.__clientToParameters
                .get(namespace)
                ?.find((x) => twoParamsEquivalent(context, x.__raw, type))));
}
function getValueTypeValue(value) {
    switch (value.valueKind) {
        case "ArrayValue":
            return value.values.map((x) => getValueTypeValue(x));
        case "BooleanValue":
        case "StringValue":
        case "NullValue":
            return value.value;
        case "NumericValue":
            return value.value.asNumber();
        case "EnumValue":
            return value.value.value ?? value.value.name;
        case "ObjectValue":
            return Object.fromEntries([...value.properties.keys()].map((x) => [
                x,
                getValueTypeValue(value.properties.get(x).value),
            ]));
        case "ScalarValue":
            // TODO: handle scalar value
            return undefined;
    }
}

async function checkExamplesDirExists(host, dir) {
    try {
        return (await host.stat(dir)).isDirectory();
    }
    catch (err) {
        return false;
    }
}
/**
 * Load all examples for a client
 *
 * @param context
 * @param apiVersion
 * @returns a map of all operations' examples, key is operation's operation id,
 * value is a map of examples, key is example's title, value is example's details
 */
async function loadExamples(context, apiVersion) {
    const diagnostics = createDiagnosticCollector();
    const examplesBaseDir = context.examplesDir ?? resolvePath(context.program.projectRoot, "examples");
    const exampleDir = apiVersion
        ? resolvePath(examplesBaseDir, apiVersion)
        : resolvePath(examplesBaseDir);
    if (!(await checkExamplesDirExists(context.program.host, exampleDir))) {
        if (context.examplesDir) {
            diagnostics.add(createDiagnostic({
                code: "example-loading",
                messageId: "noDirectory",
                format: { directory: exampleDir },
                target: NoTarget,
            }));
        }
        return diagnostics.wrap(new Map());
    }
    const map = new Map();
    const exampleFiles = await context.program.host.readDir(exampleDir);
    for (const fileName of exampleFiles) {
        try {
            const exampleFile = await context.program.host.readFile(resolvePath(exampleDir, fileName));
            const example = JSON.parse(exampleFile.text);
            if (!example.operationId || !example.title) {
                diagnostics.add(createDiagnostic({
                    code: "example-loading",
                    messageId: "noOperationId",
                    format: { filename: fileName },
                    target: NoTarget,
                }));
                continue;
            }
            if (!map.has(example.operationId.toLowerCase())) {
                map.set(example.operationId.toLowerCase(), {});
            }
            const examples = map.get(example.operationId.toLowerCase());
            if (example.title in examples) {
                diagnostics.add(createDiagnostic({
                    code: "duplicate-example-file",
                    target: NoTarget,
                    format: {
                        filename: fileName,
                        operationId: example.operationId,
                        title: example.title,
                    },
                }));
            }
            examples[example.title] = {
                relativePath: apiVersion ? resolvePath(apiVersion, fileName) : fileName,
                data: example,
            };
        }
        catch (err) {
            diagnostics.add(createDiagnostic({
                code: "example-loading",
                messageId: "default",
                format: { filename: fileName, error: err?.toString() ?? "" },
                target: NoTarget,
            }));
        }
    }
    return diagnostics.wrap(map);
}
function resolveOperationId(context, operation, honorRenaming) {
    const { program } = context;
    // if @operationId was specified use that value
    const explicitOperationId = getOperationId(program, operation);
    if (explicitOperationId) {
        return explicitOperationId;
    }
    const operationName = honorRenaming ? getLibraryName(context, operation) : operation.name;
    if (operation.interface) {
        return `${honorRenaming ? getLibraryName(context, operation.interface) : operation.interface.name}_${operationName}`;
    }
    const namespace = operation.namespace;
    if (namespace === undefined ||
        isGlobalNamespace(program, namespace) ||
        isService(program, namespace)) {
        return operationName;
    }
    return `${honorRenaming ? getLibraryName(context, namespace) : namespace.name}_${operationName}`;
}
async function handleClientExamples(context, client) {
    const diagnostics = createDiagnosticCollector();
    const examples = diagnostics.pipe(await loadExamples(context, getValidApiVersion(context, client.apiVersions)));
    const methodQueue = [...client.methods];
    while (methodQueue.length > 0) {
        const method = methodQueue.pop();
        if (method.kind === "clientaccessor") {
            methodQueue.push(...method.response.methods);
        }
        else {
            // since operation could have customization in client.tsp, we need to handle all the original operation (exclude the templated operation)
            let operation = method.__raw;
            while (operation && operation.templateMapper === undefined) {
                // try operation id with renaming
                let operationId = resolveOperationId(context, operation, true).toLowerCase();
                if (examples.has(operationId)) {
                    diagnostics.pipe(handleMethodExamples(context, method, examples.get(operationId)));
                    break;
                }
                // try operation id without renaming
                operationId = resolveOperationId(context, operation, false).toLowerCase();
                if (examples.has(operationId)) {
                    diagnostics.pipe(handleMethodExamples(context, method, examples.get(operationId)));
                    break;
                }
                operation = operation.sourceOperation;
            }
        }
    }
    return diagnostics.wrap(undefined);
}
function handleMethodExamples(context, method, examples) {
    const diagnostics = createDiagnosticCollector();
    if (method.operation.kind === "http") {
        diagnostics.pipe(handleHttpOperationExamples(method.operation, examples));
        if (method.operation.examples) {
            if (!context.__httpOperationExamples) {
                context.__httpOperationExamples = new Map();
            }
            context.__httpOperationExamples.set(method.operation.__raw, method.operation.examples);
        }
    }
    return diagnostics.wrap(undefined);
}
function handleHttpOperationExamples(operation, examples) {
    const diagnostics = createDiagnosticCollector();
    operation.examples = [];
    for (const [title, example] of Object.entries(examples)) {
        const operationExample = {
            kind: "http",
            name: title,
            description: title,
            doc: title,
            filePath: example.relativePath,
            parameters: diagnostics.pipe(handleHttpParameters(operation.bodyParam
                ? [...operation.parameters, operation.bodyParam]
                : operation.parameters, example.data, example.relativePath)),
            responses: diagnostics.pipe(handleHttpResponses(operation.responses, example.data, example.relativePath)),
            rawExample: example.data,
        };
        operation.examples.push(operationExample);
    }
    // sort examples by file path
    operation.examples.sort((a, b) => (a.filePath > b.filePath ? 1 : -1));
    return diagnostics.wrap(undefined);
}
function handleHttpParameters(parameters, example, relativePath) {
    const diagnostics = createDiagnosticCollector();
    const parameterExamples = [];
    if ("parameters" in example &&
        typeof example.parameters === "object" &&
        example.parameters !== null) {
        for (const name of Object.keys(example.parameters)) {
            let parameter = parameters.find((p) => p.serializedName === name);
            // fallback to body in example for any body parameter
            if (!parameter && name === "body") {
                parameter = parameters.find((p) => p.kind === "body");
            }
            if (parameter) {
                const value = diagnostics.pipe(getSdkTypeExample(parameter.type, example.parameters[name], relativePath));
                if (value) {
                    parameterExamples.push({
                        parameter,
                        value,
                    });
                }
            }
            else {
                addExampleValueNoMappingDignostic(diagnostics, { [name]: example.parameters[name] }, relativePath);
            }
        }
    }
    return diagnostics.wrap(parameterExamples);
}
function handleHttpResponses(responses, example, relativePath) {
    const diagnostics = createDiagnosticCollector();
    const responseExamples = [];
    if ("responses" in example &&
        typeof example.responses === "object" &&
        example.responses !== null) {
        for (const code of Object.keys(example.responses)) {
            const statusCode = parseInt(code, 10);
            let found = false;
            for (const response of responses) {
                const responseCode = response.statusCodes;
                if (responseCode === statusCode) {
                    responseExamples.push(diagnostics.pipe(handleHttpResponse(response, statusCode, example.responses[code], relativePath)));
                    found = true;
                    break;
                }
                else if (typeof responseCode === "object" &&
                    responseCode !== null &&
                    responseCode.start <= statusCode &&
                    responseCode.end >= statusCode) {
                    responseExamples.push(diagnostics.pipe(handleHttpResponse(response, statusCode, example.responses[code], relativePath)));
                    found = true;
                    break;
                }
            }
            if (!found) {
                addExampleValueNoMappingDignostic(diagnostics, { [code]: example.responses[code] }, relativePath);
            }
        }
    }
    return diagnostics.wrap(responseExamples);
}
function handleHttpResponse(response, statusCode, example, relativePath) {
    const diagnostics = createDiagnosticCollector();
    const responseExample = {
        response,
        statusCode,
        headers: [],
    };
    if (typeof example === "object" && example !== null) {
        for (const name of Object.keys(example)) {
            if (name === "description") {
                continue;
            }
            else if (name === "body") {
                if (response.type) {
                    responseExample.bodyValue = diagnostics.pipe(getSdkTypeExample(response.type, example.body, relativePath));
                }
                else {
                    addExampleValueNoMappingDignostic(diagnostics, { body: example.body }, relativePath);
                }
            }
            else if (name === "headers") {
                for (const subName of Object.keys(example.headers)) {
                    const header = response.headers.find((p) => p.serializedName === subName);
                    if (header) {
                        const value = diagnostics.pipe(getSdkTypeExample(header.type, example[name][subName], relativePath));
                        if (value) {
                            responseExample.headers.push({
                                header,
                                value,
                            });
                        }
                    }
                    else {
                        addExampleValueNoMappingDignostic(diagnostics, { [subName]: example[name][subName] }, relativePath);
                    }
                }
            }
            else {
                addExampleValueNoMappingDignostic(diagnostics, { [name]: example[name] }, relativePath);
            }
        }
    }
    return diagnostics.wrap(responseExample);
}
function getSdkTypeExample(type, example, relativePath) {
    const diagnostics = createDiagnosticCollector();
    if (isSdkIntKind(type.kind) || isSdkFloatKind(type.kind)) {
        return getSdkBaseTypeExample("number", type, example, relativePath);
    }
    else {
        switch (type.kind) {
            case "string":
            case "bytes":
                return getSdkBaseTypeExample("string", type, example, relativePath);
            case "boolean":
                return getSdkBaseTypeExample("boolean", type, example, relativePath);
            case "url":
            case "plainDate":
            case "plainTime":
                return getSdkBaseTypeExample("string", type, example, relativePath);
            case "nullable":
                if (example === null) {
                    return diagnostics.wrap({
                        kind: "null",
                        type,
                        value: null,
                    });
                }
                else {
                    return getSdkTypeExample(type.type, example, relativePath);
                }
            case "unknown":
                return diagnostics.wrap({
                    kind: "unknown",
                    type,
                    value: example,
                });
            case "constant":
                if (example === type.value) {
                    return getSdkBaseTypeExample(typeof type.value, type, example, relativePath);
                }
                else {
                    addExampleValueNoMappingDignostic(diagnostics, example, relativePath);
                    return diagnostics.wrap(undefined);
                }
            case "enum":
                if (type.values.some((v) => v.value === example)) {
                    return getSdkBaseTypeExample(typeof example, type, example, relativePath);
                }
                else {
                    addExampleValueNoMappingDignostic(diagnostics, example, relativePath);
                    return diagnostics.wrap(undefined);
                }
            case "enumvalue":
                if (type.value === example) {
                    return getSdkBaseTypeExample(typeof example, type, example, relativePath);
                }
                else {
                    addExampleValueNoMappingDignostic(diagnostics, example, relativePath);
                    return diagnostics.wrap(undefined);
                }
            case "utcDateTime":
            case "offsetDateTime":
            case "duration":
                const inner = diagnostics.pipe(getSdkTypeExample(type.wireType, example, relativePath));
                if (inner) {
                    inner.type = type;
                }
                return diagnostics.wrap(inner);
            case "union":
                return diagnostics.wrap({
                    kind: "union",
                    type,
                    value: example,
                });
            case "array":
                return getSdkArrayExample(type, example, relativePath);
            case "dict":
                return getSdkDictionaryExample(type, example, relativePath);
            case "model":
                return getSdkModelExample(type, example, relativePath);
        }
    }
    return diagnostics.wrap(undefined);
}
function getSdkBaseTypeExample(kind, type, example, relativePath) {
    const diagnostics = createDiagnosticCollector();
    if (typeof example === kind) {
        return diagnostics.wrap({
            kind,
            type,
            value: example,
        });
    }
    else {
        addExampleValueNoMappingDignostic(diagnostics, example, relativePath);
    }
    return diagnostics.wrap(undefined);
}
function getSdkArrayExample(type, example, relativePath) {
    const diagnostics = createDiagnosticCollector();
    if (Array.isArray(example)) {
        const arrayExample = [];
        for (const item of example) {
            const result = diagnostics.pipe(getSdkTypeExample(type.valueType, item, relativePath));
            if (result) {
                arrayExample.push(result);
            }
        }
        return diagnostics.wrap({
            kind: "array",
            type,
            value: arrayExample,
        });
    }
    else {
        addExampleValueNoMappingDignostic(diagnostics, example, relativePath);
        return diagnostics.wrap(undefined);
    }
}
function getSdkDictionaryExample(type, example, relativePath) {
    const diagnostics = createDiagnosticCollector();
    if (typeof example === "object") {
        if (example === null) {
            return diagnostics.wrap(undefined);
        }
        const dictionaryExample = {};
        for (const key of Object.keys(example)) {
            const result = diagnostics.pipe(getSdkTypeExample(type.valueType, example[key], relativePath));
            if (result) {
                dictionaryExample[key] = result;
            }
        }
        return diagnostics.wrap({
            kind: "dict",
            type,
            value: dictionaryExample,
        });
    }
    else {
        addExampleValueNoMappingDignostic(diagnostics, example, relativePath);
        return diagnostics.wrap(undefined);
    }
}
function getSdkModelExample(type, example, relativePath) {
    const diagnostics = createDiagnosticCollector();
    if (typeof example === "object") {
        if (example === null) {
            return diagnostics.wrap(undefined);
        }
        // handle discriminated model
        if (type.discriminatorProperty) {
            if (type.discriminatorProperty.name in example &&
                example[type.discriminatorProperty.name] in type.discriminatedSubtypes) {
                return getSdkModelExample(type.discriminatedSubtypes[example[type.discriminatorProperty.name]], example, relativePath);
            }
            else {
                addExampleValueNoMappingDignostic(diagnostics, example, relativePath);
                return diagnostics.wrap(undefined);
            }
        }
        let additionalPropertiesType;
        const additionalProperties = new Map();
        const additionalPropertiesExample = {};
        const properties = new Map();
        const propertiesExample = {};
        // get all properties type and additional properties type if exist
        const modelQueue = [type];
        while (modelQueue.length > 0) {
            const model = modelQueue.pop();
            for (let property of model.properties) {
                property = property;
                if (!properties.has(property.serializedName)) {
                    properties.set(property.serializedName, property);
                }
            }
            if (model.additionalProperties && additionalPropertiesType === undefined) {
                additionalPropertiesType = model.additionalProperties;
            }
            if (model.baseModel) {
                modelQueue.push(model.baseModel);
            }
        }
        for (const name of Object.keys(example)) {
            const property = properties.get(name);
            if (property) {
                const result = diagnostics.pipe(getSdkTypeExample(property.type, example[name], relativePath));
                if (result) {
                    propertiesExample[name] = result;
                }
            }
            else {
                additionalProperties[name] = example[name];
            }
        }
        // handle additional properties
        if (Object.keys(additionalProperties).length > 0) {
            if (additionalPropertiesType) {
                for (const [name, value] of Object.entries(additionalProperties)) {
                    const result = diagnostics.pipe(getSdkTypeExample(additionalPropertiesType, value, relativePath));
                    if (result) {
                        additionalPropertiesExample[name] = result;
                    }
                }
            }
            else {
                addExampleValueNoMappingDignostic(diagnostics, additionalProperties, relativePath);
            }
        }
        return diagnostics.wrap({
            kind: "model",
            type,
            value: propertiesExample,
            additionalPropertiesValue: Object.keys(additionalPropertiesExample).length > 0
                ? additionalPropertiesExample
                : undefined,
        });
    }
    else {
        addExampleValueNoMappingDignostic(diagnostics, example, relativePath);
        return diagnostics.wrap(undefined);
    }
}
function addExampleValueNoMappingDignostic(diagnostics, value, relativePath) {
    diagnostics.add(createDiagnostic({
        code: "example-value-no-mapping",
        target: NoTarget,
        format: {
            value: JSON.stringify(value),
            relativePath,
        },
    }));
}

function getSdkServiceOperation(context, operation, methodParameters) {
    const diagnostics = createDiagnosticCollector();
    const httpOperation = getHttpOperationWithCache(context, operation);
    if (httpOperation) {
        const sdkHttpOperation = diagnostics.pipe(getSdkHttpOperation(context, httpOperation, methodParameters));
        return diagnostics.wrap(sdkHttpOperation);
    }
    diagnostics.add(createDiagnostic({
        code: "unsupported-protocol",
        target: operation,
        format: {},
    }));
    return diagnostics.wrap(undefined);
}
function getSdkLroPagingServiceMethod(context, operation, client) {
    const diagnostics = createDiagnosticCollector();
    return diagnostics.wrap({
        ...diagnostics.pipe(getSdkLroServiceMethod(context, operation, client)),
        ...diagnostics.pipe(getSdkPagingServiceMethod(context, operation, client)),
        kind: "lropaging",
    });
}
function getSdkPagingServiceMethod(context, operation, client) {
    const diagnostics = createDiagnosticCollector();
    const basic = diagnostics.pipe(getSdkBasicServiceMethod(context, operation, client));
    // nullable response type means the underlaying operation has multiple responses and only one of them is not empty, which is what we want
    let responseType = basic.response.type;
    if (responseType?.kind === "nullable") {
        responseType = responseType.type;
    }
    // normal paging
    if (isList(context.program, operation)) {
        const pagingOperation = diagnostics.pipe(getPagingOperation(context.program, operation));
        if (responseType?.__raw?.kind !== "Model" || !pagingOperation) {
            diagnostics.add(createDiagnostic({
                code: "unexpected-pageable-operation-return-type",
                target: operation,
                format: {
                    operationName: operation.name,
                },
            }));
            // return as page method with no paging info
            return diagnostics.wrap({
                ...basic,
                kind: "paging",
            });
        }
        basic.response.resultPath = getPropertyPathFromModel(context, responseType?.__raw, (p) => p === pagingOperation.output.pageItems.property);
        const nextLinkPath = pagingOperation.output.nextLink
            ? getPropertyPathFromModel(context, responseType?.__raw, (p) => p === pagingOperation.output.nextLink.property)
            : undefined;
        context.__pagedResultSet.add(responseType);
        // tcgc will let all paging method return a list of items
        basic.response.type = diagnostics.pipe(getClientTypeWithDiagnostics(context, pagingOperation?.output.pageItems.property.type));
        return diagnostics.wrap({
            ...basic,
            kind: "paging",
            nextLinkPath,
        });
    }
    // azure core paging
    const pagedMetadata = getPagedResult(context.program, operation);
    if (responseType?.__raw?.kind !== "Model" || !pagedMetadata.itemsProperty) {
        diagnostics.add(createDiagnostic({
            code: "unexpected-pageable-operation-return-type",
            target: operation,
            format: {
                operationName: operation.name,
            },
        }));
        // return as page method with no paging info
        return diagnostics.wrap({
            ...basic,
            kind: "paging",
        });
    }
    context.__pagedResultSet.add(responseType);
    // tcgc will let all paging method return a list of items
    basic.response.type = diagnostics.pipe(getClientTypeWithDiagnostics(context, pagedMetadata.itemsProperty.type));
    basic.response.resultPath = getPropertyPathFromSegment(context, pagedMetadata.modelType, pagedMetadata.itemsSegments);
    return diagnostics.wrap({
        ...basic,
        __raw_paged_metadata: pagedMetadata,
        kind: "paging",
        nextLinkPath: getPropertyPathFromSegment(context, pagedMetadata.modelType, pagedMetadata?.nextLinkSegments),
        nextLinkOperation: pagedMetadata?.nextLinkOperation
            ? diagnostics.pipe(getSdkServiceOperation(context, pagedMetadata.nextLinkOperation, basic.parameters))
            : undefined,
    });
}
function getPropertyPathFromModel(context, model, predicate) {
    const queue = [];
    for (const prop of model.properties.values()) {
        if (predicate(prop)) {
            return getLibraryName(context, prop);
        }
        if (prop.type.kind === "Model") {
            queue.push({ model: prop.type, path: [prop] });
        }
    }
    while (queue.length > 0) {
        const { model, path } = queue.shift();
        for (const prop of model.properties.values()) {
            if (predicate(prop)) {
                return path
                    .concat(prop)
                    .map((s) => getLibraryName(context, s))
                    .join(".");
            }
            if (prop.type.kind === "Model") {
                queue.push({ model: prop.type, path: path.concat(prop) });
            }
        }
    }
    return undefined;
}
function getPropertyPathFromSegment(context, type, segments) {
    if (!segments || segments.length === 0) {
        return "";
    }
    const wireSegments = [];
    let current = type;
    for (const segment of segments) {
        const property = current.properties.get(segment);
        if (!property) {
            if (current.baseModel) {
                return getPropertyPathFromSegment(context, current.baseModel, segments);
            }
            return "";
        }
        wireSegments.push(getLibraryName(context, property));
        current = property.type;
    }
    return wireSegments.join(".");
}
function getSdkLroServiceMethod(context, operation, client) {
    const diagnostics = createDiagnosticCollector();
    const metadata = getServiceMethodLroMetadata(context, operation);
    const basicServiceMethod = diagnostics.pipe(getSdkBasicServiceMethod(context, operation, client));
    basicServiceMethod.response.type = metadata.finalResponse?.result;
    basicServiceMethod.response.resultPath = metadata.finalResponse?.resultPath;
    return diagnostics.wrap({
        ...basicServiceMethod,
        kind: "lro",
        __raw_lro_metadata: metadata.__raw,
        lroMetadata: metadata,
        operation: diagnostics.pipe(getSdkServiceOperation(context, metadata.__raw.operation, basicServiceMethod.parameters)),
    });
}
function getServiceMethodLroMetadata(context, operation) {
    const rawMetadata = getLroMetadata(context.program, operation);
    if (rawMetadata === undefined) {
        return undefined;
    }
    const diagnostics = createDiagnosticCollector();
    return {
        __raw: rawMetadata,
        finalStateVia: rawMetadata.finalStateVia,
        finalResponse: rawMetadata.finalEnvelopeResult !== undefined && rawMetadata.finalEnvelopeResult !== "void"
            ? {
                envelopeResult: diagnostics.pipe(getClientTypeWithDiagnostics(context, rawMetadata.finalEnvelopeResult)),
                result: diagnostics.pipe(getClientTypeWithDiagnostics(context, rawMetadata.finalResult)),
                resultPath: rawMetadata.finalResultPath,
            }
            : undefined,
        finalStep: rawMetadata.finalStep !== undefined ? { kind: rawMetadata.finalStep.kind } : undefined,
        pollingStep: {
            responseBody: diagnostics.pipe(getClientTypeWithDiagnostics(context, rawMetadata.pollingInfo.responseModel)),
        },
    };
}
function getSdkMethodResponse(context, operation, sdkOperation, client) {
    const responses = sdkOperation.responses;
    // TODO: put head as bool here
    const { allResponseBodies, nonBodyExists } = getAllResponseBodiesAndNonBodyExists(responses);
    const responseTypes = new Set(allResponseBodies.map((x) => getHashForType(x)));
    let type = undefined;
    if (responseTypes.size > 1) {
        // return union of all the different types
        type = {
            __raw: operation,
            kind: "union",
            access: "public",
            usage: UsageFlags.Output,
            variantTypes: allResponseBodies,
            name: createGeneratedName(context, operation, "UnionResponse"),
            isGeneratedName: true,
            clientNamespace: client.clientNamespace,
            crossLanguageDefinitionId: `${getCrossLanguageDefinitionId(context, operation)}.UnionResponse`,
            decorators: [],
        };
    }
    else if (responseTypes.size === 1) {
        type = allResponseBodies[0];
    }
    if (nonBodyExists && type) {
        type = {
            kind: "nullable",
            type: type,
            decorators: [],
            access: "public",
            usage: UsageFlags.Output,
            clientNamespace: client.clientNamespace,
        };
    }
    return {
        kind: "method",
        type,
    };
}
function getSdkBasicServiceMethod(context, operation, client) {
    const diagnostics = createDiagnosticCollector();
    const methodParameters = [];
    // we have to calculate apiVersions first, so that the information is put
    // in __tspTypeToApiVersions before we call parameters since method wraps parameter
    const operationLocation = getLocationOfOperation(operation);
    const apiVersions = getAvailableApiVersions(context, operation, operationLocation);
    let clientParams = context.__clientToParameters.get(operationLocation);
    if (!clientParams) {
        clientParams = [];
        context.__clientToParameters.set(operationLocation, clientParams);
    }
    const override = getOverriddenClientMethod(context, operation);
    const params = (override ?? operation).parameters.properties.values();
    for (const param of params) {
        if (isNeverOrVoidType(param.type))
            continue;
        const sdkMethodParam = diagnostics.pipe(getSdkMethodParameter(context, param, operation));
        if (sdkMethodParam.onClient) {
            const operationLocation = getLocationOfOperation(operation);
            if (sdkMethodParam.isApiVersionParam) {
                if (!context.__clientToParameters.get(operationLocation)?.find((x) => x.isApiVersionParam)) {
                    clientParams.push(sdkMethodParam);
                }
            }
            else if (isSubscriptionId(context, param)) {
                if (!context.__clientToParameters
                    .get(operationLocation)
                    ?.find((x) => isSubscriptionId(context, x))) {
                    clientParams.push(sdkMethodParam);
                }
            }
        }
        else {
            methodParameters.push(sdkMethodParam);
        }
    }
    const serviceOperation = diagnostics.pipe(getSdkServiceOperation(context, operation, methodParameters));
    // set the correct encode for body parameter according to the content-type
    if (serviceOperation.bodyParam &&
        serviceOperation.bodyParam.correspondingMethodParams.length === 1) {
        const methodBodyParam = serviceOperation.bodyParam.correspondingMethodParams[0];
        const contentTypes = serviceOperation.__raw.parameters.body?.contentTypes;
        const defaultContentType = contentTypes && contentTypes.length > 0 ? contentTypes[0] : "application/json";
        diagnostics.pipe(addEncodeInfo(context, methodBodyParam.__raw, methodBodyParam.type, defaultContentType));
    }
    const response = getSdkMethodResponse(context, operation, serviceOperation, client);
    const name = getLibraryName(context, operation);
    return diagnostics.wrap({
        __raw: operation,
        kind: "basic",
        name,
        access: getAccess(context, operation) ?? "public",
        parameters: methodParameters,
        doc: getDoc(context.program, operation),
        summary: getSummary(context.program, operation),
        operation: serviceOperation,
        response,
        apiVersions,
        crossLanguageDefinitionId: getCrossLanguageDefinitionId(context, operation),
        decorators: diagnostics.pipe(getTypeDecorators(context, operation)),
        generateConvenient: shouldGenerateConvenient(context, operation),
        generateProtocol: shouldGenerateProtocol(context, operation),
        isOverride: override !== undefined,
    });
}
function getSdkServiceMethod(context, operation, client) {
    const lro = getLroMetadata(context.program, operation);
    const paging = getPagedResult(context.program, operation) || isList(context.program, operation);
    if (lro && paging) {
        return getSdkLroPagingServiceMethod(context, operation, client);
    }
    else if (paging) {
        return getSdkPagingServiceMethod(context, operation, client);
    }
    else if (lro) {
        return getSdkLroServiceMethod(context, operation, client);
    }
    return getSdkBasicServiceMethod(context, operation, client);
}
function getClientDefaultApiVersion(context, client) {
    if (context.apiVersion && !["latest", "all"].includes(context.apiVersion)) {
        return context.apiVersion;
    }
    let defaultVersion = getDefaultApiVersion(context, client.service)?.value;
    if (!defaultVersion) {
        // eslint-disable-next-line @typescript-eslint/no-deprecated
        defaultVersion = getService(context.program, client.service)?.version;
    }
    return defaultVersion;
}
function getSdkInitializationType(context, client) {
    const diagnostics = createDiagnosticCollector();
    let initializationModel = getClientInitialization(context, client.type);
    let clientParams = context.__clientToParameters.get(client.type);
    if (!clientParams) {
        clientParams = [];
        context.__clientToParameters.set(client.type, clientParams);
    }
    const access = client.kind === "SdkClient" ? "public" : "internal";
    if (initializationModel) {
        for (const prop of initializationModel.properties) {
            clientParams.push(prop);
        }
        initializationModel.access = access;
    }
    else {
        const namePrefix = client.kind === "SdkClient" ? client.name : client.groupPath;
        const name = `${namePrefix.split(".").at(-1)}Options`;
        initializationModel = {
            __raw: client.service,
            doc: "Initialization class for the client",
            kind: "model",
            properties: [],
            name,
            isGeneratedName: true,
            access,
            usage: UsageFlags.Input,
            crossLanguageDefinitionId: `${getNamespaceFullName(client.service.namespace)}.${name}`,
            clientNamespace: getClientNamespace(context, client.type),
            apiVersions: context.__tspTypeToApiVersions.get(client.type),
            decorators: [],
        };
    }
    return diagnostics.wrap(initializationModel);
}
function getSdkMethodParameter(context, type, operation) {
    const diagnostics = createDiagnosticCollector();
    return diagnostics.wrap({
        ...diagnostics.pipe(getSdkModelPropertyType(context, type, operation)),
        kind: "method",
    });
}
function getSdkMethods(context, client, sdkClientType) {
    const diagnostics = createDiagnosticCollector();
    const retval = [];
    for (const operation of listOperationsInOperationGroup(context, client)) {
        retval.push(diagnostics.pipe(getSdkServiceMethod(context, operation, sdkClientType)));
    }
    for (const operationGroup of listOperationGroups(context, client)) {
        // We create a client accessor for each operation group
        const operationGroupClient = diagnostics.pipe(createSdkClientType(context, operationGroup, sdkClientType));
        const clientInitialization = getClientInitialization(context, operationGroup.type);
        const parameters = [];
        if (clientInitialization) {
            for (const property of clientInitialization.properties) {
                parameters.push(property);
            }
        }
        const name = `get${operationGroup.type.name}`;
        retval.push({
            kind: "clientaccessor",
            parameters,
            name,
            doc: getDoc(context.program, operationGroup.type),
            summary: getSummary(context.program, operationGroup.type),
            access: "internal",
            response: operationGroupClient,
            apiVersions: getAvailableApiVersions(context, operationGroup.type, client.type),
            crossLanguageDefinitionId: `${getCrossLanguageDefinitionId(context, operationGroup.type)}.${name}`,
            decorators: [],
        });
    }
    return diagnostics.wrap(retval);
}
function getEndpointTypeFromSingleServer(context, client, server) {
    const diagnostics = createDiagnosticCollector();
    const templateArguments = [];
    const defaultOverridableEndpointType = {
        kind: "endpoint",
        serverUrl: "{endpoint}",
        templateArguments: [
            {
                name: "endpoint",
                isGeneratedName: true,
                doc: "Service host",
                kind: "path",
                onClient: true,
                urlEncode: false,
                explode: false,
                style: "simple",
                allowReserved: false,
                optional: false,
                serializedName: "endpoint",
                correspondingMethodParams: [],
                type: getTypeSpecBuiltInType(context, "string"),
                isApiVersionParam: false,
                apiVersions: context.__tspTypeToApiVersions.get(client.__raw.type),
                crossLanguageDefinitionId: `${getCrossLanguageDefinitionId(context, client.__raw.service)}.endpoint`,
                decorators: [],
            },
        ],
        decorators: [],
    };
    const types = [];
    if (!server)
        return diagnostics.wrap([defaultOverridableEndpointType]);
    for (const param of server.parameters.values()) {
        const sdkParam = diagnostics.pipe(getSdkHttpParameter(context, param, undefined, undefined, "path"));
        if (sdkParam.kind === "path") {
            templateArguments.push(sdkParam);
            sdkParam.onClient = true;
            if (param.defaultValue) {
                sdkParam.clientDefaultValue = getValueTypeValue(param.defaultValue);
            }
            const apiVersionInfo = updateWithApiVersionInformation(context, param, client.__raw.type);
            sdkParam.isApiVersionParam = apiVersionInfo.isApiVersionParam;
            if (sdkParam.isApiVersionParam && apiVersionInfo.clientDefaultValue) {
                sdkParam.clientDefaultValue = apiVersionInfo.clientDefaultValue;
            }
            sdkParam.apiVersions = getAvailableApiVersions(context, param, client.__raw.type);
            sdkParam.crossLanguageDefinitionId = `${getCrossLanguageDefinitionId(context, client.__raw.service)}.${param.name}`;
        }
        else {
            diagnostics.add(createDiagnostic({
                code: "server-param-not-path",
                target: param,
                format: {
                    templateArgumentName: sdkParam.name,
                    templateArgumentType: sdkParam.kind,
                },
            }));
        }
    }
    const isOverridable = templateArguments.length === 1 && server.url.startsWith("{") && server.url.endsWith("}");
    if (templateArguments.length === 0) {
        types.push(defaultOverridableEndpointType);
        types[0].templateArguments[0].clientDefaultValue = server.url;
    }
    else {
        types.push({
            kind: "endpoint",
            serverUrl: server.url,
            templateArguments,
            decorators: [],
        });
        if (!isOverridable) {
            types.push(defaultOverridableEndpointType);
        }
    }
    return diagnostics.wrap(types);
}
function getSdkEndpointParameter(context, client) {
    const diagnostics = createDiagnosticCollector();
    const rawClient = client.__raw;
    const servers = getServers(context.program, client.__raw.service);
    const types = [];
    if (servers === undefined) {
        // if there is no defined server url, we will return an overridable endpoint
        types.push(...diagnostics.pipe(getEndpointTypeFromSingleServer(context, client, undefined)));
    }
    else {
        for (const server of servers) {
            types.push(...diagnostics.pipe(getEndpointTypeFromSingleServer(context, client, server)));
        }
    }
    let type;
    if (types.length > 1) {
        type = {
            kind: "union",
            access: "public",
            usage: UsageFlags.None,
            variantTypes: types,
            name: createGeneratedName(context, rawClient.service, "Endpoint"),
            isGeneratedName: true,
            crossLanguageDefinitionId: `${getCrossLanguageDefinitionId(context, rawClient.service)}.Endpoint`,
            clientNamespace: getClientNamespace(context, rawClient.service),
            decorators: [],
        };
    }
    else {
        type = types[0];
    }
    return diagnostics.wrap({
        kind: "endpoint",
        type,
        name: "endpoint",
        isGeneratedName: true,
        doc: "Service host",
        onClient: true,
        urlEncode: false,
        apiVersions: context.__tspTypeToApiVersions.get(rawClient.type),
        optional: false,
        isApiVersionParam: false,
        crossLanguageDefinitionId: `${getCrossLanguageDefinitionId(context, rawClient.service)}.endpoint`,
        decorators: [],
    });
}
function createSdkClientType(context, client, parent) {
    const diagnostics = createDiagnosticCollector();
    const isClient = client.kind === "SdkClient";
    let name = "";
    if (isClient) {
        name = client.name;
    }
    else {
        name = getClientNameOverride(context, client.type) ?? client.type.name;
    }
    const sdkClientType = {
        __raw: client,
        kind: "client",
        name,
        doc: getDoc(context.program, client.type),
        summary: getSummary(context.program, client.type),
        methods: [],
        apiVersions: context.__tspTypeToApiVersions.get(client.type),
        nameSpace: getClientNamespaceStringHelper(context, client.service),
        clientNamespace: getClientNamespace(context, client.type),
        initialization: diagnostics.pipe(getSdkInitializationType(context, client)),
        decorators: diagnostics.pipe(getTypeDecorators(context, client.type)),
        parent,
        // if it is client, the crossLanguageDefinitionId is the ${namespace}, if it is operation group, the crosslanguageDefinitionId is the %{namespace}.%{operationGroupName}
        crossLanguageDefinitionId: getCrossLanguageDefinitionId(context, client.type),
    };
    // NOTE: getSdkMethods recursively calls createSdkClientType
    sdkClientType.methods = diagnostics.pipe(getSdkMethods(context, client, sdkClientType));
    addDefaultClientParameters(context, sdkClientType);
    return diagnostics.wrap(sdkClientType);
}
function addDefaultClientParameters(context, client) {
    const diagnostics = createDiagnosticCollector();
    // there will always be an endpoint property
    client.initialization.properties.push(diagnostics.pipe(getSdkEndpointParameter(context, client)));
    const credentialParam = getSdkCredentialParameter(context, client.__raw);
    if (credentialParam) {
        client.initialization.properties.push(credentialParam);
    }
    let apiVersionParam = context.__clientToParameters
        .get(client.__raw.type)
        ?.find((x) => x.isApiVersionParam);
    if (!apiVersionParam) {
        for (const operationGroup of listOperationGroups(context, client.__raw)) {
            // if any sub operation groups have an api version param, the top level needs
            // the api version param as well
            apiVersionParam = context.__clientToParameters
                .get(operationGroup.type)
                ?.find((x) => x.isApiVersionParam);
            if (apiVersionParam)
                break;
        }
    }
    if (apiVersionParam) {
        client.initialization.properties.push(apiVersionParam);
    }
    let subId = context.__clientToParameters
        .get(client.__raw.type)
        ?.find((x) => isSubscriptionId(context, x));
    if (!subId && context.arm) {
        for (const operationGroup of listOperationGroups(context, client.__raw)) {
            // if any sub operation groups have an subId param, the top level needs it as well
            subId = context.__clientToParameters
                .get(operationGroup.type)
                ?.find((x) => isSubscriptionId(context, x));
            if (subId)
                break;
        }
    }
    if (subId) {
        client.initialization.properties.push(subId);
    }
}
function populateApiVersionInformation(context) {
    for (const client of listClients(context)) {
        let clientApiVersions = resolveVersions(context.program, client.service)
            .filter((x) => x.rootVersion)
            .map((x) => x.rootVersion.value);
        context.__tspTypeToApiVersions.set(client.type, filterApiVersionsWithDecorators(context, client.type, clientApiVersions));
        context.__clientToApiVersionClientDefaultValue.set(client.type, getClientDefaultApiVersion(context, client));
        for (const og of listOperationGroups(context, client)) {
            clientApiVersions = resolveVersions(context.program, og.service)
                .filter((x) => x.rootVersion)
                .map((x) => x.rootVersion.value);
            context.__tspTypeToApiVersions.set(og.type, filterApiVersionsWithDecorators(context, og.type, clientApiVersions));
            context.__clientToApiVersionClientDefaultValue.set(og.type, getClientDefaultApiVersion(context, og));
        }
    }
}
function getSdkPackage(context) {
    const diagnostics = createDiagnosticCollector();
    populateApiVersionInformation(context);
    diagnostics.pipe(handleAllTypes(context));
    const crossLanguagePackageId = diagnostics.pipe(getCrossLanguagePackageId(context));
    const allReferencedTypes = getAllReferencedTypes(context);
    const sdkPackage = {
        name: getClientNamespaceString(context),
        rootNamespace: getClientNamespaceString(context),
        clients: listClients(context).map((c) => diagnostics.pipe(createSdkClientType(context, c))),
        models: allReferencedTypes.filter((x) => x.kind === "model"),
        enums: allReferencedTypes.filter((x) => x.kind === "enum"),
        unions: allReferencedTypes.filter((x) => x.kind === "union" || x.kind === "nullable"),
        crossLanguagePackageId,
        namespaces: [],
    };
    organizeNamespaces(sdkPackage);
    return diagnostics.wrap(sdkPackage);
}
function organizeNamespaces(sdkPackage) {
    const clients = [...sdkPackage.clients];
    while (clients.length > 0) {
        const client = clients.shift();
        getSdkNamespace(sdkPackage, client.clientNamespace).clients.push(client);
        client.methods
            .filter((m) => m.kind === "clientaccessor")
            .map((m) => m.response)
            .map((c) => clients.push(c));
    }
    for (const model of sdkPackage.models) {
        getSdkNamespace(sdkPackage, model.clientNamespace).models.push(model);
    }
    for (const enumType of sdkPackage.enums) {
        getSdkNamespace(sdkPackage, enumType.clientNamespace).enums.push(enumType);
    }
    for (const unionType of sdkPackage.unions) {
        getSdkNamespace(sdkPackage, unionType.clientNamespace).unions.push(unionType);
    }
}
function getSdkNamespace(sdkPackage, namespace) {
    const segments = namespace.split(".");
    let current = sdkPackage;
    let fullName = "";
    for (const segment of segments) {
        fullName = fullName === "" ? segment : `${fullName}.${segment}`;
        const ns = current.namespaces.find((ns) => ns.name === segment);
        if (ns === undefined) {
            const newNs = {
                name: segment,
                fullName,
                clients: [],
                models: [],
                enums: [],
                unions: [],
                namespaces: [],
            };
            current.namespaces.push(newNs);
            current = newNs;
        }
        else {
            current = ns;
        }
    }
    return current;
}

function createTCGCContext(program, emitterName) {
    const diagnostics = createDiagnosticCollector();
    return {
        program,
        emitterName: diagnostics.pipe(parseEmitterName(program, emitterName ?? program.emitters[0]?.metadata?.name)),
        diagnostics: diagnostics.diagnostics,
        originalProgram: program,
        __clientToParameters: new Map(),
        __tspTypeToApiVersions: new Map(),
        __clientToApiVersionClientDefaultValue: new Map(),
        previewStringRegex: /-preview$/,
        disableUsageAccessPropagationToBase: false,
        __pagedResultSet: new Set(),
    };
}
async function createSdkContext(context, emitterName, options) {
    const diagnostics = createDiagnosticCollector();
    const protocolOptions = true; // context.program.getLibraryOptions("generate-protocol-methods");
    const convenienceOptions = true; // context.program.getLibraryOptions("generate-convenience-methods");
    const generateProtocolMethods = context.options["generate-protocol-methods"] ?? protocolOptions;
    const generateConvenienceMethods = context.options["generate-convenience-methods"] ?? convenienceOptions;
    const tcgcContext = createTCGCContext(context.program, emitterName ?? context.options["emitter-name"]);
    const sdkContext = {
        ...tcgcContext,
        emitContext: context,
        sdkPackage: undefined,
        generateProtocolMethods: generateProtocolMethods,
        generateConvenienceMethods: generateConvenienceMethods,
        packageName: context.options["package-name"],
        flattenUnionAsEnum: context.options["flatten-union-as-enum"] ?? true,
        apiVersion: options?.versioning?.strategy === "ignore" ? "all" : context.options["api-version"],
        examplesDir: context.options["examples-dir"] ?? context.options["examples-directory"],
        decoratorsAllowList: [...defaultDecoratorsAllowList, ...(options?.additionalDecorators ?? [])],
        previewStringRegex: options?.versioning?.previewStringRegex || tcgcContext.previewStringRegex,
        disableUsageAccessPropagationToBase: options?.disableUsageAccessPropagationToBase ?? false,
    };
    sdkContext.sdkPackage = diagnostics.pipe(getSdkPackage(sdkContext));
    for (const client of sdkContext.sdkPackage.clients) {
        diagnostics.pipe(await handleClientExamples(sdkContext, client));
    }
    sdkContext.diagnostics = sdkContext.diagnostics.concat(diagnostics.diagnostics);
    if (options?.exportTCGCoutput) {
        await exportTCGCOutput(sdkContext);
    }
    return sdkContext;
}
async function exportTCGCOutput(context) {
    await emitFile(context.program, {
        path: resolvePath(context.emitContext.emitterOutputDir, "tcgc-output.yaml"),
        content: stringify(context.sdkPackage, (k, v) => {
            if (typeof k === "string" && k.startsWith("__")) {
                return undefined; // skip keys starting with "__" from the output
            }
            if (k === "scheme") {
                return undefined; // remove credential schema
            }
            if (k === "rawExample") {
                return undefined; // remove raw example
            }
            return v;
        }, { lineWidth: 0 }),
    });
}
async function $onEmit(context) {
    if (!context.program.compilerOptions.noEmit) {
        const sdkContext = await createSdkContext(context);
        await exportTCGCOutput(sdkContext);
    }
}

function $onValidate(program) {
    const tcgcContext = createTCGCContext(program, "@azure-tools/typespec-client-generator-core");
    const languageScopes = getDefinedLanguageScopes(program);
    for (const scope of languageScopes) {
        validateClientNamesPerNamespace(tcgcContext, scope, program.getGlobalNamespaceType());
    }
}
function getDefinedLanguageScopes(program) {
    const languageScopes = new Set();
    for (const value of program.stateMap(clientNameKey).values()) {
        if (value[AllScopes]) {
            languageScopes.add(AllScopes);
        }
        for (const languageScope of Object.keys(value)) {
            languageScopes.add(languageScope);
        }
    }
    return languageScopes;
}
function validateClientNamesPerNamespace(tcgcContext, scope, namespace) {
    // Check for duplicate client names for models, enums, and unions
    validateClientNamesCore(tcgcContext, scope, [
        ...namespace.models.values(),
        ...namespace.enums.values(),
        ...namespace.unions.values(),
    ]);
    // Check for duplicate client names for operations
    validateClientNamesCore(tcgcContext, scope, namespace.operations.values());
    // check for duplicate client names for operations in interfaces
    for (const item of namespace.interfaces.values()) {
        validateClientNamesCore(tcgcContext, scope, item.operations.values());
    }
    // Check for duplicate client names for interfaces
    validateClientNamesCore(tcgcContext, scope, namespace.interfaces.values());
    // Check for duplicate client names for scalars
    validateClientNamesCore(tcgcContext, scope, namespace.scalars.values());
    // Check for duplicate client names for namespaces
    validateClientNamesCore(tcgcContext, scope, namespace.namespaces.values());
    // Check for duplicate client names for model properties
    for (const model of namespace.models.values()) {
        validateClientNamesCore(tcgcContext, scope, model.properties.values());
    }
    // Check for duplicate client names for enum members
    for (const item of namespace.enums.values()) {
        validateClientNamesCore(tcgcContext, scope, item.members.values());
    }
    // Check for duplicate client names for union variants
    for (const item of namespace.unions.values()) {
        validateClientNamesCore(tcgcContext, scope, item.variants.values());
    }
    // Check for duplicate client names for nested namespaces
    for (const item of namespace.namespaces.values()) {
        validateClientNamesPerNamespace(tcgcContext, scope, item);
    }
}
function validateClientNamesCore(tcgcContext, scope, items) {
    const duplicateTracker = new DuplicateTracker();
    for (const item of items) {
        const clientName = getClientNameOverride(tcgcContext, item, scope);
        if (clientName !== undefined) {
            const clientNameDecorator = item.decorators.find((x) => x.definition?.name === "@clientName");
            if (clientNameDecorator?.node !== undefined) {
                duplicateTracker.track(clientName, clientNameDecorator.node);
            }
        }
        else {
            if (item.name !== undefined && typeof item.name === "string") {
                duplicateTracker.track(item.name, item);
            }
        }
    }
    reportDuplicateClientNames(tcgcContext.program, duplicateTracker, scope);
}
function reportDuplicateClientNames(program, duplicateTracker, scope) {
    for (const [name, duplicates] of duplicateTracker.entries()) {
        for (const item of duplicates) {
            const scopeStr = scope === AllScopes ? "AllScopes" : scope;
            // If the item is a decorator application node
            if (item.kind === SyntaxKind.DecoratorExpression ||
                item.kind === SyntaxKind.AugmentDecoratorStatement) {
                reportDiagnostic(program, {
                    code: "duplicate-client-name",
                    format: { name, scope: scopeStr },
                    target: item,
                });
            }
            else {
                reportDiagnostic(program, {
                    code: "duplicate-client-name",
                    messageId: "nonDecorator",
                    format: { name, scope: scopeStr },
                    target: item,
                });
            }
        }
    }
}

/** @internal */
const $decorators = {
    "Azure.ClientGenerator.Core": {
        clientName: $clientName,
        convenientAPI: $convenientAPI,
        protocolAPI: $protocolAPI,
        client: $client,
        operationGroup: $operationGroup,
        usage: $usage,
        access: $access,
        flattenProperty: $flattenProperty,
        override: $override,
        useSystemTextJsonConverter: $useSystemTextJsonConverter,
        clientInitialization: $clientInitialization,
        paramAlias: $paramAlias,
        apiVersion: $apiVersion,
        clientNamespace: $clientNamespace,
        alternateType: $alternateType,
        scope: $scope,
    },
};

var f1 = /*#__PURE__*/Object.freeze({
    __proto__: null,
    $decorators: $decorators,
    $lib: $lib,
    $onValidate: $onValidate
});

var f0 = /*#__PURE__*/Object.freeze({
    __proto__: null,
    $access: $access,
    $alternateType: $alternateType,
    $apiVersion: $apiVersion,
    $client: $client,
    $clientInitialization: $clientInitialization,
    $clientName: $clientName,
    $clientNamespace: $clientNamespace,
    $convenientAPI: $convenientAPI,
    $decorators: $decorators,
    $flattenProperty: $flattenProperty,
    $lib: $lib,
    $onEmit: $onEmit,
    $operationGroup: $operationGroup,
    $override: $override,
    $paramAlias: $paramAlias,
    $protocolAPI: $protocolAPI,
    $scope: $scope,
    $usage: $usage,
    $useSystemTextJsonConverter: $useSystemTextJsonConverter,
    get UsageFlags () { return UsageFlags; },
    addEncodeInfo: addEncodeInfo,
    createDiagnostic: createDiagnostic,
    createSdkContext: createSdkContext,
    createStateSymbol: createStateSymbol,
    createTCGCContext: createTCGCContext,
    getAccess: getAccess,
    getAccessOverride: getAccessOverride,
    getAllModels: getAllModels,
    getAllModelsWithDiagnostics: getAllModelsWithDiagnostics,
    getAllReferencedTypes: getAllReferencedTypes,
    getAlternateType: getAlternateType,
    getClient: getClient,
    getClientInitialization: getClientInitialization,
    getClientNameOverride: getClientNameOverride,
    getClientNamespace: getClientNamespace,
    getClientNamespaceString: getClientNamespaceString,
    getClientType: getClientType,
    getClientTypeWithDiagnostics: getClientTypeWithDiagnostics,
    getCrossLanguageDefinitionId: getCrossLanguageDefinitionId,
    getCrossLanguagePackageId: getCrossLanguagePackageId,
    getDefaultApiVersion: getDefaultApiVersion,
    getEffectivePayloadType: getEffectivePayloadType,
    getGeneratedName: getGeneratedName,
    getHttpOperationExamples: getHttpOperationExamples,
    getHttpOperationParameter: getHttpOperationParameter,
    getHttpOperationWithCache: getHttpOperationWithCache,
    getInitializationType: getInitializationType,
    getIsApiVersion: getIsApiVersion,
    getKnownScalars: getKnownScalars,
    getLibraryName: getLibraryName,
    getOperationGroup: getOperationGroup,
    getOverriddenClientMethod: getOverriddenClientMethod,
    getParamAlias: getParamAlias,
    getPropertyNames: getPropertyNames,
    getSdkArrayOrDict: getSdkArrayOrDict,
    getSdkArrayOrDictWithDiagnostics: getSdkArrayOrDictWithDiagnostics,
    getSdkBuiltInType: getSdkBuiltInType,
    getSdkConstant: getSdkConstant,
    getSdkCredentialParameter: getSdkCredentialParameter,
    getSdkDurationType: getSdkDurationType,
    getSdkEnum: getSdkEnum,
    getSdkEnumValue: getSdkEnumValue,
    getSdkModel: getSdkModel,
    getSdkModelPropertyType: getSdkModelPropertyType,
    getSdkModelPropertyTypeBase: getSdkModelPropertyTypeBase,
    getSdkModelWithDiagnostics: getSdkModelWithDiagnostics,
    getSdkTuple: getSdkTuple,
    getSdkTupleWithDiagnostics: getSdkTupleWithDiagnostics,
    getSdkUnion: getSdkUnion,
    getSdkUnionEnum: getSdkUnionEnum,
    getSdkUnionEnumWithDiagnostics: getSdkUnionEnumWithDiagnostics,
    getSdkUnionWithDiagnostics: getSdkUnionWithDiagnostics,
    getTypeSpecBuiltInType: getTypeSpecBuiltInType,
    getUsage: getUsage,
    getUsageOverride: getUsageOverride,
    getWireName: getWireName,
    handleAllTypes: handleAllTypes,
    isApiVersion: isApiVersion,
    isAzureCoreModel: isAzureCoreModel,
    isInOperationGroup: isInOperationGroup,
    isOperationGroup: isOperationGroup,
    isPagedResultModel: isPagedResultModel,
    isReadOnly: isReadOnly,
    isSdkBuiltInKind: isSdkBuiltInKind,
    isSdkDateTimeEncodings: isSdkDateTimeEncodings,
    isSdkFloatKind: isSdkFloatKind,
    isSdkIntKind: isSdkIntKind,
    listClients: listClients,
    listOperationGroups: listOperationGroups,
    listOperationsInOperationGroup: listOperationsInOperationGroup,
    listSubClients: listSubClients,
    namespace: namespace,
    reportDiagnostic: reportDiagnostic,
    shouldFlattenProperty: shouldFlattenProperty,
    shouldGenerateConvenient: shouldGenerateConvenient,
    shouldGenerateProtocol: shouldGenerateProtocol
});

const TypeSpecJSSources = {
"dist/src/index.js": f0,
"dist/src/tsp-index.js": f1,
};
const TypeSpecSources = {
  "package.json": "{\"name\":\"@azure-tools/typespec-client-generator-core\",\"version\":\"0.50.0\",\"author\":\"Microsoft Corporation\",\"description\":\"TypeSpec Data Plane Generation library\",\"homepage\":\"https://azure.github.io/typespec-azure\",\"readme\":\"https://github.com/Microsoft/typespec/blob/main/README.md\",\"license\":\"MIT\",\"repository\":{\"type\":\"git\",\"url\":\"git+https://github.com/Azure/typespec-azure.git\"},\"bugs\":{\"url\":\"https://github.com/Azure/typespec-azure/issues\"},\"keywords\":[\"typespec\",\"sdk\",\"ClientGenerator\"],\"main\":\"dist/src/index.js\",\"tspMain\":\"./lib/main.tsp\",\"exports\":{\".\":{\"types\":\"./dist/src/index.d.ts\",\"default\":\"./dist/src/index.js\",\"typespec\":\"./lib/main.tsp\"},\"./testing\":{\"types\":\"./dist/src/testing/index.d.ts\",\"default\":\"./dist/src/testing/index.js\"}},\"type\":\"module\",\"engines\":{\"node\":\">=18.0.0\"},\"scripts\":{\"clean\":\"rimraf ./dist ./temp\",\"build\":\"npm run gen-extern-signature && tsc -p . && npm run lint-typespec-library\",\"watch\":\"tsc -p . --watch\",\"gen-extern-signature\":\"tspd --enable-experimental gen-extern-signature .\",\"lint-typespec-library\":\"tsp compile . --warn-as-error --import @typespec/library-linter --no-emit\",\"test\":\"vitest run\",\"test:watch\":\"vitest -w\",\"test:ui\":\"vitest --ui\",\"test:ci\":\"vitest run --coverage  --reporter=junit --reporter=default\",\"test:e2e\":\"vitest run --config ./vitest.config.e2e.ts\",\"lint\":\"eslint .  --max-warnings=0\",\"lint:fix\":\"eslint . --fix \",\"regen-docs\":\"tspd doc .  --enable-experimental  --output-dir ../../website/src/content/docs/docs/libraries/typespec-client-generator-core/reference\"},\"files\":[\"lib/*.tsp\",\"dist/**\",\"!dist/test/**\"],\"dependencies\":{\"change-case\":\"~5.4.4\",\"pluralize\":\"^8.0.0\",\"yaml\":\"~2.5.1\"},\"peerDependencies\":{\"@azure-tools/typespec-azure-core\":\"workspace:~\",\"@typespec/compiler\":\"workspace:~\",\"@typespec/http\":\"workspace:~\",\"@typespec/openapi\":\"workspace:~\",\"@typespec/rest\":\"workspace:~\",\"@typespec/versioning\":\"workspace:~\"},\"devDependencies\":{\"@azure-tools/typespec-azure-core\":\"workspace:~\",\"@azure-tools/typespec-azure-resource-manager\":\"workspace:~\",\"@types/node\":\"~22.7.9\",\"@types/pluralize\":\"^0.0.33\",\"@typespec/compiler\":\"workspace:~\",\"@typespec/http\":\"workspace:~\",\"@typespec/library-linter\":\"workspace:~\",\"@typespec/openapi\":\"workspace:~\",\"@typespec/prettier-plugin-typespec\":\"workspace:~\",\"@typespec/rest\":\"workspace:~\",\"@typespec/tspd\":\"workspace:~\",\"@typespec/xml\":\"workspace:~\",\"@vitest/coverage-v8\":\"^2.1.5\",\"@vitest/ui\":\"^2.1.2\",\"c8\":\"^10.1.2\",\"rimraf\":\"~6.0.1\",\"typescript\":\"~5.6.3\",\"vitest\":\"^2.1.5\"}}",
  "../../core/packages/compiler/lib/intrinsics.tsp": "import \"../dist/src/lib/intrinsic/tsp-index.js\";\nimport \"./prototypes.tsp\";\n\n// This file contains all the intrinsic types of typespec. Everything here will always be loaded\nnamespace TypeSpec;\n\n/**\n * Represent a byte array\n */\nscalar bytes;\n\n/**\n * A numeric type\n */\nscalar numeric;\n\n/**\n * A whole number. This represent any `integer` value possible.\n * It is commonly represented as `BigInteger` in some languages.\n */\nscalar integer extends numeric;\n\n/**\n * A number with decimal value\n */\nscalar float extends numeric;\n\n/**\n * A 64-bit integer. (`-9,223,372,036,854,775,808` to `9,223,372,036,854,775,807`)\n */\nscalar int64 extends integer;\n\n/**\n * A 32-bit integer. (`-2,147,483,648` to `2,147,483,647`)\n */\nscalar int32 extends int64;\n\n/**\n * A 16-bit integer. (`-32,768` to `32,767`)\n */\nscalar int16 extends int32;\n\n/**\n * A 8-bit integer. (`-128` to `127`)\n */\nscalar int8 extends int16;\n\n/**\n * A 64-bit unsigned integer (`0` to `18,446,744,073,709,551,615`)\n */\nscalar uint64 extends integer;\n\n/**\n * A 32-bit unsigned integer (`0` to `4,294,967,295`)\n */\nscalar uint32 extends uint64;\n\n/**\n * A 16-bit unsigned integer (`0` to `65,535`)\n */\nscalar uint16 extends uint32;\n\n/**\n * A 8-bit unsigned integer (`0` to `255`)\n */\nscalar uint8 extends uint16;\n\n/**\n * An integer that can be serialized to JSON (`−9007199254740991 (−(2^53 − 1))` to `9007199254740991 (2^53 − 1)` )\n */\nscalar safeint extends int64;\n\n/**\n * A 64 bit floating point number. (`±5.0 × 10^−324` to `±1.7 × 10^308`)\n */\nscalar float64 extends float;\n\n/**\n * A 32 bit floating point number. (`±1.5 x 10^−45` to `±3.4 x 10^38`)\n */\nscalar float32 extends float64;\n\n/**\n * A decimal number with any length and precision. This represent any `decimal` value possible.\n * It is commonly represented as `BigDecimal` in some languages.\n */\nscalar decimal extends numeric;\n\n/**\n * A 128-bit decimal number.\n */\nscalar decimal128 extends decimal;\n\n/**\n * A sequence of textual characters.\n */\nscalar string;\n\n/**\n * A date on a calendar without a time zone, e.g. \"April 10th\"\n */\nscalar plainDate {\n  /**\n   * Create a plain date from an ISO 8601 string.\n   * @example\n   *\n   * ```tsp\n   * const time = plainTime.fromISO(\"2024-05-06\");\n   * ```\n   */\n  init fromISO(value: string);\n}\n\n/**\n * A time on a clock without a time zone, e.g. \"3:00 am\"\n */\nscalar plainTime {\n  /**\n   * Create a plain time from an ISO 8601 string.\n   * @example\n   *\n   * ```tsp\n   * const time = plainTime.fromISO(\"12:34\");\n   * ```\n   */\n  init fromISO(value: string);\n}\n\n/**\n * An instant in coordinated universal time (UTC)\"\n */\nscalar utcDateTime {\n  /**\n   * Create a date from an ISO 8601 string.\n   * @example\n   *\n   * ```tsp\n   * const time = utcDateTime.fromISO(\"2024-05-06T12:20-12Z\");\n   * ```\n   */\n  init fromISO(value: string);\n}\n\n/**\n * A date and time in a particular time zone, e.g. \"April 10th at 3:00am in PST\"\n */\nscalar offsetDateTime {\n  /**\n   * Create a date from an ISO 8601 string.\n   * @example\n   *\n   * ```tsp\n   * const time = offsetDateTime.fromISO(\"2024-05-06T12:20-12-0700\");\n   * ```\n   */\n  init fromISO(value: string);\n}\n\n/**\n * A duration/time period. e.g 5s, 10h\n */\nscalar duration {\n  /**\n   * Create a duration from an ISO 8601 string.\n   * @example\n   *\n   * ```tsp\n   * const time = duration.fromISO(\"P1Y1D\");\n   * ```\n   */\n  init fromISO(value: string);\n}\n\n/**\n * Boolean with `true` and `false` values.\n */\nscalar boolean;\n\n/**\n * @dev Array model type, equivalent to `Element[]`\n * @template Element The type of the array elements\n */\n@indexer(integer, Element)\nmodel Array<Element> {}\n\n/**\n * @dev Model with string properties where all the properties have type `Property`\n * @template Element The type of the properties\n */\n@indexer(string, Element)\nmodel Record<Element> {}\n",
  "../../core/packages/compiler/lib/prototypes.tsp": "namespace TypeSpec.Prototypes;\n\nextern dec getter(target: unknown);\n\nnamespace Types {\n  interface ModelProperty {\n    @getter type(): unknown;\n  }\n\n  interface Operation {\n    @getter returnType(): unknown;\n    @getter parameters(): unknown;\n  }\n\n  interface Array<TElementType> {\n    @getter elementType(): TElementType;\n  }\n}\n",
  "../../core/packages/compiler/lib/std/main.tsp": "// TypeSpec standard library. Everything in here can be omitted by using `--nostdlib` cli flag or `nostdlib` in the config.\nimport \"./types.tsp\";\nimport \"./decorators.tsp\";\nimport \"./reflection.tsp\";\nimport \"./projected-names.tsp\";\nimport \"./visibility.tsp\";\n",
  "../../core/packages/compiler/lib/std/types.tsp": "namespace TypeSpec;\n\n/**\n * Represent a 32-bit unix timestamp datetime with 1s of granularity.\n * It measures time by the number of seconds that have elapsed since 00:00:00 UTC on 1 January 1970.\n */\n@encode(\"unixTimestamp\", int32)\nscalar unixTimestamp32 extends utcDateTime;\n\n/**\n * Represent a model\n */\n// Deprecated June 2023 sprint\n#deprecated \"object is deprecated. Please use {} for an empty model, `Record<unknown>` for a record with unknown property types, `unknown[]` for an array.\"\nmodel object {}\n\n/**\n * Represent a URL string as described by https://url.spec.whatwg.org/\n */\nscalar url extends string;\n\n/**\n * Represents a collection of optional properties.\n *\n * @template Source An object whose spread properties are all optional.\n */\n@doc(\"The template for adding optional properties.\")\n@withOptionalProperties\nmodel OptionalProperties<Source> {\n  ...Source;\n}\n\n/**\n * Represents a collection of updateable properties.\n *\n * @template Source An object whose spread properties are all updateable.\n */\n@doc(\"The template for adding updateable properties.\")\n@withUpdateableProperties\nmodel UpdateableProperties<Source> {\n  ...Source;\n}\n\n/**\n * Represents a collection of omitted properties.\n *\n * @template Source An object whose properties are spread.\n * @template Keys The property keys to omit.\n */\n@doc(\"The template for omitting properties.\")\n@withoutOmittedProperties(Keys)\nmodel OmitProperties<Source, Keys extends string> {\n  ...Source;\n}\n\n/**\n * Represents a collection of properties with only the specified keys included.\n *\n * @template Source An object whose properties are spread.\n * @template Keys The property keys to include.\n */\n@doc(\"The template for picking properties.\")\n@withPickedProperties(Keys)\nmodel PickProperties<Source, Keys extends string> {\n  ...Source;\n}\n\n/**\n * Represents a collection of properties with default values omitted.\n *\n * @template Source An object whose spread property defaults are all omitted.\n */\n@withoutDefaultValues\nmodel OmitDefaults<Source> {\n  ...Source;\n}\n\n/**\n * Applies a visibility setting to a collection of properties.\n *\n * @template Source An object whose properties are spread.\n * @template Visibility The visibility to apply to all properties.\n */\n@doc(\"The template for setting the default visibility of key properties.\")\n@withDefaultKeyVisibility(Visibility)\nmodel DefaultKeyVisibility<Source, Visibility extends valueof string> {\n  ...Source;\n}\n",
  "../../core/packages/compiler/lib/std/decorators.tsp": "import \"../../dist/src/lib/tsp-index.js\";\n\nusing TypeSpec.Reflection;\n\nnamespace TypeSpec;\n\n/**\n * Typically a short, single-line description.\n * @param summary Summary string.\n *\n * @example\n * ```typespec\n * @summary(\"This is a pet\")\n * model Pet {}\n * ```\n */\nextern dec summary(target: unknown, summary: valueof string);\n\n/**\n * Attach a documentation string.\n * @param doc Documentation string\n * @param formatArgs Record with key value pair that can be interpolated in the doc.\n *\n * @example\n * ```typespec\n * @doc(\"Represent a Pet available in the PetStore\")\n * model Pet {}\n * ```\n */\nextern dec doc(target: unknown, doc: valueof string, formatArgs?: {});\n\n/**\n * Attach a documentation string to describe the successful return types of an operation.\n * If an operation returns a union of success and errors it only describes the success. See `@errorsDoc` for error documentation.\n * @param doc Documentation string\n *\n * @example\n * ```typespec\n * @returnsDoc(\"Returns doc\")\n * op get(): Pet | NotFound;\n * ```\n */\nextern dec returnsDoc(target: Operation, doc: valueof string);\n\n/**\n * Attach a documentation string to describe the error return types of an operation.\n * If an operation returns a union of success and errors it only describes the errors. See `@returnsDoc` for success documentation.\n * @param doc Documentation string\n *\n * @example\n * ```typespec\n * @errorsDoc(\"Errors doc\")\n * op get(): Pet | NotFound;\n * ```\n */\nextern dec errorsDoc(target: Operation, doc: valueof string);\n\n/**\n * Mark this type as deprecated.\n *\n * NOTE: This decorator **should not** be used, use the `#deprecated` directive instead.\n *\n * @deprecated Use the `#deprecated` [directive](https://typespec.io/docs/language-basics/directives/#deprecated) instead.\n * @param message Deprecation message.\n *\n * @example\n *\n * Use the `#deprecated` directive instead:\n *\n * ```typespec\n * #deprecated \"Use ActionV2\"\n * op Action<Result>(): Result;\n * ```\n */\n#deprecated \"@deprecated decorator is deprecated. Use the `#deprecated` directive instead.\"\nextern dec deprecated(target: unknown, message: valueof string);\n\n/**\n * Service options.\n */\nmodel ServiceOptions {\n  /**\n   * Title of the service.\n   */\n  title?: string;\n\n  /**\n   * Version of the service.\n   */\n  version?: string;\n}\n\n/**\n * Mark this namespace as describing a service and configure service properties.\n * @param options Optional configuration for the service.\n *\n * @example\n * ```typespec\n * @service\n * namespace PetStore;\n * ```\n *\n * @example Setting service title\n * ```typespec\n * @service({title: \"Pet store\"})\n * namespace PetStore;\n * ```\n *\n * @example Setting service version\n * ```typespec\n * @service({version: \"1.0\"})\n * namespace PetStore;\n * ```\n */\nextern dec service(target: Namespace, options?: ServiceOptions);\n\n/**\n * Specify that this model is an error type. Operations return error types when the operation has failed.\n *\n * @example\n * ```typespec\n * @error\n * model PetStoreError {\n *   code: string;\n *   message: string;\n * }\n * ```\n */\nextern dec error(target: Model);\n\n/**\n * Specify a known data format hint for this string type. For example `uuid`, `uri`, etc.\n * This differs from the `@pattern` decorator which is meant to specify a regular expression while `@format` accepts a known format name.\n * The format names are open ended and are left to emitter to interpret.\n *\n * @param format format name.\n *\n * @example\n * ```typespec\n * @format(\"uuid\")\n * scalar uuid extends string;\n * ```\n */\nextern dec format(target: string | bytes | ModelProperty, format: valueof string);\n\n/**\n * Specify the the pattern this string should respect using simple regular expression syntax.\n * The following syntax is allowed: alternations (`|`), quantifiers (`?`, `*`, `+`, and `{ }`), wildcard (`.`), and grouping parentheses.\n * Advanced features like look-around, capture groups, and references are not supported.\n *\n * This decorator may optionally provide a custom validation _message_. Emitters may choose to use the message to provide\n * context when pattern validation fails. For the sake of consistency, the message should be a phrase that describes in\n * plain language what sort of content the pattern attempts to validate. For example, a complex regular expression that\n * validates a GUID string might have a message like \"Must be a valid GUID.\"\n *\n * @param pattern Regular expression.\n * @param validationMessage Optional validation message that may provide context when validation fails.\n *\n * @example\n * ```typespec\n * @pattern(\"[a-z]+\", \"Must be a string consisting of only lower case letters and of at least one character.\")\n * scalar LowerAlpha extends string;\n * ```\n */\nextern dec pattern(\n  target: string | bytes | ModelProperty,\n  pattern: valueof string,\n  validationMessage?: valueof string\n);\n\n/**\n * Specify the minimum length this string type should be.\n * @param value Minimum length\n *\n * @example\n * ```typespec\n * @minLength(2)\n * scalar Username extends string;\n * ```\n */\nextern dec minLength(target: string | ModelProperty, value: valueof integer);\n\n/**\n * Specify the maximum length this string type should be.\n * @param value Maximum length\n *\n * @example\n * ```typespec\n * @maxLength(20)\n * scalar Username extends string;\n * ```\n */\nextern dec maxLength(target: string | ModelProperty, value: valueof integer);\n\n/**\n * Specify the minimum number of items this array should have.\n * @param value Minimum number\n *\n * @example\n * ```typespec\n * @minItems(1)\n * model Endpoints is string[];\n * ```\n */\nextern dec minItems(target: unknown[] | ModelProperty, value: valueof integer);\n\n/**\n * Specify the maximum number of items this array should have.\n * @param value Maximum number\n *\n * @example\n * ```typespec\n * @maxItems(5)\n * model Endpoints is string[];\n * ```\n */\nextern dec maxItems(target: unknown[] | ModelProperty, value: valueof integer);\n\n/**\n * Specify the minimum value this numeric type should be.\n * @param value Minimum value\n *\n * @example\n * ```typespec\n * @minValue(18)\n * scalar Age is int32;\n * ```\n */\nextern dec minValue(target: numeric | ModelProperty, value: valueof numeric);\n\n/**\n * Specify the maximum value this numeric type should be.\n * @param value Maximum value\n *\n * @example\n * ```typespec\n * @maxValue(200)\n * scalar Age is int32;\n * ```\n */\nextern dec maxValue(target: numeric | ModelProperty, value: valueof numeric);\n\n/**\n * Specify the minimum value this numeric type should be, exclusive of the given\n * value.\n * @param value Minimum value\n *\n * @example\n * ```typespec\n * @minValueExclusive(0)\n * scalar distance is float64;\n * ```\n */\nextern dec minValueExclusive(target: numeric | ModelProperty, value: valueof numeric);\n\n/**\n * Specify the maximum value this numeric type should be, exclusive of the given\n * value.\n * @param value Maximum value\n *\n * @example\n * ```typespec\n * @maxValueExclusive(50)\n * scalar distance is float64;\n * ```\n */\nextern dec maxValueExclusive(target: numeric | ModelProperty, value: valueof numeric);\n\n/**\n * Mark this string as a secret value that should be treated carefully to avoid exposure\n *\n * @example\n * ```typespec\n * @secret\n * scalar Password is string;\n * ```\n */\nextern dec secret(target: string | ModelProperty);\n\n/**\n * Attaches a tag to an operation, interface, or namespace. Multiple `@tag` decorators can be specified to attach multiple tags to a TypeSpec element.\n * @param tag Tag value\n */\nextern dec tag(target: Namespace | Interface | Operation, tag: valueof string);\n\n/**\n * Specifies how a templated type should name their instances.\n * @param name name the template instance should take\n * @param formatArgs Model with key value used to interpolate the name\n *\n * @example\n * ```typespec\n * @friendlyName(\"{name}List\", T)\n * model List<Item> {\n *   value: Item[];\n *   nextLink: string;\n * }\n * ```\n */\nextern dec friendlyName(target: unknown, name: valueof string, formatArgs?: unknown);\n\n/**\n * Provide a set of known values to a string type.\n * @param values Known values enum.\n *\n * @example\n * ```typespec\n * @knownValues(KnownErrorCode)\n * scalar ErrorCode extends string;\n *\n * enum KnownErrorCode {\n *   NotFound,\n *   Invalid,\n * }\n * ```\n */\n#deprecated \"This decorator has been deprecated. Use a named union of string literals with a string variant to achieve the same result without a decorator.\"\nextern dec knownValues(target: string | numeric | ModelProperty, values: Enum);\n\n/**\n * Mark a model property as the key to identify instances of that type\n * @param altName Name of the property. If not specified, the decorated property name is used.\n *\n * @example\n * ```typespec\n * model Pet {\n *   @key id: string;\n * }\n * ```\n */\nextern dec key(target: ModelProperty, altName?: valueof string);\n\n/**\n * Specify this operation is an overload of the given operation.\n * @param overloadbase Base operation that should be a union of all overloads\n *\n * @example\n * ```typespec\n * op upload(data: string | bytes, @header contentType: \"text/plain\" | \"application/octet-stream\"): void;\n * @overload(upload)\n * op uploadString(data: string, @header contentType: \"text/plain\" ): void;\n * @overload(upload)\n * op uploadBytes(data: bytes, @header contentType: \"application/octet-stream\"): void;\n * ```\n */\nextern dec overload(target: Operation, overloadbase: Operation);\n\n/**\n * DEPRECATED: Use `@encodedName` instead.\n *\n * Provide an alternative name for this type.\n * @param targetName Projection target\n * @param projectedName Alternative name\n *\n * @example\n * ```typespec\n * model Certificate {\n *   @projectedName(\"json\", \"exp\")\n *   expireAt: int32;\n * }\n * ```\n */\n#deprecated \"Use `@encodedName` instead for changing the name over the wire.\"\nextern dec projectedName(\n  target: unknown,\n  targetName: valueof string,\n  projectedName: valueof string\n);\n\n/**\n * Provide an alternative name for this type when serialized to the given mime type.\n * @param mimeType Mime type this should apply to. The mime type should be a known mime type as described here https://developer.mozilla.org/en-US/docs/Web/HTTP/Basics_of_HTTP/MIME_types/Common_types without any suffix (e.g. `+json`)\n * @param name Alternative name\n *\n * @example\n *\n * ```typespec\n * model Certificate {\n *   @encodedName(\"application/json\", \"exp\")\n *   @encodedName(\"application/xml\", \"expiry\")\n *   expireAt: int32;\n * }\n * ```\n *\n * @example Invalid values\n *\n * ```typespec\n * @encodedName(\"application/merge-patch+json\", \"exp\")\n *              ^ error cannot use subtype\n * ```\n */\nextern dec encodedName(target: unknown, mimeType: valueof string, name: valueof string);\n\n/**\n * Specify the property to be used to discriminate this type.\n * @param propertyName The property name to use for discrimination\n *\n * @example\n *\n * ```typespec\n * @discriminator(\"kind\")\n * union Pet{ cat: Cat, dog: Dog }\n *\n * model Cat {kind: \"cat\", meow: boolean}\n * model Dog {kind: \"dog\", bark: boolean}\n * ```\n *\n * ```typespec\n * @discriminator(\"kind\")\n * model Pet{ kind: string }\n *\n * model Cat extends Pet {kind: \"cat\", meow: boolean}\n * model Dog extends Pet  {kind: \"dog\", bark: boolean}\n * ```\n */\nextern dec discriminator(target: Model | Union, propertyName: valueof string);\n\n/**\n * Known encoding to use on utcDateTime or offsetDateTime\n */\nenum DateTimeKnownEncoding {\n  /**\n   * RFC 3339 standard. https://www.ietf.org/rfc/rfc3339.txt\n   * Encode to string.\n   */\n  rfc3339: \"rfc3339\",\n\n  /**\n   * RFC 7231 standard. https://www.ietf.org/rfc/rfc7231.txt\n   * Encode to string.\n   */\n  rfc7231: \"rfc7231\",\n\n  /**\n   * Encode a datetime to a unix timestamp.\n   * Unix timestamps are represented as an integer number of seconds since the Unix epoch and usually encoded as an int32.\n   */\n  unixTimestamp: \"unixTimestamp\",\n}\n\n/**\n * Known encoding to use on duration\n */\nenum DurationKnownEncoding {\n  /**\n   * ISO8601 duration\n   */\n  ISO8601: \"ISO8601\",\n\n  /**\n   * Encode to integer or float\n   */\n  seconds: \"seconds\",\n}\n\n/**\n * Known encoding to use on bytes\n */\nenum BytesKnownEncoding {\n  /**\n   * Encode to Base64\n   */\n  base64: \"base64\",\n\n  /**\n   * Encode to Base64 Url\n   */\n  base64url: \"base64url\",\n}\n\n/**\n * Encoding for serializing arrays\n */\nenum ArrayEncoding {\n  /** Each values of the array is separated by a | */\n  pipeDelimited,\n\n  /** Each values of the array is separated by a <space> */\n  spaceDelimited,\n}\n\n/**\n * Specify how to encode the target type.\n * @param encodingOrEncodeAs Known name of an encoding or a scalar type to encode as(Only for numeric types to encode as string).\n * @param encodedAs What target type is this being encoded as. Default to string.\n *\n * @example offsetDateTime encoded with rfc7231\n *\n * ```tsp\n * @encode(\"rfc7231\")\n * scalar myDateTime extends offsetDateTime;\n * ```\n *\n * @example utcDateTime encoded with unixTimestamp\n *\n * ```tsp\n * @encode(\"unixTimestamp\", int32)\n * scalar myDateTime extends unixTimestamp;\n * ```\n *\n * @example encode numeric type to string\n *\n * ```tsp\n * model Pet {\n *   @encode(string) id: int64;\n * }\n * ```\n */\nextern dec encode(\n  target: Scalar | ModelProperty,\n  encodingOrEncodeAs: (valueof string | EnumMember) | Scalar,\n  encodedAs?: Scalar\n);\n\n/** Options for example decorators */\nmodel ExampleOptions {\n  /** The title of the example */\n  title?: string;\n\n  /** Description of the example */\n  description?: string;\n}\n\n/**\n * Provide an example value for a data type.\n *\n * @param example Example value.\n * @param options Optional metadata for the example.\n *\n * @example\n *\n * ```tsp\n * @example(#{name: \"Fluffy\", age: 2})\n * model Pet {\n *  name: string;\n *  age: int32;\n * }\n * ```\n */\nextern dec example(\n  target: Model | Enum | Scalar | Union | ModelProperty | UnionVariant,\n  example: valueof unknown,\n  options?: valueof ExampleOptions\n);\n\n/**\n * Operation example configuration.\n */\nmodel OperationExample {\n  /** Example request body. */\n  parameters?: unknown;\n\n  /** Example response body. */\n  returnType?: unknown;\n}\n\n/**\n * Provide example values for an operation's parameters and corresponding return type.\n *\n * @param example Example value.\n * @param options Optional metadata for the example.\n *\n * @example\n *\n * ```tsp\n * @opExample(#{parameters: #{name: \"Fluffy\", age: 2}, returnType: #{name: \"Fluffy\", age: 2, id: \"abc\"})\n * op createPet(pet: Pet): Pet;\n * ```\n */\nextern dec opExample(\n  target: Operation,\n  example: valueof OperationExample,\n  options?: valueof ExampleOptions\n);\n\n/**\n * Returns the model with required properties removed.\n */\nextern dec withOptionalProperties(target: Model);\n\n/**\n * Returns the model with any default values removed.\n */\nextern dec withoutDefaultValues(target: Model);\n\n/**\n * Returns the model with the given properties omitted.\n * @param omit List of properties to omit\n */\nextern dec withoutOmittedProperties(target: Model, omit: string | Union);\n\n/**\n * Returns the model with only the given properties included.\n * @param pick List of properties to include\n */\nextern dec withPickedProperties(target: Model, pick: string | Union);\n\n//---------------------------------------------------------------------------\n// Paging\n//---------------------------------------------------------------------------\n\n/**\n * Mark this operation as a `list` operation that returns a paginated list of items.\n */\nextern dec list(target: Operation);\n\n/**\n * Pagination property defining the number of items to skip.\n * @example\n * ```tsp\n * model Page<T> {\n *   @pageItems items: T[];\n * }\n * @list op listPets(@offset skip: int32, @pageSize pageSize: int8): Page<Pet>;\n * ```\n */\nextern dec offset(target: ModelProperty);\n\n/**\n * Pagination property defining the page index.\n *\n * @example\n * ```tsp\n * model Page<T> {\n *   @pageItems items: T[];\n * }\n * @list op listPets(@pageIndex page: int32, @pageSize pageSize: int8): Page<Pet>;\n * ```\n */\nextern dec pageIndex(target: ModelProperty);\n\n/**\n * Specify the pagination parameter that controls the maximum number of items to include in a page.\n *\n * @example\n * ```tsp\n * model Page<T> {\n *   @pageItems items: T[];\n * }\n * @list op listPets(@pageIndex page: int32, @pageSize pageSize: int8): Page<Pet>;\n * ```\n */\nextern dec pageSize(target: ModelProperty);\n\n/**\n * Specify the the property that contains the array of page items.\n *\n * @example\n * ```tsp\n * model Page<T> {\n *   @pageItems items: T[];\n * }\n * @list op listPets(@pageIndex page: int32, @pageSize pageSize: int8): Page<Pet>;\n * ```\n */\nextern dec pageItems(target: ModelProperty);\n\n/**\n * Pagination property defining the token to get to the next page.\n * It MUST be specified both on the request parameter and the response.\n *\n * @example\n * ```tsp\n * model Page<T> {\n *   @pageItems items: T[];\n *   @continuationToken continuationToken: string;\n * }\n * @list op listPets(@continuationToken continuationToken: string): Page<Pet>;\n * ```\n */\nextern dec continuationToken(target: ModelProperty);\n\n/**\n * Pagination property defining a link to the next page.\n *\n * It is expected that navigating to the link will return the same set of responses as the operation that returned the current page.\n *\n * @example\n * ```tsp\n * model Page<T> {\n *   @pageItems items: T[];\n *   @nextLink next: url;\n *   @prevLink prev: url;\n *   @firstLink first: url;\n *   @lastLink last: url;\n * }\n * @list op listPets(): Page<Pet>;\n * ```\n */\nextern dec nextLink(target: ModelProperty);\n\n/**\n * Pagination property defining a link to the previous page.\n *\n * It is expected that navigating to the link will return the same set of responses as the operation that returned the current page.\n *\n * @example\n * ```tsp\n * model Page<T> {\n *   @pageItems items: T[];\n *   @nextLink next: url;\n *   @prevLink prev: url;\n *   @firstLink first: url;\n *   @lastLink last: url;\n * }\n * @list op listPets(): Page<Pet>;\n * ```\n */\nextern dec prevLink(target: ModelProperty);\n\n/**\n * Pagination property defining a link to the first page.\n *\n * It is expected that navigating to the link will return the same set of responses as the operation that returned the current page.\n *\n * @example\n * ```tsp\n * model Page<T> {\n *   @pageItems items: T[];\n *   @nextLink next: url;\n *   @prevLink prev: url;\n *   @firstLink first: url;\n *   @lastLink last: url;\n * }\n * @list op listPets(): Page<Pet>;\n * ```\n */\nextern dec firstLink(target: ModelProperty);\n\n/**\n * Pagination property defining a link to the last page.\n *\n * It is expected that navigating to the link will return the same set of responses as the operation that returned the current page.\n *\n * @example\n * ```tsp\n * model Page<T> {\n *   @pageItems items: T[];\n *   @nextLink next: url;\n *   @prevLink prev: url;\n *   @firstLink first: url;\n *   @lastLink last: url;\n * }\n * @list op listPets(): Page<Pet>;\n * ```\n */\nextern dec lastLink(target: ModelProperty);\n\n//---------------------------------------------------------------------------\n// Debugging\n//---------------------------------------------------------------------------\n\n/**\n * A debugging decorator used to inspect a type.\n * @param text Custom text to log\n */\nextern dec inspectType(target: unknown, text: valueof string);\n\n/**\n * A debugging decorator used to inspect a type name.\n * @param text Custom text to log\n */\nextern dec inspectTypeName(target: unknown, text: valueof string);\n",
  "../../core/packages/compiler/lib/std/reflection.tsp": "namespace TypeSpec.Reflection;\n\nmodel Enum {}\nmodel EnumMember {}\nmodel Interface {}\nmodel Model {}\nmodel ModelProperty {}\nmodel Namespace {}\nmodel Operation {}\nmodel Scalar {}\nmodel Union {}\nmodel UnionVariant {}\nmodel StringTemplate {}\n",
  "../../core/packages/compiler/lib/std/projected-names.tsp": "// Set of projections consuming the @projectedName decorator\n#suppress \"projections-are-experimental\"\nprojection op#target {\n  to(targetName) {\n    if hasProjectedName(self, targetName) {\n      self::rename(getProjectedName(self, targetName));\n    };\n  }\n  from(targetName) {\n    if hasProjectedName(self, targetName) {\n      self::rename(self::projectionBase::name);\n    };\n  }\n}\n\n#suppress \"projections-are-experimental\"\nprojection interface#target {\n  to(targetName) {\n    if hasProjectedName(self, targetName) {\n      self::rename(getProjectedName(self, targetName));\n    };\n  }\n  from(targetName) {\n    if hasProjectedName(self, targetName) {\n      self::rename(self::projectionBase::name);\n    };\n  }\n}\n\n#suppress \"projections-are-experimental\"\nprojection model#target {\n  to(targetName) {\n    if hasProjectedName(self, targetName) {\n      self::rename(getProjectedName(self, targetName));\n    };\n\n    self::properties::forEach((p) => {\n      if hasProjectedName(p, targetName) {\n        self::renameProperty(p::name, getProjectedName(p, targetName));\n      };\n    });\n  }\n  from(targetName) {\n    if hasProjectedName(self, targetName) {\n      self::rename(self::projectionBase::name);\n    };\n\n    self::projectionBase::properties::forEach((p) => {\n      if hasProjectedName(p, targetName) {\n        self::renameProperty(getProjectedName(p, targetName), p::name);\n      };\n    });\n  }\n}\n\n#suppress \"projections-are-experimental\"\nprojection enum#target {\n  to(targetName) {\n    if hasProjectedName(self, targetName) {\n      self::rename(getProjectedName(self, targetName));\n    };\n\n    self::members::forEach((p) => {\n      if hasProjectedName(p, targetName) {\n        self::renameMember(p::name, getProjectedName(p, targetName));\n      };\n    });\n  }\n  from(targetName) {\n    if hasProjectedName(self, targetName) {\n      self::rename(self::projectionBase::name);\n    };\n\n    self::projectionBase::members::forEach((p) => {\n      if hasProjectedName(p, targetName) {\n        self::renameMember(getProjectedName(p, targetName), p::name);\n      };\n    });\n  }\n}\n\n#suppress \"projections-are-experimental\"\nprojection union#target {\n  to(targetName) {\n    if hasProjectedName(self, targetName) {\n      self::rename(getProjectedName(self, targetName));\n    };\n  }\n  from(targetName) {\n    if hasProjectedName(self, targetName) {\n      self::rename(self::projectionBase::name);\n    };\n  }\n}\n",
  "../../core/packages/compiler/lib/std/visibility.tsp": "// Copyright (c) Microsoft Corporation\n// Licensed under the MIT license.\n\nimport \"../../dist/src/lib/tsp-index.js\";\n\nusing TypeSpec.Reflection;\n\nnamespace TypeSpec;\n\n/**\n * Indicates that a property is only considered to be present or applicable (\"visible\") with\n * the in the given named contexts (\"visibilities\"). When a property has no visibilities applied\n * to it, it is implicitly visible always.\n *\n * As far as the TypeSpec core library is concerned, visibilities are open-ended and can be arbitrary\n * strings, but  the following visibilities are well-known to standard libraries and should be used\n * with standard emitters that interpret them as follows:\n *\n * - \"read\": output of any operation.\n * - \"create\": input to operations that create an entity..\n * - \"query\": input to operations that read data.\n * - \"update\": input to operations that update data.\n * - \"delete\": input to operations that delete data.\n *\n * See also: [Automatic visibility](https://typespec.io/docs/libraries/http/operations#automatic-visibility)\n *\n * @param visibilities List of visibilities which apply to this property.\n *\n * @example\n *\n * ```typespec\n * model Dog {\n *   // the service will generate an ID, so you don't need to send it.\n *   @visibility(Lifecycle.Read) id: int32;\n *   // the service will store this secret name, but won't ever return it\n *   @visibility(Lifecycle.Create, Lifecycle.Update) secretName: string;\n *   // the regular name is always present\n *   name: string;\n * }\n * ```\n */\nextern dec visibility(target: ModelProperty, ...visibilities: valueof (string | EnumMember)[]);\n\n/**\n * Indicates that a property is not visible in the given visibility class.\n *\n * This decorator removes all active visibility modifiers from the property within\n * the given visibility class.\n *\n * @param visibilityClass The visibility class to make the property invisible within.\n *\n * @example\n * ```typespec\n * model Example {\n *   @invisible(Lifecycle)\n *   hidden_property: string;\n * }\n * ```\n */\nextern dec invisible(target: ModelProperty, visibilityClass: Enum);\n\n/**\n * Removes visibility modifiers from a property.\n *\n * If the visibility modifiers for a visibility class have not been initialized,\n * this decorator will use the default visibility modifiers for the visibility\n * class as the default modifier set.\n *\n * @param target The property to remove visibility from.\n * @param visibilities The visibility modifiers to remove from the target property.\n *\n * @example\n * ```typespec\n * model Example {\n *   // This property will have the Create and Update visibilities, but not the\n *   // Read visibility, since it is removed.\n *   @removeVisibility(Lifecycle.Read)\n *   secret_property: string;\n * }\n * ```\n */\nextern dec removeVisibility(target: ModelProperty, ...visibilities: valueof EnumMember[]);\n\n/**\n * Removes properties that are not considered to be present or applicable\n * (\"visible\") in the given named contexts (\"visibilities\"). Can be used\n * together with spread to effectively spread only visible properties into\n * a new model.\n *\n * See also: [Automatic visibility](https://typespec.io/docs/libraries/http/operations#automatic-visibility)\n *\n * When using an emitter that applies visibility automatically, it is generally\n * not necessary to use this decorator.\n *\n * @param visibilities List of visibilities which apply to this property.\n *\n * @example\n * ```typespec\n * model Dog {\n *   @visibility(\"read\") id: int32;\n *   @visibility(\"create\", \"update\") secretName: string;\n *   name: string;\n * }\n *\n * // The spread operator will copy all the properties of Dog into DogRead,\n * // and @withVisibility will then remove those that are not visible with\n * // create or update visibility.\n * //\n * // In this case, the id property is removed, and the name and secretName\n * // properties are kept.\n * @withVisibility(\"create\", \"update\")\n * model DogCreateOrUpdate {\n *   ...Dog;\n * }\n *\n * // In this case the id and name properties are kept and the secretName property\n * // is removed.\n * @withVisibility(\"read\")\n * model DogRead {\n *   ...Dog;\n * }\n * ```\n */\nextern dec withVisibility(target: Model, ...visibilities: valueof (string | EnumMember)[]);\n\n/**\n * Set the visibility of key properties in a model if not already set.\n *\n * This will set the visibility modifiers of all key properties in the model if the visibility is not already _explicitly_ set,\n * but will not change the visibility of any properties that have visibility set _explicitly_, even if the visibility\n * is the same as the default visibility.\n *\n * Visibility may be explicitly set using any of the following decorators:\n *\n * - `@visibility`\n * - `@removeVisibility`\n * - `@invisible`\n *\n * @param visibility The desired default visibility value. If a key property already has visibility set, it will not be changed.\n */\nextern dec withDefaultKeyVisibility(target: Model, visibility: valueof string | EnumMember);\n\n/**\n * Sets which visibilities apply to parameters for the given operation.\n *\n * @param visibilities List of visibility strings which apply to this operation.\n */\nextern dec parameterVisibility(target: Operation, ...visibilities: valueof (string | EnumMember)[]);\n\n/**\n * Sets which visibilities apply to the return type for the given operation.\n * @param visibilities List of visibility strings which apply to this operation.\n */\nextern dec returnTypeVisibility(\n  target: Operation,\n  ...visibilities: valueof (string | EnumMember)[]\n);\n\n/**\n * Returns the model with non-updateable properties removed.\n */\nextern dec withUpdateableProperties(target: Model);\n\n/**\n * Declares the default visibility modifiers for a visibility class.\n *\n * The default modifiers are used when a property does not have any visibility decorators\n * applied to it.\n *\n * The modifiers passed to this decorator _MUST_ be members of the target Enum.\n *\n * @param visibilities the list of modifiers to use as the default visibility modifiers.\n */\nextern dec defaultVisibility(target: Enum, ...visibilities: valueof EnumMember[]);\n\n/**\n * A visibility class for resource lifecycle phases.\n *\n * These visibilities control whether a property is visible during the create, read, and update phases of a resource's\n * lifecycle.\n *\n * @example\n * ```typespec\n * model Dog {\n *  @visibility(Lifecycle.Read) id: int32;\n *  @visibility(Lifecycle.Create, Lifecycle.Update) secretName: string;\n *  name: string;\n * }\n * ```\n *\n * In this example, the `id` property is only visible during the read phase, and the `secretName` property is only visible\n * during the create and update phases. This means that the server will return the `id` property when returning a `Dog`,\n * but the client will not be able to set or update it. In contrast, the `secretName` property can be set when creating\n * or updating a `Dog`, but the server will never return it. The `name` property has no visibility modifiers and is\n * therefore visible in all phases.\n */\nenum Lifecycle {\n  Create,\n  Read,\n  Update,\n}\n\n/**\n * A visibility filter, used to specify which properties should be included when\n * using the `withVisibilityFilter` decorator.\n *\n * The filter matches any property with ALL of the following:\n * - If the `any` key is present, the property must have at least one of the specified visibilities.\n * - If the `all` key is present, the property must have all of the specified visibilities.\n * - If the `none` key is present, the property must have none of the specified visibilities.\n */\nmodel VisibilityFilter {\n  any?: EnumMember[];\n  all?: EnumMember[];\n  none?: EnumMember[];\n}\n\n/**\n * Applies the given visibility filter to the properties of the target model.\n *\n * This transformation is recursive, so it will also apply the filter to any nested\n * or referenced models that are the types of any properties in the `target`.\n *\n * @param target The model to apply the visibility filter to.\n * @param filter The visibility filter to apply to the properties of the target model.\n *\n * @example\n * ```typespec\n * model Dog {\n *   @visibility(Lifecycle.Read)\n *   id: int32;\n *\n *   name: string;\n * }\n *\n * @withVisibilityFilter(#{ all: #[Lifecycle.Read] })\n * model DogRead {\n *  ...Dog\n * }\n * ```\n */\nextern dec withVisibilityFilter(target: Model, filter: valueof VisibilityFilter);\n\n/**\n * Transforms the `target` model to include only properties that are visible during the\n * \"Update\" lifecycle phase.\n *\n * Any nested models of optional properties will be transformed into the \"CreateOrUpdate\"\n * lifecycle phase instead of the \"Update\" lifecycle phase, so that nested models may be\n * fully updated.\n *\n * @param target The model to apply the transformation to.\n *\n * @example\n * ```typespec\n * model Dog {\n *   @visibility(Lifecycle.Read)\n *   id: int32;\n *\n *   @visibility(Lifecycle.Create, Lifecycle.Update)\n *   secretName: string;\n *\n *   name: string;\n * }\n *\n * @withLifecycleUpdate\n * model DogUpdate {\n *   ...Dog\n * }\n * ```\n */\nextern dec withLifecycleUpdate(target: Model);\n\n/**\n * A copy of the input model `T` with only the properties that are visible during the\n * \"Create\" resource lifecycle phase.\n *\n * This transformation is recursive, and will include only properties that have the\n * `Lifecycle.Create` visibility modifier.\n *\n * If a `NameTemplate` is provided, the new model will be named according to the template.\n * The template uses the same syntax as the `@friendlyName` decorator.\n *\n * @template T The model to transform.\n * @template NameTemplate The name template to use for the new model.\n *\n *  * @example\n * ```typespec\n * model Dog {\n *   @visibility(Lifecycle.Read)\n *   id: int32;\n *\n *   name: string;\n * }\n *\n * model CreateDog is Create<Dog>;\n * ```\n */\n@friendlyName(NameTemplate, T)\n@withVisibilityFilter(#{ all: #[Lifecycle.Create] })\nmodel Create<T extends Reflection.Model, NameTemplate extends valueof string = \"Create{name}\"> {\n  ...T;\n}\n\n/**\n * A copy of the input model `T` with only the properties that are visible during the\n * \"Read\" resource lifecycle phase.\n *\n * This transformation is recursive, and will include only properties that have the\n * `Lifecycle.Read` visibility modifier.\n *\n * If a `NameTemplate` is provided, the new model will be named according to the template.\n * The template uses the same syntax as the `@friendlyName` decorator.\n *\n * @template T The model to transform.\n * @template NameTemplate The name template to use for the new model.\n *\n *  * @example\n * ```typespec\n * model Dog {\n *   @visibility(Lifecycle.Read)\n *   id: int32;\n *\n *   name: string;\n * }\n *\n * model ReadDog is Read<Dog>;\n * ```\n */\n@friendlyName(NameTemplate, T)\n@withVisibilityFilter(#{ all: #[Lifecycle.Read] })\nmodel Read<T extends Reflection.Model, NameTemplate extends valueof string = \"Read{name}\"> {\n  ...T;\n}\n\n/**\n * A copy of the input model `T` with only the properties that are visible during the\n * \"Update\" resource lifecycle phase.\n *\n * This transformation will include only the properties that have the `Lifecycle.Update`\n * visibility modifier, and the types of all properties will be replaced with the\n * equivalent `CreateOrUpdate` transformation.\n *\n * If a `NameTemplate` is provided, the new model will be named according to the template.\n * The template uses the same syntax as the `@friendlyName` decorator.\n *\n * @template T The model to transform.\n * @template NameTemplate The name template to use for the new model.\n *\n *  * @example\n * ```typespec\n * model Dog {\n *   @visibility(Lifecycle.Read)\n *   id: int32;\n *\n *   name: string;\n * }\n *\n * model UpdateDog is Update<Dog>;\n * ```\n */\n@friendlyName(NameTemplate, T)\n@withLifecycleUpdate\nmodel Update<T extends Reflection.Model, NameTemplate extends valueof string = \"Update{name}\"> {\n  ...T;\n}\n\n/**\n * A copy of the input model `T` with only the properties that are visible during the\n * \"Create\" or \"Update\" resource lifecycle phases.\n *\n * This transformation is recursive, and will include only properties that have the\n * `Lifecycle.Create` or `Lifecycle.Update` visibility modifier.\n *\n * If a `NameTemplate` is provided, the new model will be named according to the template.\n * The template uses the same syntax as the `@friendlyName` decorator.\n *\n * @template T The model to transform.\n * @template NameTemplate The name template to use for the new model.\n *\n *  * @example\n * ```typespec\n * model Dog {\n *   @visibility(Lifecycle.Read)\n *   id: int32;\n *\n *   name: string;\n * }\n *\n * model CreateOrUpdateDog is CreateOrUpdate<Dog>;\n * ```\n */\n@friendlyName(NameTemplate, T)\n@withVisibilityFilter(#{ any: #[Lifecycle.Create, Lifecycle.Update] })\nmodel CreateOrUpdate<\n  T extends Reflection.Model,\n  NameTemplate extends valueof string = \"CreateOrUpdate{name}\"\n> {\n  ...T;\n}\n",
  "lib/main.tsp": "import \"./decorators.tsp\";\nimport \"./augmentCore.tsp\";\nimport \"../dist/src/tsp-index.js\";\n",
  "lib/decorators.tsp": "using TypeSpec.Reflection;\n\nnamespace Azure.ClientGenerator.Core;\n\n/**\n * Changes the name of a method, parameter, property, or model generated in the client SDK\n * @param rename The rename you want applied to the object\n * @param scope The language scope you want this decorator to apply to. If not specified, will apply to all language emitters.\n * You can use \"!\" to specify negation such as \"!(java, python)\" or \"!java, !python\".\n *\n * @example\n * ```typespec\n * @clientName(\"nameInClient\")\n * op nameInService: void;\n * ```\n *\n * @example\n * ```typespec\n * @clientName(\"nameForJava\", \"java\")\n * @clientName(\"name_for_python\", \"python\")\n * @clientName(\"nameForCsharp\", \"csharp\")\n * @clientName(\"nameForJavascript\", \"javascript\")\n * op nameInService: void;\n * ```\n */\nextern dec clientName(target: unknown, rename: valueof string, scope?: valueof string);\n\n/**\n * Whether you want to generate an operation as a convenient operation.\n * @param value Whether to generate the operation as convenience method or not.\n * @param scope The language scope you want this decorator to apply to. If not specified, will apply to all language emitters.\n * You can use \"!\" to specify negation such as \"!(java, python)\" or \"!java, !python\".\n *\n * @example\n * ```typespec\n * @convenientAPI(false)\n * op test: void;\n * ```\n */\nextern dec convenientAPI(target: Operation, value?: valueof boolean, scope?: valueof string);\n\n/**\n * Whether you want to generate an operation as a protocol operation.\n * @param value Whether to generate the operation as protocol or not.\n * @param scope The language scope you want this decorator to apply to. If not specified, will apply to all language emitters.\n * You can use \"!\" to specify negation such as \"!(java, python)\" or \"!java, !python\".\n *\n * @example\n * ```typespec\n * @protocolAPI(false)\n * op test: void;\n * ```\n */\nextern dec protocolAPI(target: Operation, value?: valueof boolean, scope?: valueof string);\n\n/**\n * Create a ClientGenerator.Core client out of a namespace or interface\n * @param value Optional configuration for the service.\n * @param scope The language scope you want this decorator to apply to. If not specified, will apply to all language emitters.\n * You can use \"!\" to specify negation such as \"!(java, python)\" or \"!java, !python\".\n *\n * @example Basic client setting\n * ```typespec\n * @client\n * namespace MyService {}\n * ```\n *\n * @example Setting with other service\n * ```typespec\n * namespace MyService {}\n *\n * @client({service: MyService})\n * interface MyInterface {}\n * ```\n *\n * @example Changing client name if you don't want <Interface/Namespace>Client\n * ```typespec\n * @client({client: MySpecialClient})\n * interface MyInterface {}\n * ```\n *\n * @example\n */\nextern dec client(target: Namespace | Interface, value?: Model, scope?: valueof string);\n\n/**\n * Create a ClientGenerator.Core operation group out of a namespace or interface\n * @param scope The language scope you want this decorator to apply to. If not specified, will apply to all language emitters.\n * You can use \"!\" to specify negation such as \"!(java, python)\" or \"!java, !python\".\n *\n * @example\n * ```typespec\n * @operationGroup\n * interface MyInterface{}\n * ```\n */\nextern dec operationGroup(target: Namespace | Interface, scope?: valueof string);\n\n/**\n * Usage value.\n */\nenum Usage {\n  /**\n   * Used in request\n   */\n  input: 2,\n\n  /**\n   * Used in response\n   */\n  output: 4,\n}\n\n/**\n * Override usage for models/enums.\n * A model/enum's default usage info is always calculated by the operations that use it.\n * You could use this decorator to override the default usage info.\n * When setting usage for namespaces,\n * the usage info will be propagated to the models defined in the namespace.\n * If the model has an usage override, the model override takes precedence.\n * For example, with operation definition `op test(): OutputModel`,\n * the model `OutputModel` has default usage `Usage.output`.\n * After adding decorator `@@usage(OutputModel, Usage.input | Usage.output)`,\n * the final usage result for `OutputModel` is `Usage.input | Usage.output`.\n * The usage info for models will be propagated to models' properties,\n * parent models, discriminated sub models.\n * The override usage should not be narrow than the usage calculated by operation,\n * and different override usage should not conflict with each other,\n * otherwise a warning will be added to diagnostics list.\n * @param value The usage info you want to set for this model.\n * @param scope The language scope you want this decorator to apply to. If not specified, will apply to all language emitters.\n * You can use \"!\" to specify negation such as \"!(java, python)\" or \"!java, !python\".\n *\n * @example Expand usage for model\n * ```typespec\n * op test(): OutputModel;\n *\n * // usage result for `OutputModel` is `Usage.input | Usage.output`\n * @usage(Usage.input)\n * model OutputModel {\n *   prop: string\n * }\n * ```\n *\n * @example Propagation of usage\n * ```typespec\n * // Usage.output\n * @discriminator(\"kind\")\n * model Fish {\n *   age: int32;\n * }\n *\n * // Usage.input | Usage.output\n * @discriminator(\"sharktype\")\n * @usage(Usage.input)\n * model Shark extends Fish {\n *   kind: \"shark\";\n *   origin: Origin;\n * }\n *\n * // Usage.output\n * model Salmon extends Fish {\n *   kind: \"salmon\";\n * }\n *\n * // Usage.output\n * model SawShark extends Shark {\n *   sharktype: \"saw\";\n * }\n *\n * // Usage.output\n * model Origin {\n *   country: string;\n *   city: string;\n *   manufacture: string;\n * }\n *\n * @get\n * op getModel(): Fish;\n * ```\n */\nextern dec usage(\n  target: Model | Enum | Union | Namespace,\n  value: EnumMember | Union,\n  scope?: valueof string\n);\n\n/**\n * Access value.\n */\nenum Access {\n  /**\n   * Open to user\n   */\n  public: \"public\",\n\n  /**\n   * Hide from user\n   */\n  internal: \"internal\",\n}\n\n/**\n * Override access for operations, models and enums.\n * When setting access for namespaces,\n * the access info will be propagated to the models and operations defined in the namespace.\n * If the model has an access override, the model override takes precedence.\n * When setting access for an operation,\n * it will influence the access info for models/enums that are used by this operation.\n * Models/enums that are used in any operations with `@access(Access.public)` will be set to access \"public\"\n * Models/enums that are only used in operations with `@access(Access.internal)` will be set to access \"internal\".\n * The access info for models will be propagated to models' properties,\n * parent models, discriminated sub models.\n * The override access should not be narrow than the access calculated by operation,\n * and different override access should not conflict with each other,\n * otherwise a warning will be added to diagnostics list.\n * @param value The access info you want to set for this model or operation.\n * @param scope The language scope you want this decorator to apply to. If not specified, will apply to all language emitters.\n * You can use \"!\" to specify negation such as \"!(java, python)\" or \"!java, !python\".\n *\n * @example Set access\n * ```typespec\n * // Access.internal\n * @access(Access.internal)\n * model ModelToHide {\n *   prop: string;\n * }\n * // Access.internal\n * @access(Access.internal)\n * op test: void;\n * ```\n * @example Access propagation\n * ```typespec\n * // Access.internal\n * @discriminator(\"kind\")\n * model Fish {\n *   age: int32;\n * }\n *\n * // Access.internal\n * @discriminator(\"sharktype\")\n * model Shark extends Fish {\n *   kind: \"shark\";\n *   origin: Origin;\n * }\n *\n * // Access.internal\n * model Salmon extends Fish {\n *   kind: \"salmon\";\n * }\n *\n * // Access.internal\n * model SawShark extends Shark {\n *   sharktype: \"saw\";\n * }\n *\n * // Access.internal\n * model Origin {\n *   country: string;\n *   city: string;\n *   manufacture: string;\n * }\n *\n * // Access.internal\n * @get\n * @access(Access.internal)\n * op getModel(): Fish;\n * ```\n * @example Access influence from operation\n * ```typespec\n * // Access.internal\n * model Test1 {\n * }\n *\n * // Access.internal\n * @access(Access.internal)\n * @route(\"/func1\")\n * op func1(\n *   @body body: Test1\n * ): void;\n *\n * // Access.public\n * model Test2 {\n * }\n *\n * // Access.public\n * @route(\"/func2\")\n * op func2(\n *   @body body: Test2\n * ): void;\n *\n * // Access.public\n * model Test3 {\n * }\n *\n * // Access.public\n * @access(Access.public)\n * @route(\"/func3\")\n * op func3(\n *   @body body: Test3\n * ): void;\n *\n * // Access.public\n * model Test4 {\n * }\n *\n * // Access.internal\n * @access(Access.internal)\n * @route(\"/func4\")\n * op func4(\n *   @body body: Test4\n * ): void;\n *\n * // Access.public\n * @route(\"/func5\")\n * op func5(\n *   @body body: Test4\n * ): void;\n *\n * // Access.public\n * model Test5 {\n * }\n *\n * // Access.internal\n * @access(Access.internal)\n * @route(\"/func6\")\n * op func6(\n *   @body body: Test5\n * ): void;\n *\n * // Access.public\n * @route(\"/func7\")\n * op func7(\n *   @body body: Test5\n * ): void;\n *\n * // Access.public\n * @access(Access.public)\n * @route(\"/func8\")\n * op func8(\n *   @body body: Test5\n * ): void;\n * ```\n */\nextern dec access(\n  target: Model | Operation | Enum | Union | Namespace,\n  value: EnumMember,\n  scope?: valueof string\n);\n\n/**\n * Set whether a model property should be flattened or not.\n * @param scope The language scope you want this decorator to apply to. If not specified, will apply to all language emitters.\n * You can use \"!\" to specify negation such as \"!(java, python)\" or \"!java, !python\".\n *\n * @example\n * ```typespec\n * model Foo {\n *    @flattenProperty\n *    prop: Bar;\n * }\n * model Bar {\n * }\n * ```\n */\n#deprecated \"@flattenProperty decorator is not recommended to use.\"\nextern dec flattenProperty(target: ModelProperty, scope?: valueof string);\n\n/**\n * Override the default client method generated by TCGC from your service definition\n * @param original: The original service definition\n * @param override: The override method definition that specifies the exact client method you want\n * @param scope The language scope you want this decorator to apply to. If not specified, will apply to all language emitters.\n * You can use \"!\" to specify negation such as \"!(java, python)\" or \"!java, !python\".\n *\n * @example\n * ```typespec\n * // main.tsp\n * namespace MyService;\n *\n * model Params {\n *  foo: string;\n *  bar: string;\n * }\n * op myOperation(...Params): void; // by default, we generate the method signature as `op myOperation(foo: string, bar: string)`;\n *\n * // client.tsp\n * namespace MyCustomizations;\n *\n * @override(MyService.operation)\n * op myOperationCustomization(params: Params): void;\n *\n * // method signature is now `op myOperation(params: Params)`\n * ```\n *\n * @example\n * ```typespec\n * // main.tsp\n * namespace MyService;\n *\n * model Params {\n *  foo: string;\n *  bar: string;\n * }\n * op myOperation(...Params): void; // by default, we generate the method signature as `op myOperation(foo: string, bar: string)`;\n *\n * // client.tsp\n * namespace MyCustomizations;\n *\n * @override(MyService.operation, \"csharp\")\n * op myOperationCustomization(params: Params): void;\n *\n * // method signature is now `op myOperation(params: Params)` just for csharp\n * ```\n */\nextern dec override(original: Operation, override: Operation, scope?: valueof string);\n\n/**\n * Whether a model needs the custom JSON converter, this is only used for backward compatibility for csharp.\n * @param scope The language scope you want this decorator to apply to. If not specified, will apply to all language emitters.\n * You can use \"!\" to specify negation such as \"!(java, python)\" or \"!java, !python\".\n *\n * @example\n * ```typespec\n * @useSystemTextJsonConverter\n * model MyModel {\n *   prop: string;\n * }\n * ```\n */\nextern dec useSystemTextJsonConverter(target: Model, scope?: valueof string);\n\n/**\n * Client parameters you would like to add to the client. By default, we apply endpoint, credential, and api-version parameters. If you add clientInitialization, we will append those to the default list of parameters.\n * @param scope The language scope you want this decorator to apply to. If not specified, will apply to all language emitters.\n * You can use \"!\" to specify negation such as \"!(java, python)\" or \"!java, !python\".\n *\n * @example\n * ```typespec\n * // main.tsp\n * namespace MyService;\n *\n * op upload(blobName: string): void;\n * op download(blobName: string): void;\n *\n * // client.tsp\n * namespace MyCustomizations;\n * model MyServiceClientOptions {\n *   blobName: string;\n * }\n *\n * @@clientInitialization(MyService, MyServiceClientOptions)\n * // The generated client will have `blobName` on it. We will also\n * // elevate the existing `blobName` parameter to the client level.\n * ```\n */\nextern dec clientInitialization(\n  target: Namespace | Interface,\n  options: Model,\n  scope?: valueof string\n);\n\n/**\n * Alias the name of a client parameter to a different name. This permits you to have a different name for the parameter in client initialization then on individual methods and still refer to the same parameter.\n * @param scope The language scope you want this decorator to apply to. If not specified, will apply to all language emitters.\n * You can use \"!\" to specify negation such as \"!(java, python)\" or \"!java, !python\".\n *\n * @example\n * ```typespec\n * // main.tsp\n * namespace MyService;\n *\n * op upload(blobName: string): void;\n *\n * // client.tsp\n * namespace MyCustomizations;\n * model MyServiceClientOptions {\n *   blob: string;\n * }\n *\n * @@clientInitialization(MyService, MyServiceClientOptions)\n * @@paramAlias(MyServiceClientOptions.blob, \"blobName\")\n *\n * // The generated client will have `blobName` on it. We will also\n * // elevate the existing `blob` parameter to the client level.\n * ```\n */\nextern dec paramAlias(original: ModelProperty, paramAlias: valueof string, scope?: valueof string);\n\n/**\n * Changes the namespace of a client, model, enum or union generated in the client SDK.\n * By default, the client namespace for them will follow the TypeSpec namespace.\n * @param rename The rename you want applied to the object\n * @param scope The language scope you want this decorator to apply to. If not specified, will apply to all language emitters.\n * You can use \"!\" to specify negation such as \"!(java, python)\" or \"!java, !python\".\n *\n * @example\n * ```typespec\n * @clientNamespace(\"ContosoClient\")\n * namespace Contoso;\n * ```\n *\n * @example\n * ```typespec\n * @clientNamespace(\"ContosoJava\", \"java\")\n * @clientNamespace(\"ContosoPython\", \"python\")\n * @clientNamespace(\"ContosoCSharp\", \"csharp\")\n * @clientNamespace(\"ContosoJavascript\", \"javascript\")\n * namespace Contoso;\n * ```\n */\nextern dec clientNamespace(\n  target: Namespace | Interface | Model | Enum | Union,\n  rename: valueof string,\n  scope?: valueof string\n);\n\n/**\n * Set an alternate type for a model property, scalar, or function parameter. Note that `@encode` will be overridden by the one defined in alternate type.\n *\n * @param source The source type you want to apply the alternate type to. Only scalar types are supported.\n * @param alternate The alternate type you want applied to the target. Only scalar types are supported.\n * @param scope The language scope you want this decorator to apply to. If not specified, will apply to all language emitters.\n * You can use \"!\" to specify negation such as \"!(java, python)\" or \"!java, !python\".\n * @example\n * ```typespec\n * model Foo {\n *    date: utcDateTime;\n * }\n * @@alternateType(Foo.date, string);\n * ```\n * @example\n * ```typespec\n * scalar storageDateTime extends utcDataTime;\n * @@alternateType(storageDateTime, string, \"python\");\n * ```\n * @example\n * ```typespec\n * op test(@param @alternateType(string) date: utcDateTime): void;\n * ```\n */\nextern dec alternateType(source: ModelProperty | Scalar, alternate: Scalar, scope?: valueof string);\n\n/**\n * To define the client scope of an operation.\n * @param scope The language scope you want this decorator to apply to. If not specified, will apply to all language emitters.\n * You can use \"!\" to specify negation such as \"!(java, python)\" or \"!java, !python\".\n *\n * @example\n * ```typespec\n * @scope(\"!csharp\")\n * op test: void;\n * ```\n */\nextern dec scope(target: Operation, scope?: valueof string);\n\n/**\n * Use to override default assumptions on whether a parameter is an api-version parameter or not.\n * By default, we do matches with the `api-version` or `apiversion` string in the parameter name. Since api versions are\n * a client parameter, we will also elevate this parameter up onto the client.\n * @param value If true, we will treat this parameter as an api-version parameter. If false, we will not. Default is true.\n * @param scope The language scope you want this decorator to apply to. If not specified, will apply to all language emitters.\n * You can use \"!\" to specify negation such as \"!(java, python)\" or \"!java, !python\".\n *\n * @example\n * ```typespec\n * namespace Contoso;\n *\n * op test(\n *   @apiVersion\n *   @header(\"x-ms-version\")\n *   version: string\n * ): void;\n * ```\n */\nextern dec apiVersion(target: ModelProperty, value?: valueof boolean, scope?: valueof string);\n",
  "lib/augmentCore.tsp": ""
};
const _TypeSpecLibrary_ = {
  jsSourceFiles: TypeSpecJSSources,
  typespecSourceFiles: TypeSpecSources,
};

export { $access, $alternateType, $apiVersion, $client, $clientInitialization, $clientName, $clientNamespace, $convenientAPI, $decorators, $flattenProperty, $lib, $onEmit, $operationGroup, $override, $paramAlias, $protocolAPI, $scope, $usage, $useSystemTextJsonConverter, UsageFlags, _TypeSpecLibrary_, addEncodeInfo, createDiagnostic, createSdkContext, createStateSymbol, createTCGCContext, getAccess, getAccessOverride, getAllModels, getAllModelsWithDiagnostics, getAllReferencedTypes, getAlternateType, getClient, getClientInitialization, getClientNameOverride, getClientNamespace, getClientNamespaceString, getClientType, getClientTypeWithDiagnostics, getCrossLanguageDefinitionId, getCrossLanguagePackageId, getDefaultApiVersion, getEffectivePayloadType, getGeneratedName, getHttpOperationExamples, getHttpOperationParameter, getHttpOperationWithCache, getInitializationType, getIsApiVersion, getKnownScalars, getLibraryName, getOperationGroup, getOverriddenClientMethod, getParamAlias, getPropertyNames, getSdkArrayOrDict, getSdkArrayOrDictWithDiagnostics, getSdkBuiltInType, getSdkConstant, getSdkCredentialParameter, getSdkDurationType, getSdkEnum, getSdkEnumValue, getSdkModel, getSdkModelPropertyType, getSdkModelPropertyTypeBase, getSdkModelWithDiagnostics, getSdkTuple, getSdkTupleWithDiagnostics, getSdkUnion, getSdkUnionEnum, getSdkUnionEnumWithDiagnostics, getSdkUnionWithDiagnostics, getTypeSpecBuiltInType, getUsage, getUsageOverride, getWireName, handleAllTypes, isApiVersion, isAzureCoreModel, isInOperationGroup, isOperationGroup, isPagedResultModel, isReadOnly, isSdkBuiltInKind, isSdkDateTimeEncodings, isSdkFloatKind, isSdkIntKind, listClients, listOperationGroups, listOperationsInOperationGroup, listSubClients, namespace, reportDiagnostic, shouldFlattenProperty, shouldGenerateConvenient, shouldGenerateProtocol };
